#!/usr/bin/env bash

function pause() {
  key=""
  echo -n Hit any key to continue....
  stty -icanon
  key=`dd count=1 2>/dev/null`
  stty icanon
}

function print_intro() {
  echo ""
  echo "  PegasusN: A Scalable and Versatile Graph Mining System"
  echo "  Last modified October 1st 2017"
  echo ""
  echo "  Authors: Ha-Myung Park, Chiwan Park, and U Kang"
  echo "           Dept. of Computer Science and Engineering, Seoul National University"
  echo ""
  echo "  Type \`help\` for available commands."
  echo "  The user manual is available at https://datalab.snu.ac.kr/pegasusn"
  echo ""
  echo ""
}

function help() {
  echo ""
	echo "    add [file or directory] [graph name]"
	echo "        upload a local graph file or directory to HDFS"
	echo "    del [graph_name]"
	echo "        delete a graph"
	echo "    list"
	echo "        list graphs"
	echo "    show [type of data] [graph name]"
	echo "        show sampled data of a graph"
	echo "    compute [algorithm] [graph name]"
	echo "        run an algorithm on a graph"
	echo "    generate [model]"
	echo "        generate a synthetic graph"
	echo "    plot [type of plot] [graph name]"
	echo "        generate plots"
	echo "    exit"
	echo "        exit PegasusN"
	echo "    help"
	echo "        show this screen"
	echo ""
}

function help_add() {
  echo "    add [file or directory] [graph name]"
  echo "        upload a local grpah file or directory to HDFS"
  echo "        the graph are stored in pegasus/graphs directory."
}

function list() {
	echo -e "=== GRAPH LIST ===\n"
	hdfs dfs -ls pegasus/graphs | grep "^d" | awk '{print $8}' | awk -F"/" '{ print "- "$NF}'
	echo ""
}

function add_graph() {
  local_path=$1
  graph_name=$2
  
  hdfs dfs -test -e pegasus/graphs/$graph_name
  graph_dir=$?
  if [[ $graph_dir -eq 0 ]]; then
    echo "Error: graph $graph_name already exists."
    return 1
  fi
  
  hdfs dfs -mkdir pegasus/graphs/$graph_name
  
  if [[ -f $local_path ]]; then
    hdfs dfs -put $local_path pegasus/graphs/$graph_name
    echo "Graph $graph_name added."
  elif [[ -d $local_path ]]; then
    hdfs dfs -put $local_path/* pegasus/graphs/$graph_name
    echo "Graph $graph_name added."
  else
    echo "Error: $local_path is not a regular file or directory."
  fi
}

function del_graph() {
  graph=$1
  
  hdfs dfs -test -e pegasus/graphs/$graph_name
  graph_dir=$?
  if [[ $graph_dir -eq 0 ]]; then
    hdfs dfs -rm -r pegasus/graphs/$graph
    echo "The $graph graph is deleted."
  fi
}

function show_graph() {
  graph=$1
  path=pegasus/graphs/$graph
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the graph $graph does not exist."
    return 1
  fi
  
  hdfs dfs -test -d $path
  is_dir=$?
  if [[ $is_dir -eq 0 ]]; then
    files=$(hdfs dfs -ls pegasus/graphs/$graph/* | grep "^-r" | awk '{print $8}' | grep -v '^_')
    first_file=$(echo $files | awk -F"/" 'NR==1{ print $NF}')
    file_path=$path/$first_file
  else
    file_path=$path
  fi
  
  hdfs dfs -tail $file_path | tail -n 15
}

function show_pagerank() {
  graph=$1
  path=pegasus/pagerank/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the PageRank score for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-m-00000 | tail -n 15
}

function show_cc() {
  graph=$1
  path=pegasus/cc/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the connected component computation result for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-r-00000 | tail -n 15
}

function show_sssp() {
  graph=$1
  path=pegasus/sssp/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the shortest path computation result for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-m-00000 | tail -n 15
}

function show_diameter() {
  graph=$1
  path=pegasus/diameter/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the diameter estimation result for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-m-00000 | tail -n 15
}

function show_triangle() {
  graph=$1
  path=pegasus/triangle/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the diameter estimation result for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-m-00000 | tail -n 15
}

function show_subgraph() {
  graph=$1
  path=pegasus/subgraph/$1
  hdfs dfs -test -e $path
  exist=$?
  if [[ $exist -ne 0 ]]; then
    echo "Error: the diameter estimation result for $graph graph does not exist."
    return 1
  fi
  
  hdfs dfs -tail $path/part-m-00000 | tail -n 15
}

function main_loop() {
  while [ 1 -eq 1 ]; do
    echo -n "PegasusN> "
    read CMD REST
    case "$CMD" in
    "")
      continue ;;
    "add")
      add_graph $REST
      continue ;;
    "del")
      del_graph $REST
      continue ;;
    "compute")
      compute $REST
      continue ;;
    "generate")
      generate $REST
      continue ;;
    "plot")
      plot $REST
      continue ;;
    "help")
      case "$REST" in
        "") help ;;
        "add") help_add ;;
      esac
      continue ;;
    "list")
      list
      continue ;;
    "show")
      type=$(echo $REST | head -n1 | cut -d " " -f1)
      graph=$(echo $REST | head -n1 | cut -d " " -f2)
      case "$type" in
      "graph")
        show_graph $graph
        continue ;;
      "pagerank")
        show_pagerank $graph
        continue ;;
      "cc")
        show_cc $graph
        continue ;;
      "sssp")
        show_sssp $graph
        continue ;;
      "diameter")
        show_diameter $graph
        continue ;;
      "triangle")
        show_triangle $graph
        continue ;;
      "subgraph")
        show_subgraph $graph
        continue ;;
      esac
      continue ;;
    "exit")
      break ;;
    *)
      echo "Invalid command. Type \`help\` for available commands." ;;
    esac
  done
}
