/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: Finalization.java
 * - cc-computation step of pacc.
 * Version: 3.0
 */

package pegasus.mapreduce.cc;

import pegasus.mapreduce.common.io.LongPairWritable;
import gnu.trove.impl.Constants;
import gnu.trove.map.hash.TLongLongHashMap;
import gnu.trove.procedure.TLongProcedure;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapred.Task.Counter;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;

import java.io.IOException;

public class Finalization extends Configured implements Tool{
	
	private final Path output;
	private final String title;
	private final boolean verbose;
	private final int numRounds;
	public long inputSize;

	/**
	 * constructor
	 * @param output file path
	 * @param numRounds the number of rounds conducted until now.
	 * @param verbose if true, log is printed verbosely.
	 */
	public Finalization(Path output, int numRounds, boolean verbose){
		this.output = output;
		this.verbose = verbose;
		this.numRounds = numRounds;
		this.title = String.format("[%s]%s", this.getClass().getSimpleName(), output.getName());
	}

	/**
	 * submit the hadoop job
	 * @param args not used
	 * @return not used
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception{
		
		Job job = new Job(getConf(), title);
		
		job.setJarByClass(this.getClass());
		
		job.setMapOutputKeyClass(IntWritable.class);
		job.setMapOutputValueClass(LongPairWritable.class);
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		
		job.setMapperClass(LargeStarMapper.class);
		job.setReducerClass(LargeStarReducer.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
		
		FileSystem fs = FileSystem.get(getConf());
		
		
		for(int i=1; i<=numRounds; i++){
			if(fs.exists(output.suffix("_" + i + "/in"))){
				FileInputFormat.addInputPath(job, output.suffix("_" + i + "/in"));
			}
			if(fs.exists(output.suffix("_large_" + i + "/final"))){
				FileInputFormat.addInputPath(job, output.suffix("_large_" + i + "/final"));
			}
			
		}
		
		if(fs.exists(output.suffix("_" + numRounds + "/out"))){
			FileInputFormat.addInputPath(job, output.suffix("_" + numRounds + "/out"));
		}
		
		
		FileOutputFormat.setOutputPath(job, output);
		
		FileSystem.get(getConf()).delete(output, true);
		
		job.waitForCompletion(verbose);
		
		inputSize = job.getCounters().findCounter(Counter.MAP_INPUT_RECORDS).getValue();
		
		return 0;
	}

	static public class LargeStarMapper extends Mapper<LongWritable, LongWritable, IntWritable, LongPairWritable>{

		int numPartitions;

		/**
		 * setup before execution
		 * @param context hadoop context
		 * @throws IOException exception by hadoop
		 * @throws InterruptedException exception by hadoop
		 */
		@Override
		protected void setup(
				Context context)
				throws IOException, InterruptedException{
			
			numPartitions = context.getConfiguration().getInt("numPartitions", 0);
		}

		IntWritable p = new IntWritable();
		LongPairWritable edge = new LongPairWritable();

		/**
		 * the map function.
		 * @param u source node
		 * @param v destination node
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(LongWritable u, LongWritable v,
				Context context)
				throws IOException, InterruptedException{
			long uu = u.get();
			long vv = v.get();
			
			uu = uu < 0 ? -uu - 1: uu;
			vv = vv < 0 ? -vv - 1: vv;
			
			p.set((int) (uu % numPartitions));
			edge.set(uu, vv);
			context.write(p, edge);
		}
		
	}
	
	static public class LargeStarReducer extends Reducer<IntWritable, LongPairWritable, LongWritable, LongWritable>{

		Context context;
		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();
		
		TLongLongHashMap parent;

		/**
		 * the reduce function.
		 * It find connected components in a partition.
		 * @param key the partition number. not used.
		 * @param values edges
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void reduce(IntWritable key, Iterable<LongPairWritable> values,
				Context context)
				throws IOException, InterruptedException{
			
			parent = new TLongLongHashMap(Constants.DEFAULT_CAPACITY,
					Constants.DEFAULT_LOAD_FACTOR, -1, -1);
			
			this.context = context;

			for(LongPairWritable edge : values){
				long u = edge.i;
				long v = edge.j;
				
				if(find(u) != find(v)){
					union(u, v);
				}
				
			}

			final Context ctx = context;

			parent.forEachKey(u -> {
                try{
                    long v = find(u);
                    if(u != v){
                        ou.set(u);
                        ov.set(v);
                        ctx.write(ou, ov);
                    }
                } catch (Exception ignored){}
                return true;
            });
			
		}

		/**
		 * the find function of the union-find algorithm.
		 * parent table is updated as a side effect
		 * @param x node to process
		 * @return component id (smallest node reachable to node x)
		 */
		public long find(long x) {

			long p = parent.get(x);
			if(p != -1){
				long new_p = find(p);
				parent.put(x, new_p);
				return new_p;
			}
			else{
				return x;
			}

		}

		/**
		 * the union function of the union-find algorithm.
		 * It updates parent table to union the two
		 * @param x a node
		 * @param y another node
		 */
		public void union(long x, long y) {
			long r1 = find(x);
			long r2 = find(y);

			if(r1 > r2){
				parent.put(r1, r2);
			}
			else if(r1 < r2){
				parent.put(r2, r1);
			}
		}

		
	}
	
}
