package pegasus.mapreduce.cc.old;

import gnu.trove.impl.Constants;
import gnu.trove.map.hash.TLongLongHashMap;
import gnu.trove.procedure.TLongProcedure;

import java.io.IOException;

import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapred.Task.Counter;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;
import pegasus.mapreduce.common.io.LongPairWritable;

public class Finalization extends Configured implements Tool{
	
	private final Path output;
	private final String title;
	private final boolean verbose;
	private final int numRounds;
	public long inputSize;
	
	
	public Finalization(Path output, int numRounds, boolean verbose){
		this.output = output;
		this.verbose = verbose;
		this.numRounds = numRounds;
		this.title = String.format("[%s]%s", this.getClass().getSimpleName(), output.getName());
	}
	
	public int run(String[] args) throws Exception{
		
		Job job = new Job(getConf(), title);
		
		job.setJarByClass(this.getClass());
		
		job.setMapOutputKeyClass(IntWritable.class);
		job.setMapOutputValueClass(LongPairWritable.class);
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		
		job.setMapperClass(LargeStarMapper.class);
		job.setReducerClass(LargeStarReducer.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
		
		FileSystem fs = FileSystem.get(getConf());
		
		
		for(int i=1; i<=numRounds; i++){
			if(fs.exists(output.suffix("_" + i + "/in"))){
				FileInputFormat.addInputPath(job, output.suffix("_" + i + "/in"));
			}
			if(fs.exists(output.suffix("_large_" + i + "/in"))){
				FileInputFormat.addInputPath(job, output.suffix("_large_" + i + "/in"));
			}
			if(fs.exists(output.suffix("_large_" + i + "/final"))){
				FileInputFormat.addInputPath(job, output.suffix("_large_" + i + "/final"));
			}
			
		}
		
		if(fs.exists(output.suffix("_" + numRounds + "/out"))){
			FileInputFormat.addInputPath(job, output.suffix("_" + numRounds + "/out"));
		}
		
		
		FileOutputFormat.setOutputPath(job, output);
		
		FileSystem.get(getConf()).delete(output, true);
		
		job.waitForCompletion(verbose);
		
		inputSize = job.getCounters().findCounter(Counter.MAP_INPUT_RECORDS).getValue();
		
		return 0;
	}

	static public class LargeStarMapper extends Mapper<LongWritable, LongWritable, IntWritable, LongPairWritable>{

		int numPartitions; 
		
		@Override
		protected void setup(
				Context context)
				throws IOException, InterruptedException{
			
			numPartitions = context.getConfiguration().getInt("numPartitions", 0);
		}

		IntWritable p = new IntWritable();
		LongPairWritable edge = new LongPairWritable();
		
		@Override
		protected void map(LongWritable u, LongWritable v,
				Context context)
				throws IOException, InterruptedException{
			long uu = u.get();
			long vv = v.get();
			
			uu = uu < 0 ? -uu - 1: uu;
			vv = vv < 0 ? -vv - 1: vv;
			
			p.set((int) (uu % numPartitions));
			edge.set(uu, vv);
			context.write(p, edge);
		}
		
	}
	
	static public class LargeStarReducer extends Reducer<IntWritable, LongPairWritable, LongWritable, LongWritable>{

		Context context;
		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();
		
		TLongLongHashMap parent;
		
		@Override
		protected void reduce(IntWritable key, Iterable<LongPairWritable> values,
				Context context)
				throws IOException, InterruptedException{
			
			parent = new TLongLongHashMap(Constants.DEFAULT_CAPACITY,
					Constants.DEFAULT_LOAD_FACTOR, -1, -1);
			
			this.context = context;

			for(LongPairWritable edge : values){
				long u = edge.i;
				long v = edge.j;
				
				if(find(u) != find(v)){
					union(u, v);
				}
				
			}

			final Context ctx = context;

			parent.forEachKey(new TLongProcedure() {

				public boolean execute(long u) {
					try{
						long v = find(u);
						if(u != v){
							ou.set(u);
							ov.set(v);
							ctx.write(ou, ov);
						}
					} catch (Exception ignored){}
					return true;
				}
			});
			
		}
		
		public long find(long x) {

			long p = parent.get(x);
			if(p != -1){
				long new_p = find(p);
				parent.put(x, new_p);
				return new_p;
			}
			else{
				return x;
			}

		}

		public void union(long x, long y) {
			long r1 = find(x);
			long r2 = find(y);

			if(r1 > r2){
				parent.put(r1, r2);
			}
			else if(r1 < r2){
				parent.put(r2, r1);
			}
		}

		
	}
	
}
