package pegasus.mapreduce.cc.old;

import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.TaskCounter;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.LazyOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.MultipleOutputs;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import java.io.*;
import java.util.Arrays;


public class PALargeStarOpt extends Configured implements Tool{
	
	private final Path input;
	private final Path output;
	private final String title;
	private final boolean verbose;
	public long numChanges;
	public long inputSize;
	public long outSize;
	public long ccSize;
	
	
	public PALargeStarOpt(Path input, Path output, boolean verbose){
		this.input = input;
		this.output = output;
		this.verbose = verbose;
		this.title = String.format("[%s]%s", this.getClass().getSimpleName(), output.getName());
	}
	
	public static void main(String[] args) throws Exception{
		
		Path input = new Path(args[0]);
		Path output = new Path(args[1]);
		
		ToolRunner.run(new PALargeStarOpt(input, output, true), args);
	}

	public int run(String[] args) throws Exception{
		
		Job job = new Job(getConf(), title);
		
		job.setJarByClass(this.getClass());
		
		job.setMapOutputKeyClass(LongWritable.class);
		job.setMapOutputValueClass(LongWritable.class);
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		
		job.setMapperClass(ColorLargeStarMapper.class);
		job.setReducerClass(ColorLargeStarReducer.class);
		job.setInputFormatClass(SequenceFileInputFormat.class);
//		job.setOutputFormatClass(SequenceFileOutputFormat.class);
		LazyOutputFormat.setOutputFormatClass(job, SequenceFileOutputFormat.class);
		
		
		FileInputFormat.addInputPath(job, input);
		FileOutputFormat.setOutputPath(job, output);
		
		FileSystem fs = FileSystem.get(getConf());
		
		fs.delete(output, true);
		
		
		if(fs.exists(input)){
			job.waitForCompletion(verbose);
			this.numChanges = job.getCounters().findCounter(Counters.NUM_CHANGES).getValue();
			this.inputSize = job.getCounters().findCounter(TaskCounter.MAP_INPUT_RECORDS).getValue();
			this.outSize = job.getCounters().findCounter(Counters.OUT_SIZE).getValue();
			this.ccSize = job.getCounters().findCounter(Counters.CC_SIZE).getValue();
		}
		
		return 0;
	}

	static public class ColorLargeStarMapper extends Mapper<LongWritable, LongWritable, LongWritable, LongWritable>{

		
		@Override
		protected void map(LongWritable u, LongWritable v,
				Context context)
				throws IOException, InterruptedException{
			
			if(u.get() >= 0) context.write(u, v);
			if(v.get() >= 0) context.write(v, u);
		}
		
	}
	
	static public class ColorLargeStarReducer extends Reducer<LongWritable, LongWritable, LongWritable, LongWritable>{

		private int numPartitions;
		long[] min;
		MultipleOutputs<LongWritable, LongWritable> mout;
		
		
		
		@Override
		protected void cleanup(
				Context context)
				throws IOException, InterruptedException {
			
			mout.close();
		}

		@Override
		protected void setup(
				Context context)
				throws IOException, InterruptedException{
			
			numPartitions = context.getConfiguration().getInt("numPartitions", 1);
			min = new long[numPartitions];
			
			mout = new MultipleOutputs<LongWritable, LongWritable>(context);
			
		}

		LongWritable om = new LongWritable();
		LongWritable ov = new LongWritable();
		
		@Override
		protected void reduce(LongWritable key, Iterable<LongWritable> values,
				Context context)
				throws IOException, InterruptedException{
			
			
			long u = key.get();
			int uPartition = (int) (u % numPartitions);
			long numChanges = 0;
			boolean isStar = true;
			
			long outSize = 0;
			long ccSize = 0;
			
			Arrays.fill(min, Long.MAX_VALUE);
			
			min[uPartition] = u;
			long min_global = u;
			
			String tmpFileName = "pacc.largestar.tmp-" + context.getTaskAttemptID();
			
			DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(tmpFileName)));

			for(LongWritable _v : values){
				long v = _v.get();
				
				if(v < 0) v = -v - 1;
				else isStar = false;
				
				int vPartition = (int) (v % numPartitions);
				
				if(v < min[vPartition]) min[vPartition] = v;
				if(v < min_global) min_global = v;
				if(u < v) out.writeLong(v);
			}
			
			out.close();
			
			
			DataInputStream in = new DataInputStream(new BufferedInputStream(new FileInputStream(tmpFileName)));
			
			if(isStar){
				
				try{
					
					om.set(u);
					
					while(true){
						long v = in.readLong();
						ov.set(v);
						mout.write(ov, om, "final/part");
						ccSize++;
					}
					
				}catch(EOFException e){}
				
			}
			else{
				
				try{
					
					while(true){
						long v = in.readLong();
						int vPartition = (int) (v % numPartitions);
						long min_local = min[vPartition];
						
						ov.set(v);
						
						if(min_local != v){
							om.set(min_local);
							mout.write(ov, om, "out/part");
							outSize++;
						}
						else if(v != min_global){
							om.set(min_global);
							mout.write(ov, om, "out/part");
							outSize++;
						}
						
						if(om.get() != u) numChanges++;
					}
					
				}catch(EOFException e){}
				
			}
			
			
			
			in.close();
			
			context.getCounter(Counters.NUM_CHANGES).increment(numChanges);
			context.getCounter(Counters.CC_SIZE).increment(ccSize);
			context.getCounter(Counters.OUT_SIZE).increment(outSize);
			
			new File(tmpFileName).delete();
			
						
		}
		
	}

	
}
