/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BaseMatrixCache.java
 * - Represents a matrix block contains matrix elements
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import com.esotericsoftware.kryo.io.KryoDataOutput;
import com.esotericsoftware.kryo.io.Output;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.LocalFileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.compress.CodecPool;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.Compressor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.types.IntIntPairWritable;
import pegasus.mapreduce.common.utils.IOUtils;
import pegasus.mapreduce.common.utils.MemoryUtils;

import java.io.IOException;
import java.io.OutputStream;

/**
 * Represents a matrix block contains matrix elements.
 *
 * @param <M> The type parameter of matrix values
 */
public abstract class BaseMatrixCache<M extends Writable> implements MatrixCache<M> {
  private static final Logger LOG = LoggerFactory.getLogger(BaseMatrixCache.class);

  protected final Configuration conf;

  protected final IntArrayList colList;
  protected final IntArrayList rowList;
  protected final IntArrayList rowSizes;
  protected final IntArrayList degrees;

  private final boolean memoryFrozen;

  protected final Path path;
  private Output out;
  private KryoDataOutput dataOut;
  private Compressor compressor;

  protected final IntIntPairWritable col = new IntIntPairWritable();

  protected BaseMatrixCache(
      final String blockName, final Configuration conf, final int numCols, final int numRows,
      final int sizePerValue) throws IOException {
    // TODO: calling GC manually may cause the performance degrade.
    MemoryUtils.callGCMultipleTimes();
    this.conf = conf;

    final long expectedSize = 12 * (long) numCols + (4 + sizePerValue) * (long) numRows;
    final long availableMemory = (long) ((
        MemoryUtils.availableMemoryBytes() - Constants.SizeOfMemoryBarrier) * 0.6);
    if (expectedSize < availableMemory) {
      memoryFrozen = false;
      LOG.info("Matrix block {} is stored in memory. (expectedSize: {} MiB, availableMemory: {} MiB",
          blockName, MemoryUtils.bytesToMebibytes(expectedSize),
          MemoryUtils.bytesToMebibytes(availableMemory));

      colList = new IntArrayList(numCols);
      rowList = new IntArrayList(numRows);
      rowSizes = new IntArrayList(numCols);
      degrees = new IntArrayList(numCols);

      path = null;
      out = null;
      dataOut = null;
      compressor = null;
    } else {
      memoryFrozen = true;
      LOG.info("Matrix block {} is stored in disk. (expectedSize: {} MiB, availableMemory: {} MiB",
          blockName, MemoryUtils.bytesToMebibytes(expectedSize),
          MemoryUtils.bytesToMebibytes(availableMemory));

      colList = null;
      rowList = null;
      rowSizes = null;
      degrees = null;

      path = IOUtils.getLocalTemporaryPath(blockName);

      final LocalFileSystem localFS = FileSystem.getLocal(conf);
      final CompressionCodec codec = IOUtils.getPreferredCompressionCodec(conf);
      final OutputStream rawStream = localFS.create(path, (short) 1);
      compressor = CodecPool.getCompressor(codec);
      out = new Output(codec.createOutputStream(rawStream, compressor));
      dataOut = new KryoDataOutput(out);
    }
  }

  /**
   * Puts a column of matrix.
   * <p>
   * Note that this method first puts
   *
   * @param col  The metadata of column
   * @param rows The elements in the column
   * @throws IOException
   */
  @Override
  public void put(final IntIntPairWritable col, final M rows) throws IOException {
    if (!memoryFrozen) {
      // store to memory
      putToMemory(col, rows);
    } else {
      // store to disk
      col.write(dataOut);
      rows.write(dataOut);
    }
  }

  /**
   * Puts a column of matrix into memory store.
   *
   * @param col  The metadata of column
   * @param rows The elements in the column
   */
  protected abstract void putToMemory(final IntIntPairWritable col, final M rows);

  /**
   * Closes memory store.
   */
  @Override
  public void freeze() {
    if (out != null) {
      out.close();
      CodecPool.returnCompressor(compressor);

      out = null;
      dataOut = null;
      compressor = null;
    }
  }
}
