/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BaseMatrixIterator.java
 * - Iterates over all elements in the given matrix block.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.KryoDataInput;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.LocalFileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.compress.CodecPool;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.Decompressor;
import pegasus.mapreduce.common.utils.IOUtils;

import java.io.IOException;
import java.io.InputStream;

/**
 * Iterates over all elements in the given matrix block.
 * 
 * @param <V> The type parameter of matrix elements
 */
public abstract class BaseMatrixIterator<V extends Writable> implements MatrixIterator<V> {
  protected final Input in;
  protected final KryoDataInput dataIn;
  private final Decompressor decompressor;

  public BaseMatrixIterator(final Path path, final Configuration conf) throws IOException {
    if (path != null) {
      final LocalFileSystem localFS = FileSystem.getLocal(conf);
      final CompressionCodec codec = IOUtils.getPreferredCompressionCodec(conf);
      final InputStream rawStream = localFS.open(path);
      decompressor = CodecPool.getDecompressor(codec);
      in = new Input(codec.createInputStream(rawStream, decompressor));
      dataIn = new KryoDataInput(in);
    } else {
      in = null;
      dataIn = null;
      decompressor = null;
    }
  }

  @Override
  public void close() throws Exception {
    if (in != null) {
      in.close();
      CodecPool.returnDecompressor(decompressor);
    }
  }
}
