/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BaseVectorCache.java
 * - Abstract class for writable vectors.
 * Version: 3.0
 */

package pegasus.mapreduce.common.cache;

import org.apache.hadoop.io.Writable;

/**
 * Base implementation of vector cache
 *
 * @param <V> The type of vector element
 */
public abstract class BaseVectorCache<V extends Writable> implements VectorCache<V> {
  private BitSet containsSet;
  protected int size;
  protected int numBlocks;
  protected long numVertices;
  protected int blockId;

  public BaseVectorCache(final int numBlocks, final int blockId, final long numVertices) {
    this.numBlocks = numBlocks;
    this.numVertices = numVertices;
    this.blockId = blockId;
    this.size = (int) Math.ceil(numVertices / (double) numBlocks);
    this.containsSet = new BitSet(size);
  }

  /**
   * Returns the size of vector
   *
   * @return The size of vector
   */
  @Override
  public int size() {
    return size;
  }

  /**
   * Returns whether the i-th element is in the vector or not.
   *
   * @param i The index of element
   * @return true if the i-th element is in the vector, otherwise false
   */
  @Override
  public boolean contains(final int i) {
    return containsSet.contains(i);
  }

  /**
   * Puts an element.
   *
   * @param i     The index of element
   * @param value The value of element
   */
  @Override
  public void put(int i, V value) {
    containsSet.add(i);
  }

  /**
   * Adds the given value to the i-th element.
   *
   * @param i     The index of element
   * @param value The value to be added
   */
  @Override
  public void addTo(int i, V value) {
    containsSet.add(i);
  }

  /**
   * Clears all elements in the vector
   */
  @Override
  public void clear() {
    containsSet.clear();
  }
}
