/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BitSet.java
 * - bit vector.
 * Version: 3.0
 */

package pegasus.mapreduce.common.cache;

import java.util.Arrays;

/**
 * Represents a bit set.
 */
public final class BitSet {
  private static final long[] MASKS;
  private final long[] bits;
  private final int cardinality;

  public BitSet(final int cardinality) {
    final int adjustedSize = (int) Math.ceil(cardinality / 64.0);
    this.cardinality = cardinality;
    this.bits = new long[adjustedSize];
  }

  /**
   * Returns the cardinality of set.
   * <p>
   * Note that the cardinality is different from the size of array because each element in the array
   * contains multiple elements.
   *
   * @return The cardinality of set
   */
  public final int getCardinality() {
    return cardinality;
  }

  /**
   * Returns the size of array.
   *
   * @return The size of array
   */
  public final int getArraySize() {
    return bits.length;
  }

  /**
   * Returns whether the set contains a given id or not.
   *
   * @param id The id to be checked
   * @return true if the set contains the given id, otherwise false
   */
  public final boolean contains(final int id) {
    final int pos = id / 64;
    final long mask = MASKS[id % 64];

    return (bits[pos] & mask) != 0;
  }

  /**
   * Adds the given id to the set.
   *
   * @param id The id to be added
   */
  public final void add(final int id) {
    final int pos = id / 64;
    final long mask = MASKS[id % 64];

    bits[pos] |= mask;
  }

  /**
   * Removes the given id from set.
   *
   * @param id The id to be removed
   */
  public final void remove(final int id) {
    final int pos = id / 64;
    final long mask = MASKS[id % 64];

    bits[pos] &= ~mask;
  }

  /**
   * Clears the set.
   */
  public final void clear() {
    Arrays.fill(bits, 0L);
  }

  static {
    MASKS = new long[64];
    for (int i = 0; i < 64; ++i) {
      MASKS[i] = 1L << i;
    }
  }
}
