/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BitStringsVectorCache.java
 * - Represents a vector block contains bitstring values.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.types.BitStrings;
import pegasus.mapreduce.common.utils.IDUtils;

import java.util.Arrays;

/**
 * Represents a vector block contains bitstring values.
 */
public final class BitStringsVectorCache extends BaseVectorCache<BitStrings> {
  private final long[][] values;
  private final long[] zeroValue;
  private final BitStrings value = new BitStrings();

  public BitStringsVectorCache(
      final int numBlocks, final int blockId, final long numVertices, final BitStrings zeroValue) {
    super(numBlocks, blockId, numVertices);

    this.values = new long[Constants.DefaultNumBitstrings][size];
    this.zeroValue = new long[Constants.DefaultNumBitstrings];
    System.arraycopy(zeroValue.getBits(), 0, this.zeroValue, 0, Constants.DefaultNumBitstrings);
    clear();
  }

  /**
   * {@inheritDoc}
   *
   * @param i The index of element to be retrieved
   * @return
   */
  @Override
  public final BitStrings get(final int i) {
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      value.setBit(k, values[k][i]);
    }
    return value;
  }

  /**
   * {@inheritDoc}
   *
   * @param i     The index of element
   * @param value The value to be put
   */
  @Override
  public final void put(final int i, final BitStrings value) {
    super.put(i, value);
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      values[k][i] = value.getBit(k);
    }
  }

  /**
   * {@inheritDoc}
   *
   * @param i     The index of element
   * @param value The value to be added
   */
  @Override
  public final void addTo(final int i, final BitStrings value) {
    super.put(i, value);
    for (int k = 0; k < Constants.DefaultNumBitstrings; ++k) {
      values[k][i] |= value.getBit(k);
    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public final void clear() {
    super.clear();
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      Arrays.fill(values[i], zeroValue[i]);
    }
  }

  /**
   * {@inheritDoc}
   *
   * @param onlyNonZero The flag to iterate only non-zero elements
   * @param filterSet   The set to filter elements to be iterated
   * @return
   */
  @Override
  public VectorIterator<BitStrings> iterator(final boolean onlyNonZero, final BitSet filterSet) {
    return new VectorIterator<BitStrings>() {
      private int pos = 0;
      private int id;
      private final BitStrings value = new BitStrings();

      @Override
      public final boolean next() {
        while (pos < size && IDUtils.localToGlobal(numBlocks, blockId, pos) < numVertices) {
          if (contains(pos) &&
              (filterSet == null || filterSet.contains(pos)) &&
              (!onlyNonZero || !isEmpty())) {
            id = pos;
            for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
              value.setBit(i, values[i][pos]);
            }
            ++pos;
            return true;
          }
          ++pos;
        }
        return false;
      }

      @Override
      public final int currentId() {
        return id;
      }

      @Override
      public final BitStrings currentValue() {
        return value;
      }

      private boolean isEmpty() {
        for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
          if (values[i][pos] != 0) {
            return false;
          }
        }
        return true;
      }
    };
  }
}

