/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: DoubleVectorCache.java
 * - Stores (long, double) pairs based on a resizable, primitive hash table.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import org.apache.hadoop.io.DoubleWritable;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.IDUtils;

import java.util.Arrays;

public final class DoubleVectorCache extends BaseVectorCache<DoubleWritable> {
  private final double[] values;
  private final double zeroValue;
  private final DoubleWritable value = new DoubleWritable();

  public DoubleVectorCache(final int numBlocks, final int blockId, final long numVertices, DoubleWritable zeroValue) {
    super(numBlocks, blockId, numVertices);

    this.values = new double[size];
    this.zeroValue = zeroValue.get();
    clear();
  }

  /**
   * {@inheritDoc}
   * @param i The index of element to be retrieved
   * @return
   */
  @Override
  public final DoubleWritable get(final int i) {
    value.set(values[i]);
    return value;
  }

  /**
   * {@inheritDoc}
   * @param i     The index of element
   * @param value The value to be put
   */
  @Override
  public final void put(final int i, final DoubleWritable value) {
    super.put(i, value);
    values[i] = value.get();
  }

  /**
   * {@inheritDoc}
   * @param i     The index of element
   * @param value The value to be added
   */
  @Override
  public void addTo(final int i, final DoubleWritable value) {
    super.addTo(i, value);
    values[i] += value.get();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void clear() {
    super.clear();
    Arrays.fill(values, zeroValue);
  }

  /**
   * {@inheritDoc}
   * @param onlyNonZero The flag to iterate only non-zero elements
   * @param filterSet   The set to filter elements to be iterated
   * @return
   */
  @Override
  public VectorIterator<DoubleWritable> iterator(final boolean onlyNonZero, final BitSet filterSet) {
    return new VectorIterator<DoubleWritable>() {
      private int pos = 0;
      private int id;
      private final DoubleWritable value = new DoubleWritable();

      @Override
      public final boolean next() {
        while (pos < size && IDUtils.localToGlobal(numBlocks, blockId, pos) < numVertices) {
          if (contains(pos) &&
              (filterSet == null || filterSet.contains(pos)) &&
              (!onlyNonZero || Math.abs(values[pos] - zeroValue) > Constants.DoubleValueZeroThreshold)) {
            id = pos;
            value.set(values[pos]);
            ++pos;
            return true;
          }
          ++pos;
        }

        return false;
      }

      @Override
      public final int currentId() {
        return id;
      }

      @Override
      public final DoubleWritable currentValue() {
        return value;
      }
    };
  }
}
