/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: FileUtils.java
 * - Provides some utility methods manipulating files.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.DefaultCodec;
import org.apache.hadoop.io.compress.SnappyCodec;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.util.ReflectionUtils;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * An utility class to manipulate files.
 */
public class FileUtils {
  private static CompressionCodec preferredCodec = null;

  /**
   * Returns a temporary directory path.
   *
   * @param name    The name of directory
   * @param context Hadoop context to create new directory
   * @return A temporary directory path
   */
  public static Path getLocalTemporaryPath(final String name, final TaskAttemptContext context) {
    return new Path(System.getProperty("java.io.tmpdir"), new Path(context.getJobID().toString(), name));
  }

  /**
   * Creates an input stream from the given path of a file. This method assumes that the file is
   * compressed by the preferred compression codec.
   *
   * @param path    The path of a file
   * @param context Hadoop context to open a file
   * @return An input stream from the given file
   * @throws IOException
   */
  public static DataInputStream getTemporaryFileInputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataInputStream(codec.createInputStream(fs.open(path)));
  }

  /**
   * Creates an output stream to the given path of a file. This method compresses the values to be stored.
   *
   * @param path    The path of a file
   * @param context Hadoop context to create a file
   * @return An output stream to the given file
   * @throws IOException
   */
  public static DataOutputStream getTemporaryFileOutputStream(final Path path, final TaskAttemptContext context)
      throws IOException {
    final FileSystem fs = FileSystem.getLocal(context.getConfiguration());
    final CompressionCodec codec = getPreferredCompressionCodec(context);

    return new DataOutputStream(codec.createOutputStream(fs.create(path, (short) 1)));
  }

  /**
   * Returns the preferred compression codec.
   * <p>
   * This method returns SnappyCodec object if it is available, otherwise DefaultCodec object is returned.
   *
   * @param conf Hadoop configuration object
   * @return A compression codec object
   */
  public static CompressionCodec getPreferredCompressionCodec(final Configuration conf) {
    if (preferredCodec != null) {
      return preferredCodec;
    }
    Class<? extends CompressionCodec> codecClass;
    if (SnappyCodec.isNativeCodeLoaded()) {
      codecClass = SnappyCodec.class;
    } else {
      codecClass = DefaultCodec.class;
    }

    preferredCodec = ReflectionUtils.newInstance(codecClass, conf);
    return preferredCodec;
  }

  /**
   * Returns the preferred compression codec.
   *
   * See {getPreferredCompressionCodec(Configuration)}.
   *
   * @param context Hadoop context object
   * @return A compression codec object
   */
  public static CompressionCodec getPreferredCompressionCodec(final TaskAttemptContext context) {
    return getPreferredCompressionCodec(context.getConfiguration());
  }
}
