/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: UnweightedMatrixCache.java
 * - Contains elements of matrix without weight.
 * Version: 3.0
 */
package pegasus.mapreduce.common.cache;

import org.apache.hadoop.conf.Configuration;
import pegasus.mapreduce.common.types.IntArrayWritable;
import pegasus.mapreduce.common.types.IntIntPairWritable;

import java.io.IOException;

public class UnweightedMatrixCache extends BaseMatrixCache<IntArrayWritable> {
  public UnweightedMatrixCache(
      final String blockName, final Configuration conf, final int numCols, final int numRows) throws IOException {
    super(blockName, conf, numCols, numRows, 0);
  }

  /**
   * {@inheritDoc}
   *
   * @param col  The metadata of column
   * @param rows The elements in the column
   */
  @Override
  protected void putToMemory(final IntIntPairWritable col, final IntArrayWritable rows) {
    colList.add(col.getFirst());
    degrees.add(col.getSecond());
    rowList.addElements(rowList.size(), rows.getValues(), rows.begin(), rows.size());
    rowSizes.add(rowList.size());
  }

  /**
   * {@inheritDoc}
   *
   * @return
   * @throws IOException
   */
  @Override
  public MatrixIterator<IntArrayWritable> iterator() throws IOException {
    return new BaseMatrixIterator<IntArrayWritable>(path, conf) {
      private int readCol = 0;
      private int readRow = 0;
      private final IntIntPairWritable col = new IntIntPairWritable();
      private final IntArrayWritable rows = new IntArrayWritable();

      @Override
      public boolean next() throws IOException {
        if (in == null && readCol < colList.size()) {
          // read from memory
          col.set(colList.getInt(readCol), degrees.getInt(readCol));
          rows.set(rowList.elements(), readRow, rowSizes.getInt(readCol));
          ++readCol;
          readRow += rows.size();
          return true;
        } else if (in != null && !in.eof()) {
          // read from disk
          col.readFields(dataIn);
          rows.readFields(dataIn);
          return true;
        }
        return false;
      }

      @Override
      public int currentCol() {
        return col.getFirst();
      }

      @Override
      public int numNonzeros() {
        return col.getSecond();
      }

      @Override
      public IntArrayWritable currentRows() {
        return rows;
      }
    };
  }
}
