/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: ComputationMapper.java
 * - Provides a base implementation of PMV algorithm
 * Version: 3.0
 */
package pegasus.mapreduce.common.computation;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.compress.CodecPool;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.RecordWriter;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;
import org.apache.zookeeper.KeeperException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.cache.BitSet;
import pegasus.mapreduce.common.cache.MatrixCache;
import pegasus.mapreduce.common.cache.VectorCache;
import pegasus.mapreduce.common.cache.VectorIterator;
import pegasus.mapreduce.common.coordination.Barrier;
import pegasus.mapreduce.common.coordination.CoordinationService;
import pegasus.mapreduce.common.coordination.Counter;
import pegasus.mapreduce.common.io.ComputationTaskSplit;
import pegasus.mapreduce.common.io.MatrixRecordReader;
import pegasus.mapreduce.common.types.IntIntPairWritable;
import pegasus.mapreduce.common.utils.IDUtils;
import pegasus.mapreduce.common.utils.IOUtils;
import pegasus.mapreduce.common.utils.MemoryUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * A base class for PMV algorithm.
 *
 * @param <M>  The type parameter for matrix elements
 * @param <V>  The type parameter for input vector elements
 * @param <PV> The type parameter for intermediate vector elements
 */
public abstract class ComputationMapper
    <M extends Writable, V extends Writable, PV extends Writable, OV extends Writable>
    extends Mapper<NullWritable, NullWritable, LongWritable, OV> {
  private final Logger LOG = LoggerFactory.getLogger(ComputationMapper.class);

  private final CoordinationService coordService = new CoordinationService();

  private MatrixCache<M>[] denseBlocks;
  private MatrixCache<M>[] sparseBlocks;

  private long numVertices;
  private int denseThreshold;

  private Context context;

  private VectorCache<V> vecInput;
  private VectorCache<PV> vecIntermediate;
  private VectorCache<V> vecRemote;
  private BitSet denseSet;

  private int taskId;
  private int parallelism;
  private int numBlocks;

  private int numIterations = 1;
  private int maxIterations;

  private boolean vectorChanged;

  /**
   * Given a job description, estimates the size of memory caches for vectors and matrix, and stores the size
   * into the job description.
   *
   * @param job The Hadoop job object
   * @throws IOException
   */
  public static void setEstimatedSize(final Job job) throws IOException {
    final Configuration conf = job.getConfiguration();
    final Path[] paths = FileInputFormat.getInputPaths(job);
    if (paths == null || paths.length == 0) {
      throw new IllegalArgumentException("Input graph is not set.");
    }

    final FileSystem fs = paths[0].getFileSystem(conf);
    long numVertices = -1;
    try (final BufferedReader reader =
             new BufferedReader(new InputStreamReader(fs.open(new Path(paths[0], Constants.StatPath))))) {
      for (final String line : reader.lines().collect(Collectors.toList())) {
        if (line.startsWith("numVertices")) {
          numVertices = Long.parseLong(line.split("=")[1]);
        }
      }
    }

    conf.setLong(Constants.SizeOfVector, numVertices);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  // Utility Methods for Coordination
  ////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Updates the status message. The user can see the message in web admin.
   *
   * @param message The status message
   */
  protected void statusMessage(final String message) {
    context.setStatus(message);
  }

  /**
   * Updates the counter value.
   *
   * @param name  The name of counter
   * @param value The value to be updated
   */
  protected void reportStats(final String name, final long value) {
    coordService.getHadoopCounter(name).update(value);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  // Utility Methods for Vector I/O
  ////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Reads a set of vertices which are in the dense region.
   *
   * @param path The path to the set of vertices
   * @throws IOException
   * @throws InterruptedException
   */
  private void readDenseSet(final Path path) throws IOException, InterruptedException {
    final RecordReader<IntWritable, NullWritable> reader = IOUtils.getIntermediateVectorReader(
        context, path, NullWritable.class);
    if (reader != null) {
      while (reader.nextKeyValue()) {
        final int key = reader.getCurrentKey().get();
        denseSet.add(key);
      }
      reader.close();
    }
  }

  /**
   * Writes the result vector to HDFS.
   *
   * @param vecBlockId The block id of result vector block
   * @param vector     The vector cache
   * @throws Exception
   */
  protected void writeOutputVecOp(final int vecBlockId, final VectorCache<V> vector) throws Exception {
    final LongWritable keyOut = new LongWritable();
    final VectorIterator<V> iter = vector.iterator(false, null);
    while (iter.next()) {
      final long key = IDUtils.localToGlobal(numBlocks, vecBlockId, iter.currentId());
      keyOut.set(key);
      context.write(keyOut, finalizeOp(iter.currentId(), iter.currentValue()));
    }
  }

  /**
   * Writes the intermediate vector block to HDFS.
   *
   * @param vector  The intermediate vector cache
   * @param key     The vector key
   * @param blockId The block id of vector cache
   * @param <T>     The type of vector value
   * @throws IOException
   * @throws InterruptedException
   */
  protected <T extends Writable> void writeIntermediateVecOp(
      final VectorCache<T> vector, final String key, final int blockId) throws IOException, InterruptedException {
    final boolean denseFlag = Constants.DenseVectorKey.equalsIgnoreCase(key);
    final RecordWriter<IntWritable, T> vecWriter = IOUtils.getIntermediateVectorWriter(context, key, blockId);
    final VectorIterator<T> iter;
    if (denseFlag) {
      iter = vector.iterator(true, denseSet);
    } else {
      iter = vector.iterator(false, null);
    }

    final IntWritable keyOut = new IntWritable();
    while (iter.next()) {
      keyOut.set(iter.currentId());
      vecWriter.write(keyOut, iter.currentValue());
    }
    vecWriter.close(context);
  }

  /**
   * Runs combine2 operation on sparse regions.
   *
   * @throws Exception
   */
  private void combine2Sparse() throws Exception {
    // Avoid to run this operation with PMV_horizontal
    if (denseThreshold == 0) return;

    LOG.info("Run combine2 for sparse blocks (iteration: {}/{})", numIterations, maxIterations);
    statusMessage(String.format("iter%d: combine2(sparse)", numIterations));
    for (int i = 0; i < parallelism; ++i) {
      final int blockId = ((taskId + i + 1) % parallelism) * parallelism + taskId;

      // apply 'combine2' operation on sparse matrix block
      combine2Op(blockId, sparseBlocks[blockId / parallelism], vecInput, vecIntermediate);

      // write the result partial vector to distributed storage
      if (i == parallelism - 1) break;
      writeIntermediateVecOp(vecIntermediate, Constants.IntermediateVectorKey, blockId);
      vecIntermediate.clear();
    }

    coordService.getSemaphoreBarrier(String.format("combine2Sparse-iter%d", numIterations)).markCompleted();
  }

  /**
   * Runs combineAll operation on the results of combine2 operation on sparse regions.
   *
   * @param pvValueClass The class type parameter for intermediate vector value
   * @throws Exception
   */
  private void combineAllSparse(final Class<PV> pvValueClass) throws Exception {
    // Avoid to run this operation with PMV_horizontal
    if (denseThreshold == 0) return;

    // wait all other workers finish combine2Sparse operation
    LOG.info("Wait combine2 for sparse blocks (iteration: {}/{})", numIterations, maxIterations);
    coordService.getSemaphoreBarrier(String.format("combine2Sparse-iter%d", numIterations)).waitBarrier();

    LOG.info("Run combineAll for sparse blocks (iteration: {}/{})", numIterations, maxIterations);
    statusMessage(String.format("iter%d: combineAll(sparse)", numIterations));
    for (int i = 0; i < parallelism - 1; ++i) {
      final int blockId = (taskId + i + 1) % parallelism + taskId * parallelism;

      // read the partial vector and apply the 'combineAll' operation
      final RecordReader<IntWritable, PV> pvReader = IOUtils.getIntermediateVectorReader(
          context, Constants.IntermediateVectorKey, blockId, pvValueClass);
      if (pvReader != null) {
        combineAllOp(blockId, pvReader, vecIntermediate);
        pvReader.close();
      }
    }
  }

  /**
   * Runs combine2 and combineAll operation on dense regions. Note that the intermediate results of
   * combine2 operation immediately reduces into the result vectors.
   *
   * @param vecValueClass The type class parameter for result vector value
   * @throws Exception
   */
  private void combineDense(final Class<V> vecValueClass) throws Exception {
    if (denseThreshold < 0) return;

    LOG.info("Run combine2 and combineAll for dense blocks (iteration: {}/{})", numIterations, maxIterations);
    statusMessage(String.format("iter%d: combine(dense)", numIterations));
    combine2Op(taskId * parallelism + taskId, denseBlocks[taskId], vecInput, vecIntermediate);
    for (int i = 0; i < parallelism - 1; ++i) {
      final int blockId = (taskId + i + 1) % parallelism;

      // read the dense vector
      vecRemote.clear();
      final RecordReader<IntWritable, V> dvReader = IOUtils.getIntermediateVectorReader(
          context, Constants.DenseVectorKey, blockId, vecValueClass);
      if (dvReader != null) {
        while (dvReader.nextKeyValue()) {
          vecRemote.put(dvReader.getCurrentKey().get(), dvReader.getCurrentValue());
        }
        dvReader.close();
      }

      // apply the 'combine2' and 'combineAll' operations to dense block in once
      combine2Op(taskId * parallelism + blockId, denseBlocks[blockId], vecRemote, vecIntermediate);
    }
  }

  /**
   * Runs the iterative multiplication.
   *
   * @param context The Hadoop context object
   * @throws IOException
   * @throws InterruptedException
   */
  @Override
  public void run(final Context context) throws IOException, InterruptedException {
    try {
      final ComputationTaskSplit task = ((ComputationTaskSplit) context.getInputSplit());
      taskId = task.taskId;

      setupFramework(context);
      setupComputation(context, coordService);

      // read vector blocks
      readDenseSet(task.dvSplit.getPath());

      // read matrix blocks
      readMatrixBlocks(context, task);

      // clear unused memory
      MemoryUtils.callGCMultipleTimes();

      boolean converged = false;
      while (!converged) {
        // initialize time counter
        final long tsIterBegin = System.nanoTime();

        // clear values for prev iteration
        vectorChanged = false;
        vecIntermediate.clear();

        // 'combine2' for sparse blocks
        combine2Sparse();
        if (numIterations == 1) MemoryUtils.callGCMultipleTimes();

        // 'combineAll' for sparse blocks
        combineAllSparse(intermediateVectorValueClass());
        if (numIterations == 1) MemoryUtils.callGCMultipleTimes();

        // combine for dense blocks
        combineDense(vectorValueClass());
        if (numIterations == 1) MemoryUtils.callGCMultipleTimes();

        // apply 'apply' operation to all blocks
        statusMessage(String.format("iter%d: apply", numIterations));
        LOG.info("Run apply for all blocks (iteration: {}/{})", numIterations, maxIterations);
        applyOp(taskId, vecInput, vecIntermediate);
        if (numIterations == 1) MemoryUtils.callGCMultipleTimes();

        ++numIterations;
        converged = (maxIterations != -1 && numIterations > maxIterations) || !vectorChanged;

        final Counter<Boolean> convergedCounter = coordService.getZkCounter(
            String.format("iterConverged-%d", numIterations - 1), Boolean.class);
        convergedCounter.update(converged);

        final Barrier iterBarrier = coordService.getSemaphoreBarrier(
            String.format("iterBarrier-%d", numIterations - 1));
        iterBarrier.markCompleted();
        iterBarrier.waitBarrier();

        converged = convergedCounter.getMinimum();

        // write dense blocks for next iteration
        if (!converged) {
          if (denseThreshold >= 0) {
            statusMessage(String.format("iter%d: write dense vector", numIterations));
            LOG.info("Write dense blocks for next iteration (iteration: {}/{})",
                numIterations - 1, maxIterations);
            writeIntermediateVecOp(vecInput, Constants.DenseVectorKey, taskId);
          }

          statusMessage(String.format("iter%d: waiting", numIterations));
          LOG.info("Wait until other workers are completed");
          final Barrier cleanupBarrier = coordService.getSemaphoreBarrier(
              String.format("cleanupBarrier-%d", numIterations - 1));
          cleanupBarrier.markCompleted();
          cleanupBarrier.waitBarrier();

          LOG.info("Cleanup iteration (iteration: {}/{})", numIterations - 1, maxIterations);
          if (numIterations == 2) MemoryUtils.callGCMultipleTimes();
        }

        // report time per iteration
        if (coordService.getWorkerId() == 0) {
          final long duration = (System.nanoTime() - tsIterBegin) / 1000000;
          reportStats(String.format("time.iter%03d", numIterations - 1), duration);
        }
      }

      cleanupComputation(context, coordService);

      // sparse and dense part
      LOG.info("Write the final vector");
      statusMessage("write sparse vector");
      writeOutputVecOp(taskId, vecInput);

      LOG.info("Cleanup MapReduce job");
      cleanupFramework(context);
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Initializes the computation framework.
   *
   * @param context The Hadoop context object
   * @throws InterruptedException
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  private void setupFramework(final Context context) throws InterruptedException, IOException {
    final Configuration conf = context.getConfiguration();
    parallelism = conf.getInt(Constants.Parallelism, 1);
    numBlocks = parallelism; // TODO: number of blocks != parallelism
    maxIterations = conf.getInt(Constants.MaxIterations, 1);

    // initialize compression codec
    final CompressionCodec codec = IOUtils.getPreferredCompressionCodec(conf);
    CodecPool.returnCompressor(CodecPool.getCompressor(codec));
    CodecPool.returnDecompressor(CodecPool.getDecompressor(codec));

    // initialize coordination service
    try {
      coordService.setup(context);
    } catch (KeeperException e) {
      // ignore
    }

    // initialize graph information
    denseThreshold = conf.getInt(Constants.DenseThreshold, Constants.DefaultDenseThreshold);
    numVertices = conf.getLong(Constants.SizeOfVector, -1L);

    // initialize matrix
    denseBlocks = new MatrixCache[parallelism];
    sparseBlocks = new MatrixCache[parallelism];

    // initialize vector
    vecInput = createVectorCache(numBlocks, taskId, numVertices);
    vecIntermediate = createIntermediateVectorCache(numBlocks, numVertices);
    vecRemote = createVectorCache(numBlocks, 0, numVertices);

    denseSet = new BitSet((int) Math.ceil(numVertices / (double) numBlocks));

    // initialize context
    this.context = context;
  }

  /**
   * Cleanups the computation framework.
   *
   * @param context The Hadoop context object
   * @throws InterruptedException
   * @throws IOException
   */
  private void cleanupFramework(final Context context) throws InterruptedException, IOException {
    if (coordService.getWorkerId() == 0) {
      coordService.getHadoopCounter(Constants.ExecutedIterations).update((long) (numIterations - 1));

      final Configuration conf = context.getConfiguration();
      final Path intermediatePath = IOUtils.getIntermediateVectorPath(
          context, Constants.IntermediateVectorKey, 0).getParent();
      final FileSystem fs = intermediatePath.getFileSystem(conf);
      fs.delete(intermediatePath, true);
    }

    coordService.cleanup(context);
  }

  /**
   * Reads the pre-partitioned matrix blocks from HDFS, and caches into local memory and disks.
   *
   * @param context The Hadoop context object
   * @param task    The task information
   * @throws IOException
   * @throws InterruptedException
   */
  private void readMatrixBlocks(final Context context, final ComputationTaskSplit task)
      throws IOException, InterruptedException {
    final Configuration conf = context.getConfiguration();
    final RecordReader<IntIntPairWritable, M> reader = new MatrixRecordReader<>(matrixValueClass(), conf);

    final List<FileSplit> blocks = new ArrayList<>();
    if (denseThreshold >= 0) {
      final List<FileSplit> denseBlocks = new ArrayList<>(task.denseSplits);
      denseBlocks.sort((left, right) -> {
        final String leftName = left.getPath().getName();
        final String rightName = right.getPath().getName();
        final int leftId = Integer.parseInt(leftName.substring(leftName.length() - 5));
        final int rightId = Integer.parseInt(rightName.substring(rightName.length() - 5));
        final int leftOrder = (leftId - parallelism * taskId + parallelism - taskId) % parallelism;
        final int rightOrder = (rightId - parallelism * taskId + parallelism - taskId) % parallelism;
        return leftOrder - rightOrder;
      });

      blocks.addAll(denseBlocks);
    }
    if (denseThreshold != 0) blocks.addAll(task.sparseSplits);

    for (FileSplit block : blocks) {
      final String blockName = block.getPath().getName();
      final int blockId = ComputationTaskSplit.blockId(block);
      statusMessage(String.format("load matrix block: %s", blockName));
      final boolean isSparse = blockName.startsWith("sparse");
      final int numCols;
      final int numRows;
      if (isSparse) {
        numCols = task.sparseCols[blockId / parallelism];
        numRows = task.sparseRows[blockId / parallelism];
      } else {
        numCols = task.denseCols[blockId % parallelism];
        numRows = task.denseRows[blockId % parallelism];
      }

      final MatrixCache<M> cache = createMatrixCache(blockName, conf, numCols, numRows);
      reader.initialize(block, context);
      while (reader.nextKeyValue()) {
        cache.put(reader.getCurrentKey(), reader.getCurrentValue());
      }
      reader.close();
      cache.freeze();

      if (isSparse) sparseBlocks[blockId / parallelism] = cache;
      else denseBlocks[blockId % parallelism] = cache;
    }
  }

  /**
   * Notify that some of vector elements are changed.
   */
  protected void notifyChanged() {
    vectorChanged = true;
  }

  /**
   * Returns the number of vertices in the graph.
   *
   * @return The number of vertices
   */
  protected long numVertices() {
    return numVertices;
  }

  /**
   * Returns the current iteration number.
   *
   * @return The current iteration number
   */
  protected int currentIteration() {
    return numIterations;
  }

  /**
   * Returns the maximum number of iterations
   *
   * @return The maximum number of iterations
   */
  protected int maxIterations() {
    return maxIterations;
  }

  /**
   * Returns the number of vector blocks.
   *
   * @return The number of vector blocks
   */
  protected int numBlocks() {
    return numBlocks;
  }

  /**
   * Returns the current vector block.
   *
   * @return The current vector block
   */
  protected VectorCache<V> currentVector() {
    return vecInput;
  }

  /**
   * Returns the Hadoop context object.
   *
   * @return The Hadoop context object
   */
  protected Context computationContext() {
    return context;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  // Matrix-Vector Multiplication Interfaces for Graph Mining
  ////////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Initializes the computation.
   *
   * @param context      The Hadoop context object
   * @param coordService The coordination service object
   */
  protected void setupComputation(final Context context, final CoordinationService coordService) {
  }

  /**
   * Finalizes the computation.
   *
   * @param context      The Hadoop context object
   * @param coordService The coordination service object
   */
  protected void cleanupComputation(final Context context, final CoordinationService coordService) {
  }

  /**
   * Creates an initial vector cache with default value.
   *
   * @param numBlocks   The number of blocks
   * @param blockId     The current block id
   * @param numVertices The number of vertices
   * @return The initialized vector cache
   */
  protected abstract VectorCache<V> createVectorCache(final int numBlocks, final int blockId, final long numVertices);

  /**
   * Creates an intermediate vector cache.
   *
   * @param numBlocks   The number of blocks
   * @param numVertices The number of vertices
   * @return The intermediate vector cache
   */
  protected abstract VectorCache<PV> createIntermediateVectorCache(final int numBlocks, final long numVertices);

  /**
   * Creates a matrix cache for storing matrix elements.
   *
   * @param blockName The name of matrix block (sparse, dense)
   * @param conf      The Hadoop configuration object
   * @param numCols   The number of columns
   * @param numRows   The number of rows
   * @return The matrix cache
   * @throws IOException
   */
  protected abstract MatrixCache<M> createMatrixCache(
      final String blockName, final Configuration conf, final int numCols, final int numRows) throws IOException;

  /**
   * Returns the class object for vector element.
   *
   * @return The class object for vector element
   */
  protected abstract Class<V> vectorValueClass();

  /**
   * Returns the class object for intermediate vector element.
   *
   * @return The class object for intermediate vector element
   */
  protected abstract Class<PV> intermediateVectorValueClass();

  /**
   * Returns the class object for matrix element
   *
   * @return The class object for matrix element
   */
  protected abstract Class<M> matrixValueClass();

  /**
   * Given a matrix block and an input vector block, computes combine2 operation on the blocks.
   *
   * @param matBlockId The id of matrix block
   * @param matrix     The matrix block
   * @param vecInput   The input vector block
   * @param vecOutput  The vector block to be stored
   * @throws Exception
   */
  protected void combine2Op(
      final int matBlockId,
      final MatrixCache<M> matrix, final VectorCache<V> vecInput, final VectorCache<PV> vecOutput)
      throws Exception {
  }

  /**
   * Given an intermediate vector block and an output vector block, computes combineAll operation on the blocks.
   *
   * @param matBlockId The matrix block id
   * @param vecInput   The intemediate vector block
   * @param vecOutput  The output vector block
   * @throws Exception
   */
  protected void combineAllOp(
      final int matBlockId,
      final RecordReader<IntWritable, PV> vecInput, final VectorCache<PV> vecOutput)
      throws Exception {
  }

  /**
   * Given a result vector block and the vector block from previous iteration, computes apply operation on the blocks.
   *
   * @param vecBlockId      The vector block id
   * @param vecResult       The vector block from previous iteration
   * @param vecIntermediate The result vector block
   * @throws Exception
   */
  protected void applyOp(
      final int vecBlockId, final VectorCache<V> vecResult, final VectorCache<PV> vecIntermediate)
      throws Exception {
  }

  /**
   * Computes the final result from output vector.
   *
   * @param id  The vertex id
   * @param vec The vector value
   * @return The final result for the vertex
   * @throws Exception
   */
  @SuppressWarnings("unchecked")
  protected OV finalizeOp(final int id, final V vec) throws Exception {
    return (OV) vec;
  }
}
