/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: CoordinationService.java
 * - Provides some coordianation features to mapper based on zookeeper service
 * Version: 3.0
 */
package pegasus.mapreduce.common.coordination;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.net.DNS;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.ZooKeeper;
import org.apache.zookeeper.server.ServerCnxnFactory;
import org.apache.zookeeper.server.ZooKeeperServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.ZookeeperUtils;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.InetSocketAddress;

public class CoordinationService implements Watcher {
  private final Logger LOG = LoggerFactory.getLogger(CoordinationService.class);

  private int workerId;
  private int parallelism;

  private String zkAddr;

  // master-related variables
  private ZooKeeperServer zkServer;

  // client-related variables
  private ZooKeeper zk;
  private boolean barrierInitialized = false;
  private boolean counterInitialized = false;

  // Hadoop-related variables
  private TaskAttemptContext context;

  /**
   * Initializes the coordination service.
   *
   * @param context The Hadoop context object
   * @throws IOException
   * @throws InterruptedException
   * @throws KeeperException
   */
  public final void setup(final TaskAttemptContext context)
      throws IOException, InterruptedException, KeeperException {
    this.context = context;
    final Configuration conf = context.getConfiguration();

    workerId = context.getTaskAttemptID().getTaskID().getId();
    parallelism = conf.getInt(Constants.Parallelism, -1);
    if (parallelism == -1) {
      parallelism = context.getNumReduceTasks();
    }

    if (workerId == 0) {
      // initialize zookeeper server
      final File zkDir = new File(
          System.getProperty("java.io.tmpdir"),
          String.format("zkData-%s", context.getTaskAttemptID().getJobID())).getAbsoluteFile();
      final InetSocketAddress addr = new InetSocketAddress(DNS.getDefaultHost("default").toLowerCase(), 0);
      final ServerCnxnFactory factory = ServerCnxnFactory.createFactory(addr, Constants.MaxNumWorkers);

      zkServer = new ZooKeeperServer(zkDir, zkDir, Constants.CoordinationTickTime);
      zkServer.setMinSessionTimeout(Constants.CoordinationTickTime * 2);
      zkServer.setMaxSessionTimeout(Constants.CoordinationTickTime * 50);
      zkAddr = String.format("%s:%d", addr.getHostName(), factory.getLocalPort());

      factory.startup(zkServer);

      // write zookeeper address to DFS
      final Path zkAddrFilePath = getZkAddrPath(context);
      final FileSystem fs = zkAddrFilePath.getFileSystem(conf);
      try (final DataOutputStream out = fs.create(zkAddrFilePath)) {
        Text.writeString(out, getServiceAddress());
      }

      connectToZkServer();
    } else {
      readZookeeperAddress(context);
      connectToZkServer();
    }
    waitInitialization();
  }

  /**
   * Connects to the Zookeeper server.
   *
   * @throws IOException
   */
  private void connectToZkServer() throws IOException {
    zk = new ZooKeeper(zkAddr, Constants.CoordinationTickTime * 5, this);
  }

  /**
   * Waits until initialization complete.
   *
   * @throws KeeperException
   * @throws InterruptedException
   */
  private synchronized void waitInitialization() throws KeeperException, InterruptedException {
    while (!barrierInitialized || !counterInitialized) {
      wait();
    }
  }

  /**
   * Cleanups the coordination service.
   *
   * @param context The Hadoop context object
   * @throws InterruptedException
   */
  public final void cleanup(final TaskAttemptContext context) throws InterruptedException {
    zk.close();
    if (workerId == 0) {
      int connections = zkServer.getNumAliveConnections();
      while (connections > 0) {
        LOG.info("Wait for connected clients: {}", connections);
        Thread.sleep(Constants.CoordinationTickTime);
        connections = zkServer.getNumAliveConnections();
      }
      zkServer.getServerCnxnFactory().shutdown();
    }
  }

  /**
   * Returns the current worker id.
   *
   * @return The current worker id
   */
  public final int getWorkerId() {
    return workerId;
  }

  /**
   * Returns the Zookeeper server address.
   *
   * @return The Zookeeper server address
   */
  public final String getServiceAddress() {
    return zkAddr;
  }

  /**
   * Returns the path to Zookeeper configuration.
   *
   * @param context The Hadoop context object
   * @return The path to Zookeeper configuration
   * @throws IOException
   */
  private Path getZkAddrPath(final TaskAttemptContext context) throws IOException {
    return new Path(Constants.DFSTemporaryPath, String.format("%s-zkServer", context.getJobID().toString()));
  }

  /**
   * Reads the Zookeeper server address from configuration.
   *
   * @param context The Hadoop context object
   * @throws IOException
   * @throws InterruptedException
   */
  private void readZookeeperAddress(final TaskAttemptContext context) throws IOException, InterruptedException {
    final Path zkPath = getZkAddrPath(context);
    final FileSystem fs = zkPath.getFileSystem(context.getConfiguration());

    while (!fs.exists(zkPath) || fs.getFileStatus(zkPath).getLen() <= 0) {
      Thread.sleep(Constants.CoordinationTickTime);
    }

    try (final DataInputStream is = fs.open(zkPath)) {
      zkAddr = Text.readString(is);
    }
    LOG.info("Zookeeper instance detected: {}", zkAddr);
  }

  /**
   * Registers the watcher for Zookeeper.
   */
  private void registerWatcher() {
    try {
      zk.exists(Constants.BarrierRoot, this);
    } catch (KeeperException | InterruptedException e) {
      e.printStackTrace();
    }
  }

  /**
   * Processes a Zookeeper event.
   *
   * @param event The Zookeeper event
   */
  @Override
  public void process(final WatchedEvent event) {
    if (event.getType() == Event.EventType.None) {
      switch (event.getState()) {
        case SyncConnected:
          if (!barrierInitialized || !counterInitialized) {
            barrierInitialized |= ZookeeperUtils.exists(zk, Constants.BarrierRoot);
            counterInitialized |= ZookeeperUtils.exists(zk, Constants.CounterRoot);
            createRootStructure();
            barrierInitialized |= ZookeeperUtils.exists(zk, Constants.BarrierRoot);
            counterInitialized |= ZookeeperUtils.exists(zk, Constants.CounterRoot);
            registerWatcher();
          }
          notifyFromWatcher();
          break;
        case Disconnected:
        case Expired:
          try {
            connectToZkServer();
            registerWatcher();
          } catch (IOException e) {
            e.printStackTrace(); // FIXME!
          }
          break;
        default:
          break;
      }
    }
  }

  /**
   * Initializes directory structure in Zookeeper.
   */
  private void createRootStructure() {
    // barrier
    if (!barrierInitialized) {
      ZookeeperUtils.createPersistentNode(zk, Constants.BarrierRoot);
    }
    // counter
    if (!counterInitialized) {
      ZookeeperUtils.createPersistentNode(zk, Constants.CounterRoot);
    }
  }

  /**
   * Creates a semaphore barrier.
   *
   * @param name The name of barrier
   * @return The semaphore barrier
   */
  public Barrier getSemaphoreBarrier(final String name) {
    final SemaphoreBarrier barrier = new SemaphoreBarrier(zk, name, workerId, parallelism, workerId == 0);
    barrier.setup();
    return barrier;
  }

  /**
   * Creates a mutex barrier.
   *
   * @param name    The name of barrier
   * @param mutexId The mutex id
   * @param isMine  Whether the current worker owns the barrier or not
   * @return The mutex barrier
   */
  public Barrier getMutexBarrier(final String name, final int mutexId, final boolean isMine) {
    final MutexBarrier barrier = new MutexBarrier(zk, name, mutexId, isMine);
    barrier.setup();
    return barrier;
  }

  /**
   * Creates a Zookeeper-based counter.
   *
   * @param name  The name of counter
   * @param clazz The type of counter value
   * @param <T>   The type parameter of counter
   * @return The counter object
   */
  @SuppressWarnings("unchecked")
  public <T extends Comparable<T>> Counter<T> getZkCounter(final String name, final Class<T> clazz) {
    if (clazz == Long.class) {
      return (Counter<T>) new ZookeeperLongCounter(zk, name, workerId);
    } else if (clazz == Double.class) {
      return (Counter<T>) new ZookeeperDoubleCounter(zk, name, workerId);
    } else if (clazz == Integer.class) {
      return (Counter<T>) new ZookeeperIntegerCounter(zk, name, workerId);
    } else if (clazz == Boolean.class) {
      return (Counter<T>) new ZookeeperBooleanCounter(zk, name, workerId);
    }
    return null;
  }

  /**
   * Creates a Hadoop-based counter.
   *
   * @param name The name of counter
   * @return The counter object
   */
  public Counter<Long> getHadoopCounter(final String name) {
    return new HadoopCounter(context, name);
  }

  /**
   * Notify all slept threads.
   */
  private synchronized void notifyFromWatcher() {
    notifyAll();
  }
}

