/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: MutexBarrier.java
 * - Implementation of mutex barrier based on zookeeper
 * Version: 3.0
 */
package pegasus.mapreduce.common.coordination;

import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.ZooKeeper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.ZookeeperUtils;

public class MutexBarrier implements Watcher, Barrier {
  private static final Logger LOG = LoggerFactory.getLogger(MutexBarrier.class);
  private final ZooKeeper zk;
  private final String barrierPath;
  private boolean isMine;
  private boolean finished;

  public MutexBarrier(
      final ZooKeeper zk, final String name, final int mutexId, final boolean isMine) {
    this.zk = zk;
    this.isMine = isMine;
    this.barrierPath = String.format("%s/%s-%d", Constants.BarrierRoot, name, mutexId);
  }

  @Override
  public final void setup() {
    finished = ZookeeperUtils.exists(zk, barrierPath);
    registerWatcher();
  }

  @Override
  public final void markCompleted() {
    finished = true;
    ZookeeperUtils.createPersistentNode(zk, barrierPath);
    LOG.info("Mark barrier {}", barrierPath);
  }

  @Override
  public synchronized final void waitBarrier() {
    LOG.info("Wait barrier {}", barrierPath);
    while (!finished) {
      try {
        wait(Constants.CoordinationTickTime);
        finished = ZookeeperUtils.exists(zk, barrierPath);
      } catch (InterruptedException e) {
        // ignore
      }
      registerWatcher();
    }

    finished = false;
  }

  @Override
  public void cleanup() {
    if (isMine) {
      finished = false;
      ZookeeperUtils.delete(zk, barrierPath);
      LOG.info("Barrier cleanup: {}", barrierPath);
    }
  }

  @Override
  public void process(WatchedEvent event) {
    if (event.getType() == Event.EventType.NodeCreated) {
      finished = ZookeeperUtils.exists(zk, barrierPath);
      notifyFromWatcher();
    }

    registerWatcher();
  }

  private synchronized void notifyFromWatcher() {
    notifyAll();
  }

  private void registerWatcher() {
    try {
      zk.exists(barrierPath, this);
    } catch (KeeperException | InterruptedException e) {
      // ignore
    }
  }

  @Override
  public String toString() {
    return String.format("MutexBarrier(barrierPath='%s')", barrierPath);
  }
}
