/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: SemaphoreBarrier.java
 * - Implementation of semaphore barrier
 * Version: 3.0
 */
package pegasus.mapreduce.common.coordination;

import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.ZooKeeper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.ZookeeperUtils;

import java.util.List;

public class SemaphoreBarrier implements Barrier, Watcher {
  private static final Logger LOG = LoggerFactory.getLogger(SemaphoreBarrier.class);
  private final int workerId;
  private final ZooKeeper zk;
  private final String barrierRoot;
  private boolean initialized;
  private boolean isMine;
  private boolean[] finished;

  public SemaphoreBarrier(
      final ZooKeeper zk, final String name, final int workerId, final int parallelism, final boolean isMine) {
    this.zk = zk;
    this.workerId = workerId;
    this.barrierRoot = String.format("%s/%s", Constants.BarrierRoot, name);
    this.isMine = isMine;
    finished = new boolean[parallelism];
  }

  private void registerWatcher() {
    try {
      updateState();
      zk.getChildren(barrierRoot, this);
    } catch (KeeperException | InterruptedException e) {
      // ignore
    }
  }

  private void updateState() {
    if (initialized) {
      final List<String> children = ZookeeperUtils.getChildren(zk, barrierRoot);
      for (String child : children) {
        finished[Integer.parseInt(child)] = true;
      }
    }
  }

  @Override
  public void setup() {
    // initialize barrier structure
    initialized = ZookeeperUtils.exists(zk, barrierRoot);
    if (!initialized) {
      ZookeeperUtils.createPersistentNode(zk, barrierRoot);
      initialized = ZookeeperUtils.exists(zk, barrierRoot);
    }

    updateState();
    registerWatcher();
  }

  @Override
  public void markCompleted() {
    ZookeeperUtils.createPersistentNode(zk, String.format("%s/%d", barrierRoot, workerId));
    registerWatcher();
  }

  @Override
  public synchronized void waitBarrier() {
    LOG.info("Wait barrier {}", barrierRoot);
    boolean completed = false;
    while (!completed) {
      completed = true;
      for (boolean fin : finished) {
        if (!fin) {
          completed = false;
          break;
        }
      }

      if (!completed) {
        try {
          wait(Constants.CoordinationTickTime);
        } catch (InterruptedException e) {
          // ignore
        }
        registerWatcher();
      }
    }

    initialized = false;
    for (int i = 0; i < finished.length; ++i) {
      finished[i] = false;
    }
  }

  @Override
  public void cleanup() {
    if (isMine) {
      for (int i = 0; i < finished.length; ++i) {
        ZookeeperUtils.delete(zk, String.format("%s/%d", barrierRoot, i));
      }
    }
  }

  private synchronized void notifyFromWatcher() {
    notifyAll();
  }

  @Override
  public void process(final WatchedEvent event) {
    if (event.getType() == Event.EventType.NodeChildrenChanged) {
      updateState();
      notifyFromWatcher();
    }

    registerWatcher();
  }

  @Override
  public String toString() {
    return String.format("SemaphoreBarrier(path = %s/%d)", barrierRoot, workerId);
  }
}
