/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: ZookeeperCounter.java
 * - Implementation of counter based on zookeeper
 * Version: 3.0
 */
package pegasus.mapreduce.common.coordination;

import org.apache.zookeeper.ZooKeeper;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.ZookeeperUtils;

import java.util.Comparator;
import java.util.function.BinaryOperator;
import java.util.stream.Stream;

public abstract class ZookeeperCounter<T extends Comparable<T>> implements Counter<T> {
  private final ZooKeeper zk;
  private final String counterPath;
  private final String myCounterPath;

  public ZookeeperCounter(final ZooKeeper zk, final String name, final int workerId) {
    this.zk = zk;
    counterPath = String.format("%s/%s", Constants.CounterRoot, name);
    myCounterPath = String.format("%s/%d", counterPath, workerId);

    if (!ZookeeperUtils.exists(zk, counterPath)) {
      ZookeeperUtils.createPersistentNode(zk, counterPath);
    }
  }

  @Override
  public void update(final T value) {
    ZookeeperUtils.createOrUpdatePersistentNode(zk, myCounterPath, valueToString(value));
  }

  @Override
  public T getMinimum() {
    return getCounters().min(Comparator.naturalOrder()).orElse(positiveInfinite());
  }

  @Override
  public T getMaximum() {
    return getCounters().max(Comparator.naturalOrder()).orElse(negativeInfinite());
  }

  @Override
  public T getSum() {
    return getCounters().reduce(zero(), sumOp());
  }

  protected abstract T negativeInfinite();

  protected abstract T positiveInfinite();

  protected abstract T zero();

  protected abstract BinaryOperator<T> sumOp();

  protected abstract T valueFromString(final String string);

  protected abstract String valueToString(final T value);

  private Stream<T> getCounters() {
    return ZookeeperUtils.getChildren(zk, counterPath).parallelStream().map(
        path -> valueFromString(ZookeeperUtils.getNodeData(zk, String.format("%s/%s", counterPath, path))));
  }
}
