/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BaseMatrixOutputFormat.java
 * - A base class for matrix output format
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import com.esotericsoftware.kryo.io.Output;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.utils.IOUtils;

import java.io.IOException;

public abstract class BaseMatrixOutputFormat<V> extends FileOutputFormat<IntWritable, V> {
  protected static class OutputConfig {
    public final Output[] sparseOutput;
    public final Output[] denseOutput;
    public final int denseThreshold;

    public OutputConfig(final Output[] sparseOutput, final Output[] denseOutput, final int denseThreshold) {
      this.sparseOutput = sparseOutput;
      this.denseOutput = denseOutput;
      this.denseThreshold = denseThreshold;
    }
  }

  protected OutputConfig getOutputConfig(final TaskAttemptContext context) throws IOException {
    final Configuration conf = context.getConfiguration();
    final int workerId = context.getTaskAttemptID().getTaskID().getId();
    final int parallelism = context.getNumReduceTasks();
    final int denseThreshold = conf.getInt(Constants.DenseThreshold, Constants.DefaultDenseThreshold);

    final Path rootPath = getOutputPath(context);
    final FileSystem fs = rootPath.getFileSystem(conf);
    final CompressionCodec codec = IOUtils.getPreferredCompressionCodec(context);
    final int bufSize = conf.getInt("io.file.buffer.size", 4096);

    final Output[] sparseOutput = new Output[parallelism];
    final Output[] denseOutput = new Output[parallelism];

    for (int i = 0; i < parallelism; ++i) {
      final int blockId = workerId + i * parallelism;
      final Path sparsePath = new Path(rootPath, String.format("sparse-%05d", blockId));
      final Path densePath = new Path(rootPath, String.format("dense-%05d", blockId));

      sparseOutput[i] = new Output(codec.createOutputStream(
          fs.create(sparsePath, true, bufSize, (short) 1, fs.getDefaultBlockSize(sparsePath), context)));
      denseOutput[i] = new Output(codec.createOutputStream(
          fs.create(densePath, true, bufSize, (short) 1, fs.getDefaultBlockSize(densePath), context)));
    }

    return new OutputConfig(sparseOutput, denseOutput, denseThreshold);
  }
}
