/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: ComputationTaskSplit.java
 * - A subproblem of matrix-vector multiplication
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.*;

public class ComputationTaskSplit extends InputSplit implements Writable {
  public int taskId;
  public List<FileSplit> sparseSplits = new ArrayList<>();
  public int[] sparseRows;
  public int[] sparseCols;
  public List<FileSplit> denseSplits = new ArrayList<>();
  public int[] denseRows;
  public int[] denseCols;
  public FileSplit svSplit;
  public FileSplit dvSplit;

  @Override
  public long getLength() throws IOException, InterruptedException {
    int length = 0;
    for (final FileSplit child : sparseSplits) {
      length += child.getLength();
    }
    for (final FileSplit child : denseSplits) {
      length += child.getLength();
    }
    length += svSplit.getLength() + dvSplit.getLength();
    return length;
  }

  @Override
  public String[] getLocations() throws IOException, InterruptedException {
    final Set<String> locations = new HashSet<>();

    for (final FileSplit child : sparseSplits) {
      locations.addAll(Arrays.asList(child.getLocations()));
    }

    return locations.toArray(new String[0]);
  }

  @Override
  public void write(final DataOutput out) throws IOException {
    out.writeInt(taskId);
    out.writeInt(sparseSplits.size());
    for (int i = 0, size = sparseSplits.size(); i < size; ++i) {
      sparseSplits.get(i).write(out);
      out.writeInt(sparseCols[i]);
      out.writeInt(sparseRows[i]);
    }

    out.writeInt(denseSplits.size());
    for (int i = 0, size = denseSplits.size(); i < size; ++i) {
      denseSplits.get(i).write(out);
      out.writeInt(denseRows[i]);
      out.writeInt(denseCols[i]);
    }

    svSplit.write(out);
    dvSplit.write(out);
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    taskId = in.readInt();
    final int sparseLength = in.readInt();
    sparseSplits = new ArrayList<>(sparseLength);
    sparseCols = new int[sparseLength];
    sparseRows = new int[sparseLength];
    for (int i = 0; i < sparseLength; ++i) {
      sparseSplits.add(new FileSplit());
      sparseSplits.get(i).readFields(in);
      sparseCols[i] = in.readInt();
      sparseRows[i] = in.readInt();
    }

    final int denseLength = in.readInt();
    denseSplits = new ArrayList<>(denseLength);
    denseCols = new int[denseLength];
    denseRows = new int[denseLength];
    for (int i = 0; i < denseLength; ++i) {
      denseSplits.add(new FileSplit());
      denseSplits.get(i).readFields(in);
      denseCols[i] = in.readInt();
      denseRows[i] = in.readInt();
    }

    svSplit = new FileSplit();
    svSplit.readFields(in);
    dvSplit = new FileSplit();
    dvSplit.readFields(in);
  }

  public static int blockId(FileSplit split) {
    return Integer.parseInt(split.getPath().getName().split("-")[1]);
  }
}
