/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: Counters.java
 * - A binary file reader for matrix record
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.KryoDataInput;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.compress.CodecPool;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.Decompressor;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;
import org.apache.hadoop.util.ReflectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.types.IntIntPairWritable;
import pegasus.mapreduce.common.utils.IOUtils;

import java.io.IOException;

public class MatrixRecordReader<V extends Writable> extends RecordReader<IntIntPairWritable, V> {
  private static final Logger LOG = LoggerFactory.getLogger(MatrixRecordReader.class);
  private Input in;
  private KryoDataInput dataIn;
  private Decompressor decompressor;

  private final IntIntPairWritable key = new IntIntPairWritable();
  private final V value;

  public MatrixRecordReader(final Class<V> valueClass, final Configuration conf) {
    this.value = ReflectionUtils.newInstance(valueClass, conf);
  }

  @Override
  public void initialize(final InputSplit split, final TaskAttemptContext context)
      throws IOException, InterruptedException {
    final Configuration conf = context.getConfiguration();
    final Path path = ((FileSplit) split).getPath();
    final FileSystem fs = path.getFileSystem(conf);
    final CompressionCodec codec = IOUtils.getPreferredCompressionCodec(conf);
    // We manually retrieve decompressor from codec pool because Hadoop 2.7 has a memory leakage bug.
    decompressor = CodecPool.getDecompressor(codec);

    LOG.info("Initialize input stream: {}", path.toString());
    in = new Input(codec.createInputStream(fs.open(path), decompressor));
    dataIn = new KryoDataInput(in);
  }

  @Override
  public boolean nextKeyValue() throws IOException, InterruptedException {
    if (in.eof()) return false;

    key.readFields(dataIn);
    value.readFields(dataIn);
    return true;
  }

  @Override
  public IntIntPairWritable getCurrentKey() throws IOException, InterruptedException {
    return key;
  }

  @Override
  public V getCurrentValue() throws IOException, InterruptedException {
    return value;
  }

  @Override
  public void close() throws IOException {
    if (in != null) {
      in.close();

      // We manually return the decompressor
      CodecPool.returnDecompressor(decompressor);
      decompressor = null;
      dataIn = null;
      in = null;
    }
  }

  @Override
  public float getProgress() throws IOException, InterruptedException {
    return 0.99f;
  }
}
