/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: UnweightedAdjacencyListInputFormat.java
 * - An input format for adjancency file formatted graph without weights
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.LineRecordReader;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.types.LongArrayWritable;

import java.io.IOException;

public class UnweightedAdjacencyListInputFormat extends FileInputFormat<LongWritable, LongArrayWritable> {
  @Override
  public RecordReader<LongWritable, LongArrayWritable> createRecordReader(
      InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
    final Configuration conf = context.getConfiguration();

    return new AdjacencyListRecordReader(
        conf.get(Constants.AdjacencyListVertexDelimiter, "\t"),
        conf.get(Constants.AdjacencyListEdgeDelimiter, " "));
  }

  public static class AdjacencyListRecordReader extends RecordReader<LongWritable, LongArrayWritable> {
    private final String vertexDelimiter;
    private final String edgeDelimiter;
    private final LineRecordReader reader = new LineRecordReader();

    private final LongWritable vertex = new LongWritable();
    private final LongArrayWritable edges = new LongArrayWritable();

    public AdjacencyListRecordReader(String vertexDelimiter, String edgeDelimiter) {
      this.vertexDelimiter = vertexDelimiter;
      this.edgeDelimiter = edgeDelimiter;
    }

    @Override
    public void initialize(InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
      reader.initialize(split, context);
    }

    @Override
    public boolean nextKeyValue() throws IOException, InterruptedException {
      if (!reader.nextKeyValue()) return false;

      final Text line = reader.getCurrentValue();
      final String lineStr = line.toString();
      final int vertexLimit = line.find(vertexDelimiter);

      // vertex id
      vertex.set(Long.parseLong(lineStr.substring(0, vertexLimit)));

      // number of neighbors
      final int lengthLimit = line.find(edgeDelimiter, vertexLimit + 1);
      int length;
      if (lengthLimit == -1) length = 0;
      else length = Integer.parseInt(lineStr.substring(vertexLimit + 1, lengthLimit));

      switch (length) {
        case 0:
          return nextKeyValue();
        case 1:
          edges.set(Long.parseLong(lineStr.substring(lengthLimit + 1)));
          break;
        default:
          long[] values;
          if (edges.getValues().length < length) {
            values = new long[length];
          } else {
            values = edges.getValues();
          }

          int begin = lengthLimit + 1, end = line.find(edgeDelimiter, begin);
          for (int i = 0; end != -1; ++i) {
            values[i] = Long.parseLong(lineStr.substring(begin, end));
            begin = end + 1;
            end = line.find(edgeDelimiter, begin);
          }
          values[length - 1] = Long.parseLong(lineStr.substring(begin));
          edges.set(values, 0, length);
          break;
      }

      return true;
    }

    @Override
    public LongWritable getCurrentKey() throws IOException, InterruptedException {
      return vertex;
    }

    @Override
    public LongArrayWritable getCurrentValue() throws IOException, InterruptedException {
      return edges;
    }

    @Override
    public float getProgress() throws IOException, InterruptedException {
      return reader.getProgress();
    }

    @Override
    public void close() throws IOException {
      reader.close();
    }
  }
}
