/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: WeightedEdgeListInputFormat.java
 * - An input format implementation for edge list formatted graph with weights
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.LineRecordReader;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.types.LongDoublePairArrayWritable;

import java.io.IOException;

public class WeightedEdgeListInputFormat extends FileInputFormat<LongWritable, LongDoublePairArrayWritable> {
  @Override
  public RecordReader<LongWritable, LongDoublePairArrayWritable> createRecordReader(
      InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
    final String delimiter = context.getConfiguration().get(Constants.EdgeListDelimiter, "\t");
    return new EdgeRecordReader(delimiter);
  }

  public static class EdgeRecordReader extends RecordReader<LongWritable, LongDoublePairArrayWritable> {
    private final LongWritable vertex = new LongWritable();
    private final LongDoublePairArrayWritable edges = new LongDoublePairArrayWritable(new long[1], new double[1]);
    private final LineRecordReader reader = new LineRecordReader();
    private final String fieldDelimiter;

    public EdgeRecordReader(String fieldDelimiter) {
      this.fieldDelimiter = fieldDelimiter;
    }

    @Override
    public void initialize(InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
      reader.initialize(split, context);
    }

    @Override
    public boolean nextKeyValue() throws IOException, InterruptedException {
      if (!reader.nextKeyValue()) return false;

      final Text line = reader.getCurrentValue();
      final String lineStr = line.toString();
      final int split1 = line.find(fieldDelimiter);
      final int split2 = line.find(fieldDelimiter, split1 + 1);

      vertex.set(Long.parseLong(lineStr.substring(0, split1)));
      edges.getLongs()[0] = Long.parseLong(lineStr.substring(split1 + 1, split2));
      edges.getDoubles()[0] = Double.parseDouble(lineStr.substring(split2 + 1));

      return true;
    }

    @Override
    public LongWritable getCurrentKey() throws IOException, InterruptedException {
      return vertex;
    }

    @Override
    public LongDoublePairArrayWritable getCurrentValue() throws IOException, InterruptedException {
      return edges;
    }

    @Override
    public float getProgress() throws IOException, InterruptedException {
      return reader.getProgress();
    }

    @Override
    public void close() throws IOException {
      reader.close();
    }
  }
}
