/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: WeightedMatrixOutputFormat.java
 * - An output format implementation for weighted matrix
 * Version: 3.0
 */
package pegasus.mapreduce.common.io;

import com.esotericsoftware.kryo.io.Output;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.mapreduce.RecordWriter;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import pegasus.mapreduce.common.types.WeightedColumnEntry;

import java.io.IOException;

public class WeightedMatrixOutputFormat extends BaseMatrixOutputFormat<WeightedColumnEntry> {
  @Override
  public RecordWriter<IntWritable, WeightedColumnEntry> getRecordWriter(final TaskAttemptContext context)
      throws IOException, InterruptedException {
    final OutputConfig config = getOutputConfig(context);
    return new MultiBlockWriter(
        config.sparseOutput, config.denseOutput, config.denseThreshold, config.denseOutput.length);
  }

  public static class MultiBlockWriter extends RecordWriter<IntWritable, WeightedColumnEntry> {
    private final int denseThreshold;
    private final int numBlocks;
    private final Output[] sparseOutput;
    private final Output[] denseOutput;

    MultiBlockWriter(
        final Output[] sparseOutput,
        final Output[] denseOutput,
        final int denseThreshold,
        final int numBlocks) {
      this.sparseOutput = sparseOutput;
      this.denseOutput = denseOutput;
      this.denseThreshold = denseThreshold;
      this.numBlocks = numBlocks;
    }

    @Override
    public void write(final IntWritable blockId, final WeightedColumnEntry colEntry)
        throws IOException, InterruptedException {
      final int subBlockId = blockId.get();
      final int degree = colEntry.numEntries;
      Output out;
      if (denseThreshold < 0 || degree < denseThreshold) {
        out = sparseOutput[subBlockId];
      } else {
        out = denseOutput[subBlockId];
      }

      // write column
      out.writeInt(colEntry.col);
      out.writeInt(degree);

      // write rows
      final int size = colEntry.rows.size();
      if (size == 1) {
        out.writeInt(colEntry.rows.getInt(0));
        out.writeDouble(colEntry.weights.getDouble(0));
      } else {
        out.writeInt(colEntry.rows.getInt(0) | Integer.MIN_VALUE);
        out.writeDouble(colEntry.weights.getDouble(0));
        out.writeInt(size - 1);
        for (int i = 1; i < size; ++i) {
          out.writeInt(colEntry.rows.getInt(i));
          out.writeDouble(colEntry.weights.getDouble(i));
        }
      }
    }

    @Override
    public void close(final TaskAttemptContext context) throws IOException, InterruptedException {
      for (int i = 0; i < numBlocks; ++i) {
        sparseOutput[i].close();
        denseOutput[i].close();
      }
    }
  }
}
