/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: BitStrings.java
 * - Represents multiple bitstring for diameter estimation task.
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Arrays;

/**
 * Represents multiple (typically, 32) bitstrings.
 */
public class BitStrings implements Writable {
  private long[] bits = new long[Constants.DefaultNumBitstrings];

  /**
   * Construct bitstrings with empty values.
   */
  public BitStrings() {
  }

  /**
   * Construct bitstring with given bit values.
   *
   * @param bits The initial bit values
   */
  public BitStrings(final long[] bits) {
    setBits(bits);
  }

  /**
   * Sets the value of bitstring to the given bit values.
   *
   * @param bits The bit values to be set
   */
  public void setBits(final long[] bits) {
    System.arraycopy(bits, 0, this.bits, 0, Constants.DefaultNumBitstrings);
  }

  /**
   * Returns current bit values.
   *
   * @return The current bit values which are stored
   */
  public long[] getBits() {
    return bits;
  }

  /**
   * Returns i-th sub-bitstring.
   *
   * @param i The index of desired bitstring
   * @return The i-th sub-bitstring
   */
  public long getBit(final int i) {
    return bits[i];
  }

  /**
   * Sets i-th sub-bitstring to the given bit values.
   *
   * @param i   The index of bitstring to be updated
   * @param bit The bit values
   */
  public void setBit(final int i, final long bit) {
    bits[i] = bit;
  }

  /**
   * Applies bitwise-or operation on itself and the given bitstrings and assigns the result of operation.
   *
   * @param other The bitstrings for bitwise-or operation
   */
  public void or(final BitStrings other) {
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      bits[i] |= other.bits[i];
    }
  }

  /**
   * Returns whether all bitstrings are empty or not.
   *
   * @return true if all bitstrings are empty, otherwise false
   */
  public boolean isEmpty() {
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      if (bits[i] != 0) return false;
    }
    return true;
  }

  /**
   * Writes bitstring values to an output stream.
   *
   * @param out The output stream object to write the values
   * @throws IOException
   */
  @Override
  public void write(final DataOutput out) throws IOException {
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      out.writeLong(bits[i]);
    }
  }

  /**
   * Read bitstring values from an input stream.
   *
   * @param in The input stream object to read the values
   * @throws IOException
   */
  @Override
  public void readFields(final DataInput in) throws IOException {
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      bits[i] = in.readLong();
    }
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;

    BitStrings bitStrings = (BitStrings) o;

    return Arrays.equals(bits, bitStrings.bits);
  }

  @Override
  public int hashCode() {
    return Arrays.hashCode(bits);
  }

  /**
   * Returns the cardinality of bitstrings using Flajolet-Martin method.
   *
   * @return The cardinality of bitstrings
   * @see <a href="https://en.wikipedia.org/wiki/Flajolet%E2%80%93Martin_algorithm">Wikipedia</a>
   * @see <a href="http://algo.inria.fr/flajolet/Publications/FlMa85.pdf">Probabitistic Counting Algorithms for Data Base Applications</a>
   */
  public double cardinality() {
    double bitpos = 0;
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      final long bits = this.bits[i];
      boolean flag = false;
      for (int j = 0; j < 64; ++j) {
        if ((bits & (1L << (63 - j))) == 0) {
          bitpos += j;
          flag = true;
          break;
        }
      }

      if (!flag) bitpos += 64;
    }

    bitpos /= Constants.DefaultNumBitstrings;

    return Math.pow(2, bitpos) / Constants.PHI;
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
      builder.append(Long.toHexString(bits[i]));
    }
    return builder.toString();
  }
}