/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: IntArrayWritable.java
 * - Represents an array contains int values
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import it.unimi.dsi.fastutil.ints.IntArrayList;
import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Represents an array contains int values.
 */
public class IntArrayWritable implements Writable {
  private int[] values;
  private int begin;
  private int end;

  public IntArrayWritable() {
    begin = end = 0;
    values = new int[Constants.InitArrayWritableSize];
  }

  public IntArrayWritable(final int value) {
    begin = 0;
    end = 1;
    values = new int[Constants.InitArrayWritableSize];
    values[0] = value;
  }

  public IntArrayWritable(final int... values) {
    begin = 0;
    end = values.length;
    this.values = values;
  }

  public final int size() {
    return end - begin;
  }

  public final int[] getValues() {
    return values;
  }

  public final int begin() {
    return begin;
  }

  public final int end() {
    return end;
  }

  public final int get(final int i) {
    return values[begin + i];
  }

  public final void set(final int value) {
    this.values[0] = value;
    begin = 0;
    end = 1;
  }

  public final void set(final int[] values) {
    this.values = values;
    begin = 0;
    end = values.length;
  }

  public final void set(final int[] values, final int begin, final int end) {
    this.values = values;
    this.begin = begin;
    this.end = end;
  }

  public final void set(final IntArrayList values) {
    this.values = values.elements();
    this.begin = 0;
    this.end = values.size();
  }

  @Override
  public void write(DataOutput out) throws IOException {
    final int size = size();
    if (size > 1) {
      out.writeInt(values[begin] | Integer.MIN_VALUE);
      out.writeInt(size - 1);
      for (int i = begin + 1; i < end; ++i) {
        out.writeInt(values[i]);
      }
    } else {
      out.writeInt(values[begin]);
    }
  }

  @Override
  public void readFields(final DataInput in) throws IOException {
    final int head = in.readInt();
    if ((head & Integer.MIN_VALUE) != 0) {
      end = in.readInt() + 1;
      if (end > values.length) {
        values = new int[end];
      }
      begin = 0;
      values[0] = head & Integer.MAX_VALUE;
      for (int i = 1; i < end; ++i) {
        values[i] = in.readInt();
      }
    } else {
      values[0] = head;
      begin = 0;
      end = 1;
    }
  }
}
