/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: IntDoublePairArrayWritable.java
 * - Represents an array contains (int, double) pair values
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Represents an array contains (int, double) pair values.
 */
public class IntDoublePairArrayWritable implements Writable {
  private int[] ints;
  private double[] doubles;

  private int begin;
  private int end;

  public IntDoublePairArrayWritable() {
    begin = end = 0;
    ints = new int[Constants.InitArrayWritableSize];
    doubles = new double[Constants.InitArrayWritableSize];
  }

  public IntDoublePairArrayWritable(final int[] ints, final double[] doubles) {
    this.ints = ints;
    this.doubles = doubles;
    begin = 0;
    end = ints.length;
  }

  public final int size() {
    return end - begin;
  }

  public final int[] getInts() {
    return ints;
  }

  public final double[] getDoubles() {
    return doubles;
  }

  public final int begin() {
    return begin;
  }

  public final int end() {
    return end;
  }

  public final int getInt(final int i) {
    return ints[begin + i];
  }

  public final double getDouble(final int i) {
    return doubles[begin + i];
  }

  public final void set(final int intValue, final double doubleValue) {
    ints[0] = intValue;
    doubles[0] = doubleValue;
    begin = 0;
    end = 1;
  }

  public final void set(final int[] ints, final double[] doubles, final int begin, final int end) {
    this.ints = ints;
    this.doubles = doubles;
    this.begin = begin;
    this.end = end;
  }

  public final void set(final IntArrayList ints, final DoubleArrayList doubles) {
    this.ints = ints.elements();
    this.doubles = doubles.elements();
    this.begin = 0;
    this.end = ints.size();
  }

  @Override
  public void write(DataOutput out) throws IOException {
    final int size = size();
    if (size > 1) {
      out.writeInt(ints[begin] | Integer.MIN_VALUE);
      out.writeDouble(doubles[begin]);
      out.writeInt(size - 1);
      for (int i = begin + 1; i < end; ++i) {
        out.writeInt(ints[i]);
        out.writeDouble(doubles[i]);
      }
    } else {
      out.writeInt(ints[begin]);
      out.writeDouble(doubles[begin]);
    }
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    final int headInt = in.readInt();
    final double headDouble = in.readDouble();
    if ((headInt & Integer.MIN_VALUE) != 0) {
      end = in.readInt() + 1;
      if (end > ints.length) {
        ints = new int[end];
        doubles = new double[end];
      }
      begin = 0;
      ints[0] = headInt & Integer.MAX_VALUE;
      doubles[0] = headDouble;
      for (int i = 1; i < end; ++i) {
        ints[i] = in.readInt();
        doubles[i] = in.readDouble();
      }
    } else {
      ints[0] = headInt;
      doubles[0] = headDouble;
      begin = 0;
      end = 1;
    }
  }
}
