/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: LongDoublePairArrayWritable.java
 * - Represents an array contains (long, double) pair values
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.longs.LongArrayList;
import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Represents an array contains (long, double) pair values.
 */
public class LongDoublePairArrayWritable implements Writable {
  private long[] longs;
  private double[] doubles;

  private int begin;
  private int end;

  public LongDoublePairArrayWritable() {
    begin = end = 0;
    longs = new long[Constants.InitArrayWritableSize];
    doubles = new double[Constants.InitArrayWritableSize];
  }

  public LongDoublePairArrayWritable(final long[] longs, final double[] doubles) {
    this.longs = longs;
    this.doubles = doubles;
    begin = 0;
    end = longs.length;
  }

  public final int size() {
    return end - begin;
  }

  public final long[] getLongs() {
    return longs;
  }

  public final double[] getDoubles() {
    return doubles;
  }

  public final int begin() {
    return begin;
  }

  public final int end() {
    return end;
  }

  public final long getLong(final int i) {
    return longs[begin + i];
  }

  public final double getDouble(final int i) {
    return doubles[begin + i];
  }

  public final void set(final long longValue, final double doubleValue) {
    longs[0] = longValue;
    doubles[0] = doubleValue;
    begin = 0;
    end = 1;
  }

  public final void set(final long[] longs, final double[] doubles, final int begin, final int end) {
    this.longs = longs;
    this.doubles = doubles;
    this.begin = begin;
    this.end = end;
  }

  public final void set(final LongArrayList longs, final DoubleArrayList doubles) {
    this.longs = longs.elements();
    this.doubles = doubles.elements();
    this.begin = 0;
    this.end = longs.size();
  }

  @Override
  public void write(DataOutput out) throws IOException {
    final int size = size();
    if (size > 1) {
      out.writeLong(longs[begin] | Long.MIN_VALUE);
      out.writeDouble(doubles[begin]);
      out.writeInt(size - 1);
      for (int i = begin + 1; i < end; ++i) {
        out.writeLong(longs[i]);
        out.writeDouble(doubles[i]);
      }
    } else {
      out.writeLong(longs[begin]);
      out.writeDouble(doubles[begin]);
    }
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    final long headLong = in.readLong();
    final double headDouble = in.readDouble();
    if ((headLong & Long.MIN_VALUE) != 0) {
      end = in.readInt() + 1;
      if (end > longs.length) {
        longs = new long[end];
        doubles = new double[end];
      }
      begin = 0;
      longs[0] = headLong & Long.MAX_VALUE;
      doubles[0] = headDouble;
      for (int i = 1; i < end; ++i) {
        longs[i] = in.readLong();
        doubles[i] = in.readDouble();
      }
    } else {
      longs[0] = headLong;
      doubles[0] = headDouble;
      begin = 0;
      end = 1;
    }
  }
}