/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: LongIntPairMapWritable.java
 * - Represents a map contains (long, int) pair values
 * Version: 3.0
 */
package pegasus.mapreduce.common.types;

import it.unimi.dsi.fastutil.longs.Long2IntMap;
import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap;
import org.apache.hadoop.io.Writable;
import pegasus.mapreduce.common.Constants;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Iterator;

/**
 * Represents a map contains (long, int) pair values.
 */
public class LongIntPairMapWritable implements Writable {
  private Long2IntOpenHashMap map;

  public LongIntPairMapWritable() {
    this(Constants.InitArrayWritableSize);
  }

  public LongIntPairMapWritable(final int size) {
    map = new Long2IntOpenHashMap(size, 1.0f);
    map.defaultReturnValue(0);
  }

  public LongIntPairMapWritable(final LongIntPairMapWritable other) {
    final int size = Math.max(other.size(), 4);
    map = new Long2IntOpenHashMap(size);
    merge(other);
  }

  public int get(final long id) {
    return map.get(id);
  }

  public long getMaximumByValue() {
    long maxKey = -1L;
    int maxValue = Integer.MIN_VALUE;

    final Iterator<Long2IntMap.Entry> iter = map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      if (entry.getIntValue() > maxValue || (entry.getIntValue() == maxValue && entry.getLongKey() < maxKey)) {
        maxKey = entry.getLongKey();
        maxValue = entry.getIntValue();
      }
    }

    return maxKey;
  }

  public void addTo(final long id, final int value) {
    map.addTo(id, value);
  }

  public void merge(final LongIntPairMapWritable other) {
    final Iterator<Long2IntMap.Entry> iter = other.map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      map.addTo(entry.getLongKey(), entry.getIntValue());
    }
  }

  public void clear() {
    map.clear();
  }

  public int size() {
    return map.size();
  }

  @Override
  public void write(final DataOutput out) throws IOException {
    out.writeInt(map.size());

    final Iterator<Long2IntMap.Entry> iter = map.long2IntEntrySet().fastIterator();
    while (iter.hasNext()) {
      final Long2IntMap.Entry entry = iter.next();
      out.writeLong(entry.getLongKey());
      out.writeInt(entry.getIntValue());
    }
  }

  @Override
  public void readFields(final DataInput in) throws IOException {
    final int size = in.readInt();
    for (int i = 0; i < size; ++i) {
      map.put(in.readLong(), in.readInt());
    }
  }
}
