/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: ZookeeperUtils.java
 * - A utility class that manipulates Zookeeper data
 * Version: 3.0
 */
package pegasus.mapreduce.common.utils;

import org.apache.zookeeper.CreateMode;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.ZooDefs;
import org.apache.zookeeper.ZooKeeper;

import java.util.ArrayList;
import java.util.List;

public class ZookeeperUtils {
  public static void createPersistentNode(final ZooKeeper zk, final String path) {
    createNode(zk, path, CreateMode.PERSISTENT);
  }

  public static void createOrUpdatePersistentNode(final ZooKeeper zk, final String path, final String value) {
    final byte[] valueBytes = value.getBytes();
    if (exists(zk, path)) {
      try {
        zk.setData(path, valueBytes, -1);
      } catch (KeeperException | InterruptedException e) {
        // ignore
      }
    } else {
      createNode(zk, path, valueBytes, CreateMode.PERSISTENT);
    }
  }

  public static String getNodeData(final ZooKeeper zk, final String path) {
    try {
      return new String(zk.getData(path, false, null));
    } catch (KeeperException | InterruptedException e) {
      return "";
    }
  }

  public static void createEphemeralNode(final ZooKeeper zk, final String path) {
    createNode(zk, path, CreateMode.EPHEMERAL);
  }

  private static void createNode(final ZooKeeper zk, final String path, final CreateMode createMode) {
    createNode(zk, path, new byte[0], createMode);
  }

  private static void createNode(
      final ZooKeeper zk, final String path, final byte[] values, final CreateMode createMode) {
    try {
      zk.create(path, values, ZooDefs.Ids.OPEN_ACL_UNSAFE, createMode);
    } catch (KeeperException e) {
      switch (e.code()) {
        case NODEEXISTS:
          break;
        default:
          break;
      }
    } catch (InterruptedException e) {
      // ignore
    }
  }

  public static boolean exists(final ZooKeeper zk, final String path) {
    try {
      return zk.exists(path, false) != null;
    } catch (KeeperException | InterruptedException e) {
      return false;
    }
  }

  public static List<String> getChildren(final ZooKeeper zk, final String path) {
    try {
      return zk.getChildren(path, false);
    } catch (KeeperException | InterruptedException e) {
      return new ArrayList<>();
    }
  }

  public static void delete(final ZooKeeper zk, final String path) {
    try {
      zk.delete(path, -1);
    } catch (KeeperException | InterruptedException ignored) {
    }
  }
}
