/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: DiameterComputation.java
 * - Computes the radius and diameter of a graph
 * Version: 3.0
 */
package pegasus.mapreduce.diameter;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapreduce.*;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.zookeeper.KeeperException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.cache.*;
import pegasus.mapreduce.common.computation.ComputationMapper;
import pegasus.mapreduce.common.coordination.Barrier;
import pegasus.mapreduce.common.coordination.CoordinationService;
import pegasus.mapreduce.common.io.ComputationInputFormat;
import pegasus.mapreduce.common.io.VectorOutputFormat;
import pegasus.mapreduce.common.types.BitStrings;
import pegasus.mapreduce.common.types.IntArrayWritable;
import pegasus.mapreduce.common.types.IntDoublePairWritable;
import pegasus.mapreduce.common.utils.IDUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.List;
import java.util.Random;
import java.util.stream.Collectors;

public class DiameterComputation extends Configured implements Tool {
  private static final Logger LOG = LoggerFactory.getLogger(DiameterComputation.class);

  public static void main(final String[] args) throws Exception {
    int exitCode = ToolRunner.run(new DiameterComputation(), args);
    System.exit(exitCode);
  }

  /**
   * Runs the diameter and radii computation via CLI.
   *
   * @param args The CLI arguments
   * @return 0 if the computation succeeds, otherwise -1
   * @throws Exception
   */
  @Override
  public int run(final String[] args) throws Exception {
    if (args.length < 3) return -1;

    final Path input = new Path(args[0]);
    final Path intermediate = new Path(args[0] + ".bitvector");
    final Path output = new Path(args[1]);
    final int parallelism = Integer.valueOf(args[2]);
    final int maxIterations = getConf().getInt("maxIterations", -1);

    final int numIterations = runMVMultiplication(input, intermediate, parallelism, maxIterations);
    if (numIterations < 0) return numIterations;

    final int exitCode2 = runRadii(intermediate, output, parallelism, numIterations);
    if (exitCode2 < 0) return exitCode2;

    double effectiveDiameter = estimateEffectiveDiameter(numIterations);
    LOG.info("Effective diameter: {}", effectiveDiameter);

    return 0;
  }

  /**
   * Runs matrix-vector mulitplication for diameter computation
   *
   * @param input         The path to the pre-partitioned matrix
   * @param output        The path to the result of multiplication
   * @param parallelism   The number of workers
   * @param maxIterations The maximum number of iterations
   * @return The number of iterations if the computation succeeds, otherwise -1
   * @throws InterruptedException
   * @throws IOException
   * @throws ClassNotFoundException
   */
  private int runMVMultiplication(
      final Path input, final Path output, final int parallelism, final int maxIterations)
      throws InterruptedException, IOException, ClassNotFoundException {
    final FileSystem fs = FileSystem.get(getConf());
    fs.delete(output, true);

    final Job job = Job.getInstance(
        getConf(), String.format("PMV - DiameterComputation (step1, %s)", input.toString()));
    job.setJarByClass(DiameterComputation.class);

    final Configuration conf = job.getConfiguration();
    conf.setInt(Constants.MaxIterations, maxIterations);
    conf.setInt(Constants.Parallelism, parallelism);

    // input
    FileInputFormat.setInputPaths(job, input);
    FileInputFormat.setInputDirRecursive(job, true);
    job.setInputFormatClass(ComputationInputFormat.class);
    ComputationMapper.setEstimatedSize(job);

    // mapper
    job.setMapperClass(DiameterMapper.class);
    job.setMapOutputKeyClass(LongWritable.class);
    job.setMapOutputValueClass(DoubleWritable.class);

    // reducer
    job.setNumReduceTasks(0); // map-only job

    // output
    FileOutputFormat.setOutputPath(job, output);
    job.setOutputFormatClass(SequenceFileOutputFormat.class);
    job.setOutputKeyClass(LongWritable.class);
    job.setOutputValueClass(IntDoublePairWritable.class);

    // run multiplication job and find number of iterations
    if (!job.waitForCompletion(true)) return -1;
    final Counter numIter = job.getCounters().findCounter(Constants.CounterGroup, Constants.ExecutedIterations);
    return (int) numIter.getValue();
  }

  /**
   * Computes the radii of graph using the result of matrix-vector multiplication.
   *
   * @param input         The path to the result of multiplication
   * @param output        The path to the radii
   * @param parallelism   The number of workers
   * @param numIterations The number of iterations
   * @return 0 if the computations succeeds, otherwise -1
   * @throws InterruptedException
   * @throws IOException
   * @throws ClassNotFoundException
   */
  private int runRadii(final Path input, final Path output, final int parallelism, final int numIterations)
      throws InterruptedException, IOException, ClassNotFoundException {
    final Job job = Job.getInstance(
        getConf(), String.format("PMV - DiameterComputation (step2, %s)", input.toString()));
    job.setJarByClass(DiameterComputation.class);

    final Configuration conf = job.getConfiguration();
    conf.setInt(Constants.DiameterNumIterations, numIterations);

    // input
    FileInputFormat.setInputPaths(job, input);
    job.setInputFormatClass(SequenceFileInputFormat.class);

    // output
    FileOutputFormat.setOutputPath(job, output);
    job.setOutputFormatClass(VectorOutputFormat.class);

    // mapper
    job.setMapperClass(RadiiMapper.class);
    job.setMapOutputKeyClass(LongWritable.class);
    job.setMapOutputValueClass(IntDoublePairWritable.class);

    // reducer
    job.setReducerClass(RadiiReducer.class);
    job.setNumReduceTasks(parallelism);

    return job.waitForCompletion(true) ? 0 : -1;
  }

  /**
   * Estimates the effective diameter.
   *
   * @param numIterations The number of iterations
   * @return the estimated diameter if the computation succeeds, otherwise -1
   * @throws IOException
   */
  private double estimateEffectiveDiameter(final int numIterations) throws IOException {
    final FileSystem fs = FileSystem.get(getConf());
    final double[] neighborhoods = new double[numIterations];
    try (final BufferedReader reader = new BufferedReader(
        new InputStreamReader(fs.open(new Path(Constants.DiameterNeighborhoodFile))))) {
      final List<String> lines = reader.lines().collect(Collectors.toList());
      for (final String line : lines) {
        final String[] split = line.split("\t");
        neighborhoods[Integer.parseInt(split[0])] = Double.parseDouble(split[1]);
      }
    }

    double threshold = neighborhoods[numIterations - 1] * 0.9;
    for (int i = 1; i < numIterations; ++i) {
      if (neighborhoods[i] >= threshold) {
        double decimal = (threshold - neighborhoods[i - 1]) / (neighborhoods[i] - neighborhoods[i - 1]);
        return i - 1 + decimal;
      }
    }
    return -1;
  }

  /**
   * ComputationMapper for Diameter computation
   */
  public static class DiameterMapper
      extends ComputationMapper<IntArrayWritable, BitStrings, BitStrings, IntDoublePairWritable> {
    private static final Logger LOG = LoggerFactory.getLogger(DiameterMapper.class);
    private final Random random = new Random();
    private final BitStrings initValue = new BitStrings();

    /**
     * Given a vertex id, initialize a random bitstring for the vertex. Note that the initialization process
     * should be deterministic because the initialization process can be executed in parallel.
     *
     * @param id The id of vertex
     * @return The initialized random bitstring
     */
    private BitStrings initBitStrings(final long id) {
      random.setSeed(id);

      final long[] bits = initValue.getBits();
      for (int i = 0; i < Constants.DefaultNumBitstrings; ++i) {
        final double threshold = random.nextDouble();
        double prob = 0;
        for (int j = 0; j < 64; ++j) {
          prob += Math.pow(2, -j - 1);
          if (prob > threshold) {
            bits[i] = 1L << (63 - j);
            break;
          }
        }
      }

      return initValue;
    }

    /**
     * {@inheritDoc}
     *
     * @param numBlocks   The number of blocks
     * @param blockId     The current block id
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<BitStrings> createVectorCache(
        final int numBlocks, final int blockId, final long numVertices) {
      final VectorCache<BitStrings> cache = new BitStringsVectorCache(
          numBlocks, blockId, numVertices, new BitStrings());
      for (int i = 0, size = cache.size(); i < size; ++i) {
        final long globalId = IDUtils.localToGlobal(numBlocks, blockId, i);
        cache.put(i, initBitStrings(globalId));
      }
      return cache;
    }

    /**
     * {@inheritDoc}
     *
     * @param numBlocks   The number of blocks
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<BitStrings> createIntermediateVectorCache(final int numBlocks, final long numVertices) {
      return new BitStringsVectorCache(numBlocks, 0, numVertices, new BitStrings());
    }

    /**
     * {@inheritDoc}
     *
     * @param blockName The name of matrix block (sparse, dense)
     * @param conf      The Hadoop configuration object
     * @param numCols   The number of columns
     * @param numRows   The number of rows
     * @return
     * @throws IOException
     */
    @Override
    protected MatrixCache<IntArrayWritable> createMatrixCache(
        final String blockName, final Configuration conf, final int numCols, final int numRows)
        throws IOException {
      return new UnweightedMatrixCache(blockName, conf, numCols, numRows);
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<BitStrings> vectorValueClass() {
      return BitStrings.class;
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<BitStrings> intermediateVectorValueClass() {
      return BitStrings.class;
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<IntArrayWritable> matrixValueClass() {
      return IntArrayWritable.class;
    }

    /**
     * {@inheritDoc}
     *
     * @param matBlockId The id of matrix block
     * @param matrix     The matrix block
     * @param vecInput   The input vector block
     * @param vecOutput  The vector block to be stored
     * @throws Exception
     */
    @Override
    protected void combine2Op(
        final int matBlockId,
        final MatrixCache<IntArrayWritable> matrix, final VectorCache<BitStrings> vecInput,
        final VectorCache<BitStrings> vecOutput) throws Exception {
      final int numBlocks = numBlocks();
      final int srcBlockId = matBlockId % numBlocks;
      final int dstBlockId = matBlockId / numBlocks;
      final boolean selfBlock = srcBlockId == dstBlockId;

      final MatrixIterator<IntArrayWritable> iter = matrix.iterator();
      while (iter.next()) {
        final int col = iter.currentCol();
        final BitStrings colVec;
        if (vecInput.contains(col)) {
          colVec = vecInput.get(col);
        } else {
          colVec = initBitStrings(IDUtils.localToGlobal(numBlocks, srcBlockId, col));
        }

        final IntArrayWritable rows = iter.currentRows();
        for (int i = 0, size = rows.size(); i < size; ++i) {
          final int row = rows.get(i);
          vecOutput.addTo(row, colVec);
        }
      }
      iter.close();

      if (selfBlock) {
        final VectorIterator<BitStrings> vecIter = vecInput.iterator(false, null);
        while (vecIter.next()) {
          final int col = vecIter.currentId();
          final BitStrings colVec = vecIter.currentValue();
          vecOutput.addTo(col, colVec);
        }
      }
    }

    /**
     * {@inheritDoc}
     *
     * @param matBlockId The matrix block id
     * @param vecInput   The intemediate vector block
     * @param vecOutput  The output vector block
     * @throws Exception
     */
    @Override
    protected void combineAllOp(
        final int matBlockId,
        final RecordReader<IntWritable, BitStrings> vecInput,
        final VectorCache<BitStrings> vecOutput) throws Exception {
      while (vecInput.nextKeyValue()) {
        final int row = vecInput.getCurrentKey().get();
        final BitStrings value = vecInput.getCurrentValue();

        vecOutput.addTo(row, value);
      }
    }

    /**
     * {@inheritDoc}
     *
     * @param vecBlockId      The vector block id
     * @param vecResult       The vector block from previous iteration
     * @param vecIntermediate The result vector block
     * @throws Exception
     */
    @Override
    protected void applyOp(
        final int vecBlockId,
        final VectorCache<BitStrings> vecResult,
        final VectorCache<BitStrings> vecIntermediate) throws Exception {
      final int numBlocks = numBlocks();
      final int nIter = currentIteration();
      final LongWritable vertex = new LongWritable();
      final IntDoublePairWritable cardinality = new IntDoublePairWritable();

      final VectorIterator<BitStrings> iter = vecResult.iterator(false, null);
      while (iter.next()) {
        final int id = iter.currentId();
        final BitStrings oldValue = iter.currentValue();
        final BitStrings newValue = vecIntermediate.get(id);

        if (newValue.isEmpty() || oldValue.equals(newValue)) {
          // converged
          vertex.set(IDUtils.localToGlobal(numBlocks, vecBlockId, id));
          cardinality.set(nIter, oldValue.cardinality());
          computationContext().write(vertex, cardinality);
        } else {
          // not converged
          notifyChanged();
          vecResult.put(id, newValue);
        }
      }
    }

    /**
     * {@inheritDoc}
     *
     * @param id  The vertex id
     * @param vec The vector value
     * @return
     * @throws Exception
     */
    @Override
    protected IntDoublePairWritable finalizeOp(final int id, final BitStrings vec) throws Exception {
      throw new RuntimeException("finalizeOp should be ignored");
    }

    /**
     * {@inheritDoc}
     *
     * @param vecBlockId The block id of result vector block
     * @param vector     The vector cache
     * @throws Exception
     */
    @Override
    protected void writeOutputVecOp(final int vecBlockId, final VectorCache<BitStrings> vector) throws Exception {
      // ignore default output vector operation
    }
  }

  /**
   * Mapper for radii computation
   */
  public static class RadiiMapper
      extends Mapper<LongWritable, IntDoublePairWritable, LongWritable, IntDoublePairWritable> {
    // identity mapper
  }

  /**
   * Reducer for radii computation
   */
  public static class RadiiReducer extends Reducer<LongWritable, IntDoublePairWritable, LongWritable, IntWritable> {
    private final IntWritable radii = new IntWritable();
    private int numIterations;
    private double[] cardinalities;
    private double[] sumCardinalities;

    @SuppressWarnings("unchecked")
    @Override
    protected void setup(Context context) throws IOException, InterruptedException {
      numIterations = context.getConfiguration().getInt(Constants.DiameterNumIterations, -1);
      cardinalities = new double[numIterations];
      sumCardinalities = new double[numIterations];
    }

    /**
     * Given a vertex with estimation number of reachable set for each iteration,
     * computes the effective radii for the vertex.
     *
     * @param key     The vertex id
     * @param values  The estmiation number of reachable set for each iteration
     * @param context The Hadoop context object
     * @throws IOException
     * @throws InterruptedException
     */
    @Override
    protected void reduce(
        final LongWritable key, final Iterable<IntDoublePairWritable> values, final Context context)
        throws IOException, InterruptedException {
      // collect cardinalities
      int maxHop = Integer.MIN_VALUE;
      for (int i = 0; i < numIterations; ++i) cardinalities[i] = 0;
      for (final IntDoublePairWritable value : values) {
        final int hop = value.getInt();
        cardinalities[hop - 1] = value.getDouble();
        maxHop = Math.max(maxHop, hop);
      }
      for (int i = maxHop; i < numIterations; ++i) cardinalities[i] = cardinalities[i - 1];

      // estimate effective radii for each vertex
      final double threshold = cardinalities[numIterations - 1] * 0.9;
      for (int i = 0, size = numIterations; i < size; ++i) {
        if (cardinalities[i] >= threshold) {
          radii.set(i);
          context.write(key, radii);
          break;
        }
      }

      // aggregate results of neighborhood functions
      for (int i = 0; i < numIterations; ++i) sumCardinalities[i] += cardinalities[i];
    }

    @Override
    protected void cleanup(Context context) throws IOException, InterruptedException {
      // initialize coordination service
      final CoordinationService coordService = new CoordinationService();
      try {
        coordService.setup(context);
      } catch (KeeperException e) {
        throw new RuntimeException("Coordination service initialization failed.");
      }

      // aggregate
      final Barrier barrier = coordService.getSemaphoreBarrier(Constants.DiameterNeighborhood);
      barrier.setup();
      for (int i = 0; i < numIterations; ++i) {
        coordService.getZkCounter(Constants.DiameterNeighborhood + String.valueOf(i), Double.class)
            .update(sumCardinalities[i]);
      }
      barrier.markCompleted();

      if (coordService.getWorkerId() == 0) {
        barrier.waitBarrier();

        final FileSystem fs = FileSystem.get(context.getConfiguration());
        try (final OutputStreamWriter writer = new OutputStreamWriter(
            fs.create(new Path(Constants.DiameterNeighborhoodFile)))) {
          for (int i = 0; i < numIterations; ++i) {
            final double cardinality = coordService.getZkCounter(
                Constants.DiameterNeighborhood + String.valueOf(i), Double.class).getSum();
            writer.write(String.format("%d\t%f\n", i, cardinality));
          }
        }
      }
    }
  }
}
