/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: PageRankComputation.java
 * - PageRank implementation on PMV framework
 * Version: 3.0
 */
package pegasus.mapreduce.pagerank;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.cache.*;
import pegasus.mapreduce.common.computation.ComputationMapper;
import pegasus.mapreduce.common.coordination.Barrier;
import pegasus.mapreduce.common.coordination.CoordinationService;
import pegasus.mapreduce.common.coordination.Counter;
import pegasus.mapreduce.common.io.ComputationInputFormat;
import pegasus.mapreduce.common.io.VectorOutputFormat;
import pegasus.mapreduce.common.types.IntArrayWritable;
import pegasus.mapreduce.common.utils.IDUtils;

import java.io.IOException;

/**
 * Computes the PageRank score for the given graph.
 */
public class PageRankComputation extends Configured implements Tool {
  public static void main(final String[] args) throws Exception {
    int exitCode = ToolRunner.run(new PageRankComputation(), args);
    System.exit(exitCode);
  }

  /**
   * Runs the PageRank computation via CLI.
   *
   * @param args The CLI arguments
   * @return 0 if the computation succeeds, otherwise -1
   * @throws Exception
   */
  @Override
  public int run(final String[] args) throws Exception {
    if (args.length < 3) return -1;

    final Path input = new Path(args[0]);
    final Path output = new Path(args[1]);
    final int parallelism = Integer.valueOf(args[2]);
    final int maxIterations = getConf().getInt("maxIterations", -1);
    final boolean personalized = getConf().getBoolean("personalized", false);
    final long sourceVertex = getConf().getLong("sourceVertex", -1L);
    final double dampingFactor = getConf().getDouble("dampingFactor", Constants.DefaultPageRankDampingFactor);
    final double threshold = getConf().getDouble("threshold", Constants.DefaultPageRankThreshold);

    final Job job = Job.getInstance(getConf(), String.format("PMV - PageRankComputation (%s)", input.toString()));
    job.setJarByClass(PageRankComputation.class);

    final Configuration conf = job.getConfiguration();
    conf.setInt(Constants.MaxIterations, maxIterations);
    conf.setInt(Constants.Parallelism, parallelism);
    conf.setBoolean(Constants.PersonalizedPageRank, personalized);
    conf.setLong(Constants.PageRankSourceVertex, sourceVertex);
    conf.setDouble(Constants.PageRankDampingFactor, dampingFactor);
    conf.setDouble(Constants.PageRankThreshold, threshold);
    conf.setInt("mapreduce.task.timeout", 0);

    // input
    FileInputFormat.setInputPaths(job, input);
    FileInputFormat.setInputDirRecursive(job, true);
    job.setInputFormatClass(ComputationInputFormat.class);
    ComputationMapper.setEstimatedSize(job);

    // mapper
    job.setMapperClass(PageRankMapper.class);
    job.setMapOutputKeyClass(LongWritable.class);
    job.setMapOutputValueClass(DoubleWritable.class);

    // reducer
    job.setNumReduceTasks(0); // map-only job

    // output
    FileOutputFormat.setOutputPath(job, output);
    job.setOutputFormatClass(VectorOutputFormat.class);
    job.setOutputKeyClass(LongWritable.class);
    job.setOutputValueClass(DoubleWritable.class);

    return job.waitForCompletion(true) ? 0 : -1;
  }

  /**
   * ComputationMapper for PageRank
   */
  public static class PageRankMapper
      extends ComputationMapper<IntArrayWritable, DoubleWritable, DoubleWritable, DoubleWritable> {
    private final static Logger LOG = LoggerFactory.getLogger(PageRankMapper.class);

    private double threshold;
    private double rankSum;
    private double alpha;
    private long sourceVertex;
    private boolean personalized;
    private final DoubleWritable zero = new DoubleWritable(0.0);

    /**
     * Initializes computation and retrieves configuration.
     *
     * @param context      The Hadoop context object
     * @param coordService The coordination service object
     */
    @Override
    protected void setupComputation(final Context context, final CoordinationService coordService) {
      final Configuration conf = context.getConfiguration();

      // retrieve parameters for PageRank algorithm
      alpha = conf.getDouble(Constants.PageRankDampingFactor, Constants.DefaultPageRankDampingFactor);
      personalized = conf.getBoolean(Constants.PersonalizedPageRank, false);
      sourceVertex = conf.getLong(Constants.PageRankSourceVertex, -1L);
      threshold = conf.getDouble(Constants.PageRankThreshold, Constants.DefaultPageRankThreshold);
    }

    /**
     * Finalizes the PageRank computation. This method computes the normalization constant for PageRank scores
     * to deal with dead-end vertices.
     *
     * @param context      The Hadoop context object
     * @param coordService The coordination service object
     */
    @Override
    protected void cleanupComputation(final Context context, final CoordinationService coordService) {
      // calculate sum of PageRank scores
      double partialSum = 0;
      final VectorIterator<DoubleWritable> iter = currentVector().iterator(false, null);
      while (iter.next()) {
        partialSum += iter.currentValue().get();
      }

      // update sum of PageRank scores
      final Counter<Double> rankSumCounter = coordService.getZkCounter(
          Constants.PageRankSumCounter, Double.class);
      rankSumCounter.update(partialSum);

      final Barrier barrier = coordService.getSemaphoreBarrier(Constants.PageRankSumBarrier);
      barrier.markCompleted();
      barrier.waitBarrier();

      rankSum = rankSumCounter.getSum();
    }

    /**
     * {@inheritDoc}
     *
     * @param numBlocks   The number of blocks
     * @param blockId     The current block id
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<DoubleWritable> createVectorCache(
        final int numBlocks, final int blockId, final long numVertices) {
      final DoubleVectorCache cache = new DoubleVectorCache(numBlocks, blockId, numVertices, zero);
      final DoubleWritable init = new DoubleWritable(1.0 / numVertices());
      for (int i = 0, size = cache.size(); i < size; ++i) {
        if (!personalized) {
          cache.put(i, init);
        } else {
          if (IDUtils.localToGlobal(numBlocks, blockId, i) == sourceVertex) {
            cache.put(i, new DoubleWritable(1.0));
          }
        }
      }
      return cache;
    }

    /**
     * {@inheritDoc}
     *
     * @param numBlocks   The number of blocks
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<DoubleWritable> createIntermediateVectorCache(
        final int numBlocks, final long numVertices) {
      return new DoubleVectorCache(numBlocks, 0, numVertices, zero);
    }

    /**
     * {@inheritDoc}
     *
     * @param blockName The name of matrix block (sparse, dense)
     * @param conf      The Hadoop configuration object
     * @param numCols   The number of columns
     * @param numRows   The number of rows
     * @return
     * @throws IOException
     */
    @Override
    protected MatrixCache<IntArrayWritable> createMatrixCache(
        final String blockName, final Configuration conf, final int numCols, final int numRows)
        throws IOException {
      return new UnweightedMatrixCache(blockName, conf, numCols, numRows);
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<DoubleWritable> vectorValueClass() {
      return DoubleWritable.class;
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<DoubleWritable> intermediateVectorValueClass() {
      return DoubleWritable.class;
    }

    /**
     * {@inheritDoc}
     *
     * @return
     */
    @Override
    protected Class<IntArrayWritable> matrixValueClass() {
      return IntArrayWritable.class;
    }

    /**
     * {@inheritDoc}
     *
     * @param matBlockId The id of matrix block
     * @param matrix     The matrix block
     * @param vecInput   The input vector block
     * @param vecOutput  The vector block to be stored
     * @throws Exception
     */
    @Override
    protected void combine2Op(
        final int matBlockId,
        final MatrixCache<IntArrayWritable> matrix, final VectorCache<DoubleWritable> vecInput,
        final VectorCache<DoubleWritable> vecOutput) throws Exception {
      final DoubleWritable value = new DoubleWritable();
      final MatrixIterator<IntArrayWritable> iter = matrix.iterator();
      while (iter.next()) {
        final int col = iter.currentCol();
        final int degree = iter.numNonzeros();
        final IntArrayWritable rows = iter.currentRows();
        value.set(vecInput.get(col).get() / degree);
        for (int i = 0, size = rows.size(); i < size; ++i) {
          final int row = rows.get(i);
          vecOutput.addTo(row, value);
        }
      }
      iter.close();
    }

    /**
     * {@inheritDoc}
     *
     * @param matBlockId The matrix block id
     * @param vecInput   The intemediate vector block
     * @param vecOutput  The output vector block
     * @throws Exception
     */
    @Override
    protected void combineAllOp(
        final int matBlockId,
        final RecordReader<IntWritable, DoubleWritable> vecInput, final VectorCache<DoubleWritable> vecOutput)
        throws Exception {
      while (vecInput.nextKeyValue()) {
        final int id = vecInput.getCurrentKey().get();
        vecOutput.addTo(id, vecInput.getCurrentValue());
      }
    }

    /**
     * {@inheritDoc}
     *
     * @param vecBlockId      The vector block id
     * @param vecResult       The vector block from previous iteration
     * @param vecIntermediate The result vector block
     * @throws Exception
     */
    @Override
    protected void applyOp(
        final int vecBlockId,
        final VectorCache<DoubleWritable> vecResult, final VectorCache<DoubleWritable> vecIntermediate)
        throws Exception {
      final long numVertices = numVertices();
      final DoubleWritable value = new DoubleWritable();
      final VectorIterator<DoubleWritable> iter = vecResult.iterator(false, null);
      while (iter.next()) {
        final int id = iter.currentId();
        final double prevScore = iter.currentValue().get();
        final double newScore = alpha * vecIntermediate.get(id).get() + (1 - alpha) / numVertices;
        if (Math.abs(prevScore - newScore) > threshold) {
          notifyChanged();
          value.set(newScore);
          vecResult.put(id, value);
        }
      }
    }

    /**
     * {@inheritDoc}
     * <p>
     * This method normalizes the PageRank score using the normalizing constant.
     *
     * @param id  The vertex id
     * @param vec The vector value
     * @return
     * @throws Exception
     */
    @Override
    protected DoubleWritable finalizeOp(final int id, final DoubleWritable vec) throws Exception {
      vec.set(vec.get() / rankSum);
      return vec;
    }
  }
}
