/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: SSSPComputation.java
 * - Single source shortest path implementation on PMV framework
 * Version: 3.0
 */
package pegasus.mapreduce.sssp;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import pegasus.mapreduce.common.Constants;
import pegasus.mapreduce.common.cache.*;
import pegasus.mapreduce.common.computation.ComputationMapper;
import pegasus.mapreduce.common.coordination.CoordinationService;
import pegasus.mapreduce.common.io.ComputationInputFormat;
import pegasus.mapreduce.common.io.VectorOutputFormat;
import pegasus.mapreduce.common.types.IntDoublePairArrayWritable;
import pegasus.mapreduce.common.utils.IDUtils;

import java.io.IOException;

/**
 * Computes the shortest path from given single source.
 */
public class SSSPComputation extends Configured implements Tool {
  public static void main(String[] args) throws Exception {
    int exitCode = ToolRunner.run(new SSSPComputation(), args);
    System.exit(exitCode);
  }

  /**
   * Runs the shortest path computation via CLI.
   *
   * @param args The CLI arguments
   * @return 0 if the computation succeeds, otherwise -1
   * @throws Exception
   */
  @Override
  public int run(String[] args) throws Exception {
    if (args.length < 4) return -1;

    final Path input = new Path(args[0]);
    final Path output = new Path(args[1]);
    final int parallelism = Integer.parseInt(args[2]);
    final int sourceVertex = Integer.parseInt(args[3]);
    final int maxIterations = getConf().getInt("maxIterations", -1);

    final Job job = Job.getInstance(getConf(), String.format("PMV - SSSPComputation (%s)", input.toString()));
    job.setJarByClass(SSSPComputation.class);

    final Configuration conf = job.getConfiguration();
    conf.setInt(Constants.Parallelism, parallelism);
    conf.setInt(Constants.SSSPSourceVertex, sourceVertex);
    conf.setInt(Constants.MaxIterations, maxIterations);
    conf.setInt("mapreduce.task.timeout", 0);
    ComputationMapper.setEstimatedSize(job);

    // input
    FileInputFormat.setInputPaths(job, input);
    FileInputFormat.setInputDirRecursive(job, true);
    job.setInputFormatClass(ComputationInputFormat.class);
    ComputationMapper.setEstimatedSize(job);

    // mapper
    job.setMapperClass(SSSPMapper.class);
    job.setMapOutputKeyClass(LongWritable.class);
    job.setMapOutputValueClass(DoubleWritable.class);

    // reducer
    job.setNumReduceTasks(0); // map-only job

    // output
    FileOutputFormat.setOutputPath(job, output);
    job.setOutputFormatClass(VectorOutputFormat.class);
    job.setOutputKeyClass(LongWritable.class);
    job.setOutputValueClass(DoubleWritable.class);

    return job.waitForCompletion(true) ? 0 : -1;
  }

  /**
   * ComputationMapper for shortest path computation
   */
  public static class SSSPMapper
      extends ComputationMapper<IntDoublePairArrayWritable, DoubleWritable, DoubleWritable, DoubleWritable> {
    private long sourceVertex;
    private final DoubleWritable infinite = new DoubleWritable(Double.MAX_VALUE / 2);

    /**
     * {@inheritDoc}
     * 
     * @param context      The Hadoop context object
     * @param coordService The coordination service object
     */
    @Override
    protected void setupComputation(final Context context, final CoordinationService coordService) {
      sourceVertex = context.getConfiguration().getLong(Constants.SSSPSourceVertex, -1L);
    }

    /**
     * {@inheritDoc}
     * 
     * @param numBlocks   The number of blocks
     * @param blockId     The current block id
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<DoubleWritable> createVectorCache(
        final int numBlocks, final int blockId, final long numVertices) {
      final DoubleVectorCache cache = new DoubleVectorCache(numBlocks, blockId, numVertices, infinite);
      for (int i = 0, size = cache.size(); i < size; ++i) {
        if (IDUtils.localToGlobal(numBlocks, blockId, i) == sourceVertex) {
          cache.put(i, new DoubleWritable());
        } else {
          cache.put(i, infinite);
        }
      }
      return cache;
    }

    /**
     * {@inheritDoc}
     * 
     * @param numBlocks   The number of blocks
     * @param numVertices The number of vertices
     * @return
     */
    @Override
    protected VectorCache<DoubleWritable> createIntermediateVectorCache(
        final int numBlocks, final long numVertices) {
      return new DoubleVectorCache(numBlocks, 0, numVertices, infinite);
    }

    /**
     * {@inheritDoc}
     * 
     * @param blockName The name of matrix block (sparse, dense)
     * @param conf      The Hadoop configuration object
     * @param numCols   The number of columns
     * @param numRows   The number of rows
     * @return
     * @throws IOException
     */
    @Override
    protected MatrixCache<IntDoublePairArrayWritable> createMatrixCache(
        final String blockName, final Configuration conf, final int numCols, final int numRows)
        throws IOException {
      return new WeightedMatrixCache(blockName, conf, numCols, numRows);
    }

    /**
     * {@inheritDoc}
     * 
     * @return
     */
    @Override
    protected Class<DoubleWritable> vectorValueClass() {
      return DoubleWritable.class;
    }

    /**
     * {@inheritDoc}
     * 
     * @return
     */
    @Override
    protected Class<DoubleWritable> intermediateVectorValueClass() {
      return DoubleWritable.class;
    }

    /**
     * {@inheritDoc}
     * 
     * @return
     */
    @Override
    protected Class<IntDoublePairArrayWritable> matrixValueClass() {
      return IntDoublePairArrayWritable.class;
    }

    /**
     * {@inheritDoc}
     * 
     * @param matBlockId The id of matrix block
     * @param matrix     The matrix block
     * @param vecInput   The input vector block
     * @param vecOutput  The vector block to be stored
     * @throws Exception
     */
    @Override
    protected void combine2Op(
        final int matBlockId,
        final MatrixCache<IntDoublePairArrayWritable> matrix,
        final VectorCache<DoubleWritable> vecInput,
        final VectorCache<DoubleWritable> vecOutput) throws Exception {
      final DoubleWritable min = new DoubleWritable();
      final MatrixIterator<IntDoublePairArrayWritable> iter = matrix.iterator();
      while (iter.next()) {
        final int col = iter.currentCol();
        final double colDist = vecInput.get(col).get();
        if (colDist < infinite.get()) {
          final IntDoublePairArrayWritable rows = iter.currentRows();
          for (int i = 0, size = rows.size(); i < size; ++i) {
            final int row = rows.getInt(i);
            final double weight = rows.getDouble(i);
            min.set(Math.min(vecOutput.get(row).get(), colDist + weight));
            vecOutput.put(row, min);
          }
        }
      }
      iter.close();
    }

    /**
     * {@inheritDoc}
     * 
     * @param matBlockId The matrix block id
     * @param vecInput   The intemediate vector block
     * @param vecOutput  The output vector block
     * @throws Exception
     */
    @Override
    protected void combineAllOp(
        final int matBlockId,
        final RecordReader<IntWritable, DoubleWritable> vecInput,
        final VectorCache<DoubleWritable> vecOutput)
        throws Exception {
      final DoubleWritable min = new DoubleWritable();
      while (vecInput.nextKeyValue()) {
        final int row = vecInput.getCurrentKey().get();
        final double dist = vecInput.getCurrentValue().get();
        min.set(Math.min(dist, vecOutput.get(row).get()));
        vecOutput.put(row, min);
      }
    }

    /**
     * {@inheritDoc}
     * 
     * @param vecBlockId      The vector block id
     * @param vecResult       The vector block from previous iteration
     * @param vecIntermediate The result vector block
     * @throws Exception
     */
    @Override
    protected void applyOp(
        final int vecBlockId,
        final VectorCache<DoubleWritable> vecResult,
        final VectorCache<DoubleWritable> vecIntermediate) throws Exception {
      final DoubleWritable value = new DoubleWritable();
      final VectorIterator<DoubleWritable> iter = vecResult.iterator(false, null);
      while (iter.next()) {
        final int id = iter.currentId();
        final double oldDist = iter.currentValue().get();
        final double newDist = vecIntermediate.get(id).get();

        if (oldDist > newDist) {
          notifyChanged();
          value.set(newDist);
          vecResult.put(id, value);
        }
      }
    }
  }
}
