/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: SeedGenerator.java
 * - Generate an input file for the subgraph enumeration step of PSE
 * Version: 3.0
 */

package pegasus.mapreduce.subgraph;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.stream.Collectors;

public class SeedGenerator {

	/**
	 * Generate subproblems according to the number of colors and the number of nodes in a given query graph.
	 * Two types of subproblems is generated: The first type contains 'numPatternNodes' colors, and the
	 * second type contains 'numPatternNodes - 1' colors.
	 * @param numColors the number of colors
	 * @param numPatternNodes the number of nodes in a query graph
	 * @return
	 */
	public static List<String> dual(int numColors, int numPatternNodes) {
		
		long[] interProblemIds = Combination.get(numColors,numPatternNodes);
		long[] intraProblemIds = Combination.get(numColors,numPatternNodes-1);

		@SuppressWarnings("unchecked")
		List<Long>[] intraProblems = (List<Long>[]) new ArrayList[intraProblemIds.length];
		for(int i = 0; i < intraProblemIds.length; i++)
			intraProblems[i] = new ArrayList<Long>();
		
		
		for(int c=numPatternNodes-1; c > 0; c--){
			
			long[] problemIds = Combination.get(numColors, c);
			
			for(int i=0; i < problemIds.length; i++){
				
				long problemId = problemIds[i];

				int minNumProblems = Integer.MAX_VALUE;
				for(int j = 0; j < intraProblemIds.length; j++){
					if((intraProblemIds[j] & problemId) == problemId)
						minNumProblems = Math.min(minNumProblems, intraProblems[j].size());
				}
				
				List<Integer> minIndexes = new ArrayList<>(); 
				for(int j = 0; j < intraProblemIds.length; j++){
					if((intraProblemIds[j] & problemId) == problemId && intraProblems[j].size() == minNumProblems)
						minIndexes.add(j);
				}
				
				int minIndex = minIndexes.get(new Random().nextInt(minIndexes.size()));
				
				intraProblems[minIndex].add(problemId);
			}
			
		}
		
		List<String> res = new ArrayList<>();
		
		for(int i=0; i<interProblemIds.length; i++)
			res.add(String.format("%030d", 0) + Long.toBinaryString((1l << numColors) + interProblemIds[i]).substring(1));
		
		for(int i=0; i<intraProblems.length; i++)
			res.add(intraProblems[i].stream().map(x -> Long.toBinaryString((1l << numColors) + x).substring(1)).collect(Collectors.joining("\t")));
		
		return res;
	}
	
	
	private static class Combination{
		
		private static int n;
		private static int r;
		private static long[] items;
		private static int idx_item;

		/**
		 * get every possible bit-sequences that r bits are set among n bits.
		 * @param n
		 * @param r
		 * @return
		 */
		public static long[] get(int n, int r){
			Combination.n = n;
			Combination.r = r;
			
			int numItems = 1;
			for(int i = 0; i < r; i++){
				numItems *= n-i;
				numItems /= i+1;
			}
			
			items = new long[numItems];
			idx_item = 0;
			
			dfs(0);
			
			return items;
		}
		
		private static void dfs(long item){
			if(Long.bitCount(item) == r){
				items[idx_item++] = item; //add item
			}
			else{
				//find the position of msb(most significant bit) + 1
				int i;
				for(i=0; i < n && (item >>> i ) > 0; i++);
				
				for(; i < n; i++){
					dfs(item | (1l << i));
				}
			}
			
		}
		
	}
	
}
