/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: EdgeOrder.java
 * - Make an oriented graph from an undirected graph by vertex degree.
 * Version: 3.0
 */

package pegasus.mapreduce.triangle;

import java.io.IOException;
import java.util.StringTokenizer;

import org.apache.hadoop.fs.FileSystem;
import pegasus.mapreduce.common.io.LongPairWritable;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile.CompressionType;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import gnu.trove.list.array.TLongArrayList;

public class EdgeOrder extends Configured implements Tool{

	/**
	 * The main entry point
	 * @param args [0]: input file path. [1]: output file path.
	 * @throws Exception by hadoop
	 */
	public static void main(String args[]) throws Exception{
		ToolRunner.run(new EdgeOrder(), args);
	}
	
	public int run(String[] args) throws Exception{
		
		String input = args[0];
		String output = args[1];
		String intermediate = output + ".tmp";
		
		step1(input, intermediate);
		step2(intermediate, output);

		FileSystem.get(getConf()).delete(new Path(intermediate), true);

		return 0;
	}

	/**
	 * Submit the hadoop job for the first step of EdgeOrder
	 * @param input file path
	 * @param intermediate file path
	 * @throws Exception by hadoop
	 */
	private void step1(String input, String intermediate) throws Exception {
		
		Configuration conf = getConf();

		
		Job job = Job.getInstance(conf, String.format("[EdgeOrder-step1] %s -> %s", input, intermediate));
		job.setJarByClass(this.getClass());
		job.setMapperClass(EOStep1Mapper.class);
		job.setReducerClass(EOStep1Reducer.class);
		job.setMapOutputKeyClass(LongWritable.class);
		job.setMapOutputValueClass(LongWritable.class);
		job.setOutputKeyClass(LongPairWritable.class);
		job.setOutputValueClass(IntWritable.class);
		
		job.setOutputFormatClass(SequenceFileOutputFormat.class);

		FileInputFormat.setInputPaths(job, new Path(input));
		FileOutputFormat.setOutputPath(job, new Path(intermediate));
		
		job.waitForCompletion(false);
		
	}

	/**
	 * Submit the hadoop job for the second step of EdgeOrder
	 * @param intermediate file path
	 * @param output file path
	 * @throws Exception by hadoop
	 */
	private void step2(String intermediate, String output) throws Exception {
		
		Configuration conf = getConf();

		
		Job job = Job.getInstance(conf, String.format("[EdgeOrder-step2] %s -> %s", intermediate, output));
		job.setJarByClass(this.getClass());
		job.setMapperClass(EOStep2Mapper.class);
		job.setReducerClass(EOStep2Reducer.class);
		job.setMapOutputKeyClass(LongPairWritable.class);
		job.setMapOutputValueClass(IntWritable.class);
		job.setOutputKeyClass(LongWritable.class);
		job.setOutputValueClass(LongWritable.class);
		
		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setOutputFormatClass(SequenceFileOutputFormat.class);

		FileInputFormat.setInputPaths(job, new Path(intermediate));
		FileOutputFormat.setOutputPath(job, new Path(output));
		
		job.waitForCompletion(false);
		
	}

	public static class EOStep1Mapper extends Mapper<Object, Text, LongWritable, LongWritable>{
	
		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();

		/**
		 * The map function of the first step
		 * @param key not used
		 * @param value an edge as one line of text
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(Object key, Text value, Mapper<Object, Text, LongWritable, LongWritable>.Context context)
				throws IOException, InterruptedException {
			
			StringTokenizer st = new StringTokenizer(value.toString());
			long u = Long.parseLong(st.nextToken());
			long v = Long.parseLong(st.nextToken());
			
			if(u == v) return;
			
			ou.set(u);
			ov.set(v);
			
			context.write(ou, ov);
			context.write(ov, ou);
			
		}
		
	}

	public static class EOStep1Reducer extends Reducer<LongWritable, LongWritable, LongPairWritable, IntWritable>{
	
		LongPairWritable pair = new LongPairWritable();
		IntWritable ouDeg = new IntWritable();

		/**
		 * The reduce function of the first step
		 * @param _u a vertex
		 * @param uN neighbors of the vertex
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void reduce(LongWritable _u, Iterable<LongWritable> uN,
				Reducer<LongWritable, LongWritable, LongPairWritable, IntWritable>.Context context)
				throws IOException, InterruptedException {
			
			long u = _u.get();
			
			TLongArrayList buffer = new TLongArrayList();
			
			for(LongWritable _v : uN ) {
				buffer.add(_v.get());
			}

			buffer.sort();
			
			long prev = -1;
			int uDeg = 0;
			int bufferLen = buffer.size();
			for(int i=0; i<bufferLen;i++) {
				long v = buffer.getQuick(i);
				if(prev != v) {
					uDeg++;
					prev = v;
				}
			}
			
			ouDeg.set(uDeg);
			
			prev = -1;
			for(int i=0; i<bufferLen;i++) {
				long v = buffer.getQuick(i);
				if(prev != v) {
	
					pair.set(u, v);
					context.write(pair, ouDeg);
					
					prev = v;
				}
			}
			
		}
		
	}

	public static class EOStep2Mapper extends Mapper<LongPairWritable, IntWritable, LongPairWritable, IntWritable>{

		LongPairWritable pair_reverse = new LongPairWritable();
		IntWritable uDeg_dest = new IntWritable();

		/**
		 * The map function of the second step
		 * @param pair an edge. The high 32 bits are of a source vertex and thelow 32 bits are of a destination vertex.
		 * @param uDeg The degree of the source vertex
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(LongPairWritable pair, IntWritable uDeg,
				Mapper<LongPairWritable, IntWritable, LongPairWritable, IntWritable>.Context context)
				throws IOException, InterruptedException {
			
			long u = pair.i;
			long v = pair.j;
			
			if(u < v) {
				context.write(pair, uDeg);
			}
			else {
				pair_reverse.set(v, u);
				uDeg_dest.set(-uDeg.get());
				
				context.write(pair_reverse, uDeg_dest);
			}
		}
		
	}
	
	public static class EOStep2Reducer extends Reducer<LongPairWritable, IntWritable, LongWritable, LongWritable>{

		LongWritable ou = new LongWritable();
		LongWritable ov = new LongWritable();

		/**
		 * The reduce function of the second step
		 * @param pair an edge (u, v) such that u < v
		 * @param Degs a list of degree of source vertices. If a degree is less then 0, the degree is of destination vertex.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void reduce(LongPairWritable pair, Iterable<IntWritable> Degs,
				Reducer<LongPairWritable, IntWritable, LongWritable, LongWritable>.Context context)
				throws IOException, InterruptedException {
			
			long u = pair.i;
			long v = pair.j;

			
			
			int uDeg = -1, vDeg = -1;
			
			for(IntWritable _deg : Degs) {
				
				int deg = _deg.get();
				
				if(deg < 0) vDeg = -deg;
				else uDeg = deg;
				
			}
			
			
			
			ou.set(u);
			ov.set(v);
			
			if(uDeg < vDeg || (uDeg == vDeg && u < v) )
				context.write(ou, ov);
			else
				context.write(ov, ou);
			
		}
		
	}

	
}
