/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: PTE.java
 * - PTE with the sc optimization.
 * - count or list triangles in a graph
 * Version: 3.0
 */

package pegasus.mapreduce.triangle;

import java.io.IOException;
import java.util.*;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.NLineInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.NullOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.xerial.snappy.SnappyInputStream;

import com.esotericsoftware.kryo.KryoException;
import com.esotericsoftware.kryo.io.Input;

import gnu.trove.list.array.TIntArrayList;

public class PTE extends Configured implements Tool{

	long SHUFFLED_BYTES = 0;
	long HDFS_BYTES_READ = 0;
	long TRIANGLES = 0;

	/**
	 * The main entry point
	 * @param args [0]: input file path. [1]: output file path.
	 *             Tool runner arguments:
	 *          -D numColors the number of colors
	 * @throws Exception by hadoop
	 */
	public static void main(String args[]) throws Exception{
		ToolRunner.run(new PTE(), args);
	}

	/**
	 * submit the hadoop job of PTE
	 * @param args [0]: input file path. [1]: output file path (counting mode if not given).
	 *             Tool runner arguments:
	 *          -D numColors the number of colors
	 * @return not used
	 * @throws Exception by hadoop
	 */
	public int run(String[] args) throws Exception{
		
		long pt, et;

		Configuration conf = getConf();
		int numColors = conf.getInt("numColors", 0);

		if(args.length < 1 || numColors == 0){
			System.out.println("parameters: [inputFilePath] [outputFilePath (counting mode if not given)] " +
					"-DnumColors [numColors] ");
			return 0;
		}

		String input = args[0];
		String output = null;
		if(args.length < 2) System.err.println("counting mode.");
		else{
			output = args[1];
			System.err.println("listing mode. The output file is stored in " + output);
		}
		
		String part = input + ".part";
		String seed = input + ".seed";

		pt = System.currentTimeMillis();
		
		PrePartition pp = new PrePartition();
		pp.setConf(conf);
		pp.partitionGraph(input, part, numColors);
		int nodeSize = pp.nodeSize;
		System.out.println("nodeSize: " + nodeSize);
		
		pt = (System.currentTimeMillis() - pt)/1000;

		createSeedFile(seed, numColors);
		
		et = System.currentTimeMillis();
		enumerateTriangles(numColors, nodeSize, part, seed, output);
		et = (System.currentTimeMillis() - et)/1000;
		
		System.out.println("partitionTime\tEnumerationTime\tShuffle\tNetwork\tTriangles");
		System.out.println("[" + this.getClass().getSimpleName() + "]" + input + "\t" + pt + "\t"
				+ et + "\t" + SHUFFLED_BYTES + "\t" + HDFS_BYTES_READ + "\t" + TRIANGLES);
		
		FileSystem.get(getConf()).delete(new Path(part), true);
		FileSystem.get(getConf()).delete(new Path(seed), true);
		
		return 0;
	}

	/**
	 * submit a hadoop job for triangle enumeration
	 * @param numColors the number of colors
	 * @param nodeSize the number of nodes in a graph
	 * @param part directory path of partitioned graphs
	 * @param seed file path
	 * @param output file path
	 * @throws IOException by hadoop
	 * @throws InterruptedException by hadoop
	 * @throws ClassNotFoundException by hadoop
	 */
	private void enumerateTriangles(int numColors, int nodeSize, String part,
									String seed, String output) throws IOException,
			InterruptedException, ClassNotFoundException{

		Job job;

		if(output != null){
			//listing mode
			job = Job.getInstance(getConf(), "[" + this.getClass().getSimpleName() + "_listing]" + seed);
			job.setMapperClass(TriangleListingMapper.class);
			job.setOutputFormatClass(TextOutputFormat.class);
			job.setOutputKeyClass(Text.class);
			job.setOutputValueClass(NullWritable.class);
			FileOutputFormat.setOutputPath(job, new Path(output));
		}
		else{
			//counting mode
			job = Job.getInstance(getConf(), "[" + this.getClass().getSimpleName() + "_counting]" + seed);
			job.setMapperClass(TriangleCountingMapper.class);
			job.setOutputFormatClass(NullOutputFormat.class);
		}

		Configuration conf = job.getConfiguration();
		conf.set("mapreduce.map.java.opts", conf.get("mapreduce.map.java.opts", "") +
				" -XX:+UseG1GC -XX:+UseLargePagesInMetaspace");


		job.setJarByClass(getClass());
		job.setNumReduceTasks(0);
		conf.setInt("mapred.line.input.format.linespermap", 1);
		conf.set("part", part);
		conf.setInt("numColors", numColors);
		conf.setInt("nodeSize", nodeSize);
		job.setInputFormatClass(NLineInputFormat.class);
		NLineInputFormat.setInputPaths(job, new Path(seed));


		job.getConfiguration().setLong("mapred.task.timeout", 0L);
		job.getConfiguration().setLong("mapred.min.split.size", Long.MAX_VALUE);
		
		job.waitForCompletion(false);
		
		HDFS_BYTES_READ = job.getCounters().findCounter("FileSystemCounters", "HDFS_BYTES_READ").getValue();
		TRIANGLES = job.getCounters().findCounter("PTE", "triangles").getValue();

		System.err.println("triangles : " + job.getCounters().findCounter("PTE", "triangles").getValue());
	}

	/**
	 * create seed file for triangle enumeration of PTE
	 * @param seedFile seed file path
	 * @param p the number of colors
	 * @throws IOException by hadoop
	 */
	private void createSeedFile(String seedFile, int p) throws IOException{
		
		Configuration conf = getConf();
		
		List<String> seedList = new ArrayList<>();
		
		FSDataOutputStream seedOut = FileSystem.get(conf).create(new Path(seedFile));
		
		for(int i=0; i < p-2; i++){
			for(int j=i+1; j < p-1; j++){
				for(int k=j+1; k < p; k++){
					
					seedList.add(String.format("%10d %10d %10d\n", i,j,k));
				}
			}
		}
		
		for(int i=0; i < p-1; i++){
			for(int j=i+1; j < p; j++){
				seedList.add(String.format("%10d %10d\n", i,j));
			}
		}
		
		Collections.shuffle(seedList);
		
		System.out.printf("[Info] %d subproblem(s) to process\n", seedList.size());
		for(String s : seedList){
			seedOut.writeBytes(s);
		}
		
		seedOut.close();

	}

	public static class TriangleListingMapper extends TriangleCountingMapper{

		Context context;
		Text out = new Text();
		NullWritable nul = NullWritable.get();

		/**
		 * setup before execution.
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context) throws IOException, InterruptedException {
			super.setup(context);
			this.context = context;
		}

		/**
		 * Listing triangles.
		 * It finds all triangles (u, v, w) that the colors of the vertices u, v, and w are i, j, and k, respectively.
		 * @param ij adjacency list of an edge set
		 * @param ik adjacency list of an edge set
		 * @param jk adjacency list of an edge set
		 * @param i the first node color
		 * @param j the second node color
		 * @param k the last node color
		 * @return the number of triangles of colors i, j, and k.
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected long enumTri(int[][] ij, int[][] ik, int[][] jk,
							   int i, int j, int k) throws IOException, InterruptedException {

			long t = System.currentTimeMillis();

			long count = 0;

			for(int u = 0; u < nodeSize; u++){
				if(ij[u] != null){
					for(int v : ij[u]){
						Iterator<Integer> it = intersectIterator(ik[u], jk[v]);

						try {
							//noinspection InfiniteLoopStatement
							while (true) {
								int w = it.next();
								String outString = (((long) u) * numColors + i) + "\t" +
										(((long) v) * numColors + j) + "\t" +
										(((long) w) * numColors + k);
								out.set(outString);
								context.write(out, nul);
								count++;
							}
						} catch (NoSuchElementException ignored){}
					}
				}
			}

			System.out.println("counting: " + ((System.currentTimeMillis() - t)/1000.0) + "s");

			return count;

		}

		/**
		 * intersection of two arrays
		 * @param uAdj a sorted int array
		 * @param vAdj another sorted int array
		 * @return the iterator of the two arrays' intersection
		 */
		private Iterator<Integer> intersectIterator(int[] uAdj, int[] vAdj){

			if(uAdj == null || vAdj == null) return Collections.emptyIterator();
			else{
				return new Iterator<Integer>(){

					int uCur = 0;
					int vCur = 0;
					long uD = uAdj.length;
					long vD = vAdj.length;

					int hd = Integer.MIN_VALUE;
					boolean hdDefined = false;

					@Override
					public boolean hasNext() {

						while (!hdDefined && uCur < uD && vCur < vD){
							if (uAdj[uCur] < vAdj[vCur]) uCur += 1;
							else if (vAdj[vCur] < uAdj[uCur]) vCur += 1;
							else {
								hd = uAdj[uCur];
								hdDefined = true;
								uCur += 1;
								vCur += 1;
							}
						}

						return hdDefined;
					}

					@Override
					public Integer next() {
						if(hasNext()){
							hdDefined = false;
							return hd;
						}
						else throw new NoSuchElementException();
					}
				};
			}
		}
	}

	public static class TriangleCountingMapper extends Mapper<Object, Text, Text, NullWritable>{

		int numColors;
		int nodeSize;
		Configuration conf;
		FileSystem fs;
		Path part;
		
		
		long operationCounter = 0;

		/**
		 * setup before execution
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void setup(Context context)
				throws IOException, InterruptedException{
			conf = context.getConfiguration();
			numColors = conf.getInt("numColors", 0);
			nodeSize = context.getConfiguration().getInt("nodeSize", 0);
			fs = FileSystem.get(conf);
			part = new Path(conf.get("part"));
		}

		/**
		 * the map function of triangle enumeration step of PTE
		 * @param key value meaning less
		 * @param value problem
		 * @param context of hadoop
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		@Override
		protected void map(Object key, Text value,
				Context context)
				throws IOException, InterruptedException{
			
			int[] problem = getProblem(value);
			System.out.println(Arrays.toString(problem) + "-subproblem");
			long count = 0;
			operationCounter = 0;
			
			if(problem.length == 2){
				//(i,j)-subproblem
				
				System.out.printf("free memory: %.2fMB\n", availableMemoryBytes(true) / 1000000.0);

				int i = problem[0];
				int j = problem[1];

				int[][] ii = getAdj(part.suffix("/edgeset-" + (i * numColors + i)));
				int[][] ij = getAdj(part.suffix("/edgeset-" + (i * numColors + j)));
				int[][] ji = getAdj(part.suffix("/edgeset-" + (j * numColors + i)));
				int[][] jj = getAdj(part.suffix("/edgeset-" + (j * numColors + j)));
				
				if(i + 1 == j){
					count += enumTri(ii, ii, ii, i, i, i);
				}
				
				if(problem[1] == numColors-1 && problem[0] == 0){
					count += enumTri(jj, jj, jj, j, j, j);
				}
				
				count += enumTri(ii,ij,ij, i, i, j);
				count += enumTri(ij,ii,ji, i, j, i);
				count += enumTri(ij,ij,jj, i, j, j);
				count += enumTri(ji,ji,ii, j, i, i);
				count += enumTri(ji,jj,ij, j, i, j);
				count += enumTri(jj,ji,ji, j, j, i);
				
				
			}
			else{
				//(i,j,k)-subproblem
				
				System.out.printf("free memory: %.2fMB\n", availableMemoryBytes(true) / 1000000.0);
				
				int i = problem[0];
				int j = problem[1];
				int k = problem[2];
				
				int[][] ij = getAdj(part.suffix("/edgeset-" + (problem[0] * numColors + problem[1])));
				int[][] ik = getAdj(part.suffix("/edgeset-" + (problem[0] * numColors + problem[2])));
				int[][] jk = getAdj(part.suffix("/edgeset-" + (problem[1] * numColors + problem[2])));
				
				count += enumTri(ij,ik,jk, i, j, k);
				
				int[][] kj = getAdj(part.suffix("/edgeset-" + (problem[2] * numColors + problem[1])));
				
				count += enumTri(ik,ij,kj, i, k, j);
				
				int[][] ji = getAdj(part.suffix("/edgeset-" + (problem[1] * numColors + problem[0])));
				
				count += enumTri(ji,jk,ik, j, i, k);
				
				ik = null;
				System.gc();
				
				int[][] ki = getAdj(part.suffix("/edgeset-" + (problem[2] * numColors + problem[0])));
				
				count += enumTri(jk,ji,ki, j, k, i);
				
				jk = null;
				System.gc();
				
				count += enumTri(ki,kj,ij, k, i, j);
				
				ij = null;
				System.gc();
				
				count += enumTri(kj,ki,ji, k, j, i);
				
				ji = null;
				kj = null;
				ki = null;
				System.gc();
				
			}
			
			context.getCounter("PTE", "triangles").increment(count);
			
		}

		/**
		 * parse a problem
		 * @param line the text representing a problem to parse
		 * @return the problem as an int array
		 */
		private int[] getProblem(Text line){
			StringTokenizer st = new StringTokenizer(line.toString());

			TIntArrayList _problem = new TIntArrayList();

			try{
				while(true){
					_problem.add(Integer.parseInt(st.nextToken()));
				}
			}catch(Exception e){}

			int[] problem = _problem.toArray();

			return problem;
		}

		/**
		 * load an edge set of an adjacency list format
		 * @param graphPath the path of a graph
		 * @return the adjacency list of an edge set
		 * @throws IOException of hadoop
		 */
		private int[][] getAdj(Path graphPath) throws IOException{
			long t = System.currentTimeMillis();

			int[][] adj = new int[nodeSize][];

//			DataInputStream in = new DataInputStream(new SnappyInputStream(fs.open(graphPath).getWrappedStream()));
			
			Input in = new Input(new SnappyInputStream(fs.open(graphPath).getWrappedStream()));
			
			try{
				while(true){
					int u = in.readInt(true);
					int uSize = in.readInt(true);
					
					adj[u] = new int[uSize];
					int[] uAdj = adj[u];
					
					for(int i=0; i<uSize; i++){
						uAdj[i] = in.readInt(true);
					}
				}
			}catch(KryoException e){}
			
			in.close();
			
			System.out.printf("loading[%s]: %.3fs (free memory: %.2fMB)\n",
					graphPath.toString(),
					(System.currentTimeMillis() - t) / 1000.0,
					availableMemoryBytes(true) / 1000000.0);
			
			return adj;
		}

		/**
		 * Counting triangles.
		 * It counts the number of all triangles (u, v, w) that the colors of the vertices
		 * u, v, and w are i, j, and k, respectively.
		 * @param ij adjacency list of an edge set
		 * @param ik adjacency list of an edge set
		 * @param jk adjacency list of an edge set
		 * @param i the first node color
		 * @param j the second node color
		 * @param k the last node color
		 * @return the number of triangles of colors i, j, and k.
		 * @throws IOException by hadoop
		 * @throws InterruptedException by hadoop
		 */
		protected long enumTri(int[][] ij, int[][] ik, int[][] jk, int i, int j, int k) throws IOException, InterruptedException {
			
			long t = System.currentTimeMillis();
			
			long count = 0;
			
			for(int u = 0; u < nodeSize; u++){
				if(ij[u] != null){
					for(int v : ij[u]){
							count += countIntersect(ik[u], jk[v]);
					}
				}
			}
			
			System.out.println("counting: " + ((System.currentTimeMillis() - t)/1000.0) + "s");
			
			return count;
			
		}

		/**
		 * count the number of intersection between two arrays
		 * @param uAdj a sorted array
		 * @param vAdj another array
		 * @return the number of intersection between the two arrays
		 */
		long countIntersect(int[] uAdj, int[] vAdj){

			if(uAdj == null || vAdj == null){
				return 0;
			}

			long count = 0;
			int uCur = 0;
			int vCur = 0;
			long uD = uAdj.length;
			long vD = vAdj.length;

			while(uCur < uD && vCur < vD){
				if(uAdj[uCur] < vAdj[vCur]) uCur++;
				else if(vAdj[vCur] < uAdj[uCur]) vCur++;
				else{
					count++;
					uCur++;
					vCur++;
				}
			}

			return count;
		}


		/**
		 * calculate available memory
		 * @param gc do garbage collect if true
		 * @return the size of available memory
		 */
		public static long availableMemoryBytes(boolean gc){
			
			if(gc){
				System.gc(); System.gc();
			}
			
			return Runtime.getRuntime().maxMemory()
			- (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory());
		}
		
	}
	
}
