/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PACCBasic.scala
  * - Find Connected components.
  * - This version uses 'filter' to separate out edges and filtered edges.
  * Version: 3.0
  */

package pegasus.spark.cc

import java.util.StringTokenizer

import org.apache.log4j.{Level, Logger}
import org.apache.spark.SparkContext
import org.apache.spark.rdd.RDD

class PACCBasic extends PACC{

  override val APP_NAME: String = "pacc-basic"
  override val EXEC_NAME: String = "paccbasic"

  /**
    * submit the spark job.
    *
    * @param inputPath input file path.
    * @param localThreshold if the number of remaining edges are lower than this value,
    *                      pacc run a single machine algorithm (LocalCC).
    * @param numPartitions the number of partitions
    * @param sc spark context.
    * @return an RDD containing connected components
    */
  override def run(inputPath: String, localThreshold: Int,
                   numPartitions: Int, sc: SparkContext): RDD[(Long, Long)] = {

    //initialize
    var out = sc.textFile(inputPath).map{ line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong
      (u, v)
    }

    var numEdges = out.count()

    var converge = false

    var others: RDD[(Long, Long)] = sc.emptyRDD[(Long, Long)]

    do{
      // partitioning step
      if(numEdges > localThreshold) {
        val (lout, lothers, l_change, lout_size, lcc_size, lin_size) = largeStar(out, numPartitions)

        others ++= lothers

        val (sout, sin, s_change, sout_size, sin_size) = smallStar(lout, numPartitions)
        out = sout
        others ++= sin

        PACCBasic.logger.info(f"lout: $lout_size, lcc: $lcc_size, lin: $lin_size sout: $sout_size, sin: $sin_size")

        converge = l_change == 0 && s_change == 0
        numEdges = sout_size

      }
      else{
        // LocalCC
        out = UnionFind.run(out.map{ case (u, v) => (if (u < 0) ~u else u, v) })
        converge = true
      }
    }while(!converge)

    // computation step
    ccComputation(out ++ others, numPartitions)
  }

  /**
    * PA-Large-Star Operation.
    * For each node n, this operation links each large neighbor v to the minimum node mcu(p(v))
    * in the same partition p(v) that contains the neighbor v.
    *
    * @param inputRDD the input rdd
    * @param numPartitions the number of partitions
    * @return (RDD of out edges, RDD of filtered edges, # changed edges, # of 'out' edges,
    *         # filtered 'cc' edges, # filtered 'in' edges)
    */
  def largeStar(inputRDD: RDD[(Long, Long)], numPartitions: Int): (RDD[(Long, Long)],
    RDD[(Long, Long)], Long, Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val LCC_SIZE = sc.longAccumulator
    val LIN_SIZE = sc.longAccumulator
    val LOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) =>
        if (u < 0) Seq((v, u))
        else Seq((u, v), (v, u))
    }.groupByKey()

    val LOTHERS = true
    val LOUT = false

    val res_all = groupedRDD.flatMap {
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        mcu((u % numPartitions).toInt) = u

        var isStar = true

        val uN_large = uN.filter { v_raw =>
          val v = if (v_raw >= 0) {
            isStar = false
            v_raw
          } else ~v_raw

          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v > u
        }.toSeq.distinct.toArray

        val mu = mcu.min

        if (isStar) uN_large.map { v_raw => LCC_SIZE.add(1); ((LOTHERS, ~v_raw), u) }
        else{
          uN_large.map{ v_raw =>
            val vIsLeaf = v_raw < 0
            val v: Long = if(vIsLeaf) ~v_raw else v_raw

            val vp = (v % numPartitions).toInt
            val mcu_vp = mcu(vp)

            if(v != mcu_vp) {

              if(mcu_vp != u) NUM_CHANGES.add(1)

              if(vIsLeaf){
                LIN_SIZE.add(1)
                ((LOTHERS, v), mcu_vp)
              }
              else{
                LOUT_SIZE.add(1)
                ((LOUT, v), mcu_vp)
              }
            }
            else{// v is a local minimum but not the global minimum because 'it' has only large neighbors.
              LOUT_SIZE.add(1)
              ((LOUT, v), mu)
            }
          }
        }
    }

    val lother = res_all.filter(_._1._1 == LOTHERS).map{case ((_, u), m) => (u, m)}
    val lout = res_all.filter(_._1._1 == LOUT).map{case ((_, u), m) => (u, m)}

    res_all.count()

    (lout, lother, NUM_CHANGES.value, LOUT_SIZE.value, LCC_SIZE.value, LIN_SIZE.value)

  }


  /**
    * PA-Small-Star Operation.
    * For each node n, this operation links each small neighbor v to the minimum node mcu(p(v))
    * in the same partition p(v) that contains the neighbor v.
    *
    * @param inputRDD the input rdd
    * @param numPartitions the number of partitions
    * @return (RDD of out edges, RDD of filtered edges, # changed edges, # of 'out' edges,
    *         # filtered 'in' edges)
    */
  def smallStar(inputRDD: RDD[(Long, Long)], numPartitions: Int): (RDD[(Long, Long)],
    RDD[(Long, Long)], Long, Long, Long) = {

    val sc = inputRDD.sparkContext

    val NUM_CHANGES = sc.longAccumulator
    val SIN_SIZE = sc.longAccumulator
    val SOUT_SIZE = sc.longAccumulator

    val groupedRDD = inputRDD.flatMap{
      case (u, v) => Seq((u, v), (v, u))
    }.groupByKey()

    val SIN = true
    val SOUT = false

    val res_all = groupedRDD.flatMap{
      case (u, uN) =>

        val mcu = Array.fill[Long](numPartitions)(Long.MaxValue)
        val up = (u % numPartitions).toInt
        mcu(up) = u

        var isLeaf = true

        val uN_small = uN.filter { v =>

          if(v > u) isLeaf = false
          val vp = (v % numPartitions).toInt
          mcu(vp) = Math.min(v, mcu(vp))

          v < u
        }.toArray.toStream

        val mu = mcu.min

        val sout = uN_small filter {_ != mu} map { v =>

          val vp = (v % numPartitions).toInt

          NUM_CHANGES.add(1)
          SOUT_SIZE.add(1)

          if (v != mcu(vp))
            ((SOUT, v), mcu(vp))
          else // v is mcu but not mu
            ((SOUT, v), mu)

        }

        if (u != mcu(up)) {
          if (isLeaf) {
            SIN_SIZE.add(1)
            sout ++ Stream(((SIN, u), mcu(up)))
          }
          else {
            SOUT_SIZE.add(1)
            sout ++ Stream(((SOUT, u), mcu(up)))
          }
        }
        else if (u != mu) {
          SOUT_SIZE.add(1)
          sout ++ Stream(((SOUT, if (isLeaf) ~u else u), mu))
        }
        else sout
    }

    val sin = res_all.filter(_._1._1 == SIN).map{case ((_, u), m) => (u, m)}
    val sout = res_all.filter(_._1._1 == SOUT).map{case ((_, u), m) => (u, m)}

    res_all.count()

    (sout, sin, NUM_CHANGES.value, SOUT_SIZE.value, SIN_SIZE.value)

  }
}

object PACCBasic{
  private val logger = Logger.getLogger(getClass)
  logger.setLevel(Level.INFO)

  def main(args: Array[String]): Unit = new PACCBasic().runByCLI(args)
}
