/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: MatrixLoader.scala
  * - Loads matrix from HDFS
  * Version: 3.0
  */
package pegasus.spark.common

import java.io.{BufferedReader, InputStreamReader}
import java.util.function.IntFunction

import org.apache.hadoop.fs.Path
import org.apache.spark.SparkContext
import org.apache.spark.rdd.RDD
import org.apache.spark.storage.StorageLevel
import pegasus.spark.common.io.{UnweightedMatrixReader, WeightedMatrixReader}
import pegasus.spark.common.partitioner.TaskPartitioner
import pegasus.spark.common.types.{MatrixBlock, UnweightedMatrixBlock, WeightedMatrixBlock}

object MatrixLoader {
  private def parseStatsLine(line: String): (Int, Int) = {
    val token = line.split("=")
    (token(0).substring(token(0).length - 5).toInt, token(1).toInt)
  }

  def loadSubproblems(sc: SparkContext, path: String): (Long, RDD[(Int, Seq[SubProblem])]) = {
    val statsPath = new Path(path, Constants.StatsPath)
    val fs = statsPath.getFileSystem(sc.hadoopConfiguration)
    val reader = new BufferedReader(new InputStreamReader(fs.open(statsPath)))
    try {
      val lines = reader.lines().toArray(new IntFunction[Array[String]] {
        override def apply(size: Int): Array[String] = Array.ofDim(size)
      })

      val numVertices = lines.filter(_.startsWith("numVertices")).map(_.split("=")(1).toLong).head
      val numSparseCols = lines.filter(_.startsWith("numSparseCols")).map(parseStatsLine).toMap
      val numSparseRows = lines.filter(_.startsWith("numSparseRows")).map(parseStatsLine).toMap
      val numDenseCols = lines.filter(_.startsWith("numDenseCols")).map(parseStatsLine).toMap
      val numDenseRows = lines.filter(_.startsWith("numDenseRows")).map(parseStatsLine).toMap
      val numBlocks = math.sqrt(numSparseCols.size).toInt

      val problems = (0 until numBlocks).map(taskId => {
        val sparseProblems = (0 until numBlocks).map(subTaskId => {
          val blockId = taskId + subTaskId * numBlocks
          SubProblem(path, isDense = false, blockId, numSparseCols(blockId), numSparseRows(blockId))
        })
        val denseProblems = (0 until numBlocks).map(subTaskId => {
          val blockId = taskId * numBlocks + subTaskId
          SubProblem(path, isDense = true, blockId, numDenseCols(blockId), numDenseRows(blockId))
        })

        (taskId, (sparseProblems ++ denseProblems).asInstanceOf[Seq[SubProblem]])
      })

      (numVertices, sc.parallelize(problems).partitionBy(new TaskPartitioner(numBlocks)).cache())
    } finally {
      reader.close()
    }
  }

  private def readUnweightedMatrixBlock(path: String, blockId: Int, numCols: Int, numRows: Int): MatrixBlock[Int] = {
    val reader = new UnweightedMatrixReader(path)
    val block = new UnweightedMatrixBlock(blockId, numCols, numRows)
    try {
      while (reader.hasNext) {
        val (col, deg, rows) = reader.next()
        block += (col, deg, rows)
      }
      block
    } finally {
      reader.close()
    }
  }

  private def readWeightedMatrixBlock(
    path: String,
    blockId: Int,
    numCols: Int,
    numRows: Int
  ): MatrixBlock[(Int, Double)] = {
    val reader = new WeightedMatrixReader(path)
    val block = new WeightedMatrixBlock(blockId, numCols, numRows)
    try {
      while (reader.hasNext) {
        val (col, deg, rows) = reader.next()
        block += (col, deg, rows)
      }

      block
    } finally {
      reader.close()
    }
  }

  def loadUnweightedMatrix(
    problemRDD: RDD[(Int, Seq[SubProblem])],
    storageLevel: StorageLevel = StorageLevel.MEMORY_AND_DISK
  ): (RDD[(Int, MatrixBlock[Int])], RDD[(Int, MatrixBlock[Int])]) = {
    val sparseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(!_.isDense).map(block => {
            (p._1, readUnweightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)
    val denseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(_.isDense).map(block => {
            (p._1, readUnweightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)

    sparseRDD.foreachPartition(_ => ())
    denseRDD.foreachPartition(_ => ())

    (sparseRDD, denseRDD)
  }

  def loadWeightedMatrix(
    problemRDD: RDD[(Int, Seq[SubProblem])],
    storageLevel: StorageLevel = StorageLevel.MEMORY_AND_DISK
  ): (RDD[(Int, MatrixBlock[(Int, Double)])], RDD[(Int, MatrixBlock[(Int, Double)])]) = {
    val sparseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(!_.isDense).map(block => {
            (p._1, readWeightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)
    val denseRDD = problemRDD
      .mapPartitions(iter => {
        iter.flatMap(p => {
          p._2.filter(_.isDense).map(block => {
            (p._1, readWeightedMatrixBlock(block.matBlockPath, block.matBlockId, block.numCols, block.numRows))
          })
        })
      }, preservesPartitioning = true)
      .persist(storageLevel)

    sparseRDD.foreachPartition(_ => ())
    denseRDD.foreachPartition(_ => ())

    (sparseRDD, denseRDD)
  }
}
