/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: Multiplication.scala
  * - A base implementation for matrix-vector multiplication
  * Version: 3.0
  */
package pegasus.spark.common

import pegasus.spark.common.io.{VectorReader, VectorWriter}
import pegasus.spark.common.types.{MatrixBlock, VectorBlock}

trait Multiplication[M, V, PV] extends Serializable {
  /**
    * Creates a vector block.
    *
    * @param blockId     The block id of vector
    * @param numVertices The number of vertices
    * @param numBlocks   The number of block
    * @param init        Whether to initialize vector element or not
    * @param context     The multiplication context object
    * @return A created vector block
    */
  def createVectorBlock(
    blockId: Int,
    numVertices: Long,
    numBlocks: Int,
    init: Boolean,
    context: MultiplicationContext
  ): VectorBlock[V]

  /**
    * Creates a vector reader.
    *
    * @param vectorPath The path to vector blocks
    * @param blockId    The id of vector block
    * @return A created vector reader
    */
  def createVectorReader(vectorPath: String, blockId: Int): VectorReader[V]

  /**
    * Creates a vector writer.
    *
    * @param vectorPath The path to vector blocks
    * @param blockId    The id of vector block
    * @return A created vector writer
    */
  def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[V]

  /**
    * Creates an intermediate vector block.
    *
    * @param numVertices The number of vertices
    * @param numBlocks   The number of blocks
    * @param blockId     The block id
    * @param context     The multiplication context
    * @return A created intermediate vector block
    */
  def createIntermediateVectorBlock(
    numVertices: Long,
    numBlocks: Int,
    blockId: Int,
    context: MultiplicationContext
  ): VectorBlock[PV]

  /**
    * Creates an intermediate vector reader.
    *
    * @param vectorPath The path to intermediate vector file
    * @param blockId    The block id
    * @return The intermeidate vector reader
    */
  def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[PV]

  /**
    * Creates an intermediate vector writer.
    *
    * @param vectorPath The path to intermediate vector file
    * @param blockId    The block id
    * @return
    */
  def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[PV]

  /**
    * Runs combine2 operation given a matrix block and a vector block.
    *
    * @param matrix  The matrix block
    * @param vector  The vector block
    * @param result  A vector block to contain the result
    * @param context The multiplication context
    */
  def combine2Op(
    matrix: MatrixBlock[M],
    vector: VectorBlock[V],
    result: VectorBlock[PV],
    context: MultiplicationContext
  ): Unit

  /**
    * Runs combineAll operation given partial results.
    *
    * @param input   The partial results
    * @param output  A vector block to contain the result
    * @param context The multiplication context
    */
  def combineAllOp(input: VectorReader[PV], output: VectorBlock[PV], context: MultiplicationContext): Unit

  /**
    * Runs apply operation on the result.
    *
    * @param result       The result from previous iteration
    * @param intermediate The intermediate result
    * @param context      The multiplication context
    */
  def applyOp(result: VectorBlock[V], intermediate: VectorBlock[PV], context: MultiplicationContext): Unit
}
