/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PMVUtils.scala
  * - Provides utility methods for PMV
  * Version: 3.0
  */
package pegasus.spark.common

import java.io.File
import java.nio.ByteBuffer
import java.util.concurrent.ConcurrentHashMap

import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.io.{ByteBufferPool, ElasticByteBufferPool}
import org.apache.spark.deploy.SparkHadoopUtil
import org.apache.spark.io.{CompressionCodec, LZ4CompressionCodec, LZFCompressionCodec, SnappyCompressionCodec}
import org.apache.spark.{SparkConf, SparkEnv}

object PMVUtils {
  private val JavaRuntime = Runtime.getRuntime
  private val byteBufferPool: ByteBufferPool = new ElasticByteBufferPool

  def callGCMultipleTimes(): Unit = callGCMultipleTimes(3)

  def callGCMultipleTimes(times: Int): Unit = {
    var i = 0
    while (i < times) {
      JavaRuntime.gc()
      i += 1
    }
  }

  def maxMemoryBytes: Long = JavaRuntime.maxMemory

  def availableMemoryBytes: Long = JavaRuntime.maxMemory - JavaRuntime.totalMemory + JavaRuntime.freeMemory

  def obtainByteBuffer(size: Int): ByteBuffer = byteBufferPool.getBuffer(true, size)

  def returnByteBuffer(buf: ByteBuffer): Unit = byteBufferPool.putBuffer(buf)

  def localDirs(): Array[String] = {
    val conf = sparkConf()
    if (conf == null) {
      Array(System.getProperty("java.io.tmpdir", "/tmp"))
    } else if (System.getenv("CONTAINER_ID") != null) {
      System.getenv("LOCAL_DIRS").split(",")
    } else if (System.getenv("SPARK_EXECUTOR_DIRS") != null) {
      System.getenv("SPARK_EXECUTOR_DIRS").split(File.pathSeparator)
    } else if (System.getenv("SPARK_LOCAL_DIRS") != null) {
      System.getenv("SPARK_LOCAL_DIRS").split(",")
    } else {
      conf.get("spark.local.dir", System.getProperty("java.io.tmpdir")).split(",")
    }
  }

  def sparkConf(): SparkConf = Option(SparkEnv.get).map(_.conf).getOrElse(new SparkConf())

  def hadoopConf(): Configuration = SparkHadoopUtil.get.conf

  private val codecPool = new ConcurrentHashMap[String, CompressionCodec]()

  def compressionCodec(): CompressionCodec = {
    val conf = sparkConf()
    val codecName = conf.get("spark.io.compression.codec", "lz4")
    Option(codecPool.get(codecName)) match {
      case Some(codec) => codec
      case None =>
        val codec = codecName match {
          case "lz4" => new LZ4CompressionCodec(conf)
          case "lzf" => new LZFCompressionCodec(conf)
          case "snappy" => new SnappyCompressionCodec(conf)
        }
        codecPool.put(codecName, codec)

        codec
    }
  }

  def globalToLocal(numBlocks: Int, globalId: Long): Int = (globalId / numBlocks).toInt

  def localToGlobal(numBlocks: Int, blockId: Int, localId: Int): Long = localId.toLong * numBlocks + blockId
}
