/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: Prepartitioning.scala
  * - Partitions matrix into multiple blocks
  * Version: 3.0
  */
package pegasus.spark.common

import java.io.{BufferedWriter, OutputStreamWriter}

import it.unimi.dsi.fastutil.doubles.DoubleArrayList
import it.unimi.dsi.fastutil.ints.IntArrayList
import org.apache.hadoop.fs.Path
import org.apache.spark.rdd.RDD
import pegasus.spark.common.io._
import pegasus.spark.common.partitioner.VertexPartitioner

object Prepartitioning {
  def unweightedMatrix(
    edges: RDD[(Long, Long)],
    numBlocks: Int,
    degreeThreshold: Int,
    rootPath: String
  ): Unit = {
    val sc = edges.sparkContext
    val numCores = sc.getConf.getInt("spark.executor.cores", 1)

    val stats = edges
      .partitionBy(new VertexPartitioner(numBlocks))
      .mapPartitionsWithIndex((taskId, iter) => {
        var maxVertex = Long.MinValue
        val degreeThresholdLocal = degreeThreshold
        val numBlocksLocal = numBlocks
        val grouper = new LongExternalGrouper(taskId, numCores)
        iter.foreach(e => {
          val srcLocal = PMVUtils.globalToLocal(numBlocksLocal, e._1)
          grouper.append(srcLocal, e._2)
          maxVertex = math.max(maxVertex, math.max(e._1, e._2))
        })

        val numColsSparse = Array.ofDim[Int](numBlocksLocal)
        val numRowsSparse = Array.ofDim[Int](numBlocksLocal)
        val numColsDense = Array.ofDim[Int](numBlocksLocal)
        val numRowsDense = Array.ofDim[Int](numBlocksLocal)

        val sout = Array.tabulate(numBlocksLocal)(i =>
          new UnweightedMatrixWriter(f"$rootPath/${Constants.SparseMatrixBlock}-${i * numBlocksLocal + taskId}%05d"))
        val dout = Array.tabulate(numBlocksLocal)(i =>
          new UnweightedMatrixWriter(f"$rootPath/${Constants.DenseMatrixBlock}-${i * numBlocksLocal + taskId}%05d"))
        val dvOut = new UnitVectorWriter(f"$rootPath/${Constants.DenseVectorBlock}-$taskId%05d")

        val blocks = Array.fill(numBlocksLocal)(new IntArrayList())
        grouper.groupedIterator().foreach(v => {
          val (col, deg, rows) = v
          blocks.foreach(_.clear())
          rows.foreach(row => {
            blocks((row % numBlocksLocal).toInt).add(PMVUtils.globalToLocal(numBlocksLocal, row))
          })
          val (numCols, numRows, out) = if (deg < degreeThresholdLocal) {
            (numColsSparse, numRowsSparse, sout)
          } else {
            dvOut.write(col, ())
            (numColsDense, numRowsDense, dout)
          }
          blocks.iterator.zipWithIndex.foreach(b => {
            val (block, idx) = b
            if (block.size() > 0) {
              out(idx).write(col, deg, block.toIntArray)
              numCols(idx) += 1
              numRows(idx) += block.size()
            }
          })
        })

        sout.foreach(_.close())
        dout.foreach(_.close())
        dvOut.close()

        (0 until numBlocksLocal).map(i => {
          val blockId = i * numBlocksLocal + taskId
          (blockId, maxVertex, numColsSparse(i), numRowsSparse(i), numColsDense(i), numRowsDense(i))
        }).toIterator
      })
      .collect()

    val numVertices = stats.iterator.map(_._2).max + 1
    val statsPath = new Path(rootPath, Constants.StatsPath)
    val fs = statsPath.getFileSystem(sc.hadoopConfiguration)
    val writer = new BufferedWriter(new OutputStreamWriter(fs.create(statsPath)))
    try {
      writer.write(s"numVertices=$numVertices\n")
      stats.iterator.map(v => f"numSparseCols${v._1}%05d=${v._3}\n").foreach(writer.write)
      stats.iterator.map(v => f"numSparseRows${v._1}%05d=${v._4}\n").foreach(writer.write)
      stats.iterator.map(v => f"numDenseCols${v._1}%05d=${v._5}\n").foreach(writer.write)
      stats.iterator.map(v => f"numDenseRows${v._1}%05d=${v._6}\n").foreach(writer.write)
    } finally {
      writer.close()
    }
  }

  def weightedMatrix(
    edges: RDD[(Long, Long, Double)],
    numBlocks: Int,
    degreeThreshold: Int,
    rootPath: String
  ): Unit = {
    val sc = edges.sparkContext
    val numCores = sc.getConf.getInt("spark.executor.cores", 1)

    val stats = edges.map(e => (e._1, (e._2, e._3)))
      .partitionBy(new VertexPartitioner(numBlocks))
      .mapPartitionsWithIndex((taskId, iter) => {
        var maxVertex = Long.MinValue
        val degreeThresholdLocal = degreeThreshold
        val numBlocksLocal = numBlocks
        val grouper = new LongDoubleExternalGrouper(taskId, numCores)
        iter.foreach(e => {
          val srcLocal = PMVUtils.globalToLocal(numBlocksLocal, e._1)
          grouper.append(srcLocal, e._2)
          maxVertex = math.max(maxVertex, math.max(e._1, e._2._1))
        })

        val numColsSparse = Array.ofDim[Int](numBlocksLocal)
        val numRowsSparse = Array.ofDim[Int](numBlocksLocal)
        val numColsDense = Array.ofDim[Int](numBlocksLocal)
        val numRowsDense = Array.ofDim[Int](numBlocksLocal)

        val sout = Array.tabulate(numBlocksLocal)(i =>
          new WeightedMatrixWriter(f"$rootPath/${Constants.SparseMatrixBlock}-${i * numBlocksLocal + taskId}%05d"))
        val dout = Array.tabulate(numBlocksLocal)(i =>
          new WeightedMatrixWriter(f"$rootPath/${Constants.DenseMatrixBlock}-${i * numBlocksLocal + taskId}%05d"))
        val dvOut = new UnitVectorWriter(f"$rootPath/${Constants.DenseVectorBlock}-$taskId%05d")

        val rowBlocks = Array.fill(numBlocksLocal)(new IntArrayList())
        val weightBlocks = Array.fill(numBlocksLocal)(new DoubleArrayList())
        grouper.groupedIterator().foreach(v => {
          val (col, deg, rows) = v
          rowBlocks.foreach(_.clear())
          rows.foreach(row => {
            val blockId = (row._1 % numBlocksLocal).toInt
            rowBlocks(blockId).add(PMVUtils.globalToLocal(numBlocksLocal, row._1))
            weightBlocks(blockId).add(row._2)
          })
          val (numCols, numRows, out) = if (deg < degreeThresholdLocal) {
            (numColsSparse, numRowsSparse, sout)
          } else {
            dvOut.write(col, ())
            (numColsDense, numRowsDense, dout)
          }
          rowBlocks.iterator.zipWithIndex.foreach(b => {
            val (block, idx) = b
            if (block.size() > 0) {
              val wBlock = weightBlocks(idx)
              out(idx).write(col, deg, block.toIntArray().zip(wBlock.toDoubleArray))
              numCols(idx) += 1
              numRows(idx) += block.size()
            }
          })
        })

        sout.foreach(_.close())
        dout.foreach(_.close())
        dvOut.close()

        (0 until numBlocksLocal).map(i => {
          val blockId = i * numBlocksLocal + taskId
          (blockId, maxVertex, numColsSparse(i), numRowsSparse(i), numColsDense(i), numRowsDense(i))
        }).toIterator
      }).collect()

    val numVertices = stats.iterator.map(_._2).max + 1
    val statsPath = new Path(rootPath, Constants.StatsPath)
    val fs = statsPath.getFileSystem(sc.hadoopConfiguration)
    val writer = new BufferedWriter(new OutputStreamWriter(fs.create(statsPath)))
    try {
      writer.write(s"numVertices=$numVertices\n")
      stats.iterator.map(v => f"numSparseCols${v._1}%05d=${v._3}\n").foreach(writer.write)
      stats.iterator.map(v => f"numSparseRows${v._1}%05d=${v._4}\n").foreach(writer.write)
      stats.iterator.map(v => f"numDenseCols${v._1}%05d=${v._5}\n").foreach(writer.write)
      stats.iterator.map(v => f"numDenseRows${v._1}%05d=${v._6}\n").foreach(writer.write)
    } finally {
      writer.close()
    }
  }
}
