/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  * * Redistributions of source code must retain the above copyright
  * notice, this list of conditions and the following disclaimer.
  * * Redistributions in binary form must reproduce the above copyright
  * notice, this list of conditions and the following disclaimer in the
  * documentation and/or other materials provided with the distribution.
  * * Neither the name of the Seoul National University nor the
  * names of its contributors may be used to endorse or promote products
  * derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: ExternalGrouper.scala
  * - Groups (key, value) records using external-sort algorithm.
  * Version: 3.0
  */
package pegasus.spark.common.io

import it.unimi.dsi.fastutil.ints.AbstractIntComparator
import it.unimi.dsi.fastutil.{Arrays, Swapper}
import pegasus.spark.common.PMVUtils

import scala.collection.mutable
import scala.reflect.ClassTag
import scala.reflect.io.File

abstract class ExternalGrouper[@specialized(Int, Long, Double) T: ClassTag](
  taskId: Int, sizePerEntry: Int, numCores: Int
) {
  protected var bufferSize = 0
  protected var current = 0

  // we use hard-corded lazy initialization for values to avoid duplicated initialization.
  // this is a bug of scala compiler. If the bug is fixed, we can change these fields to `protected val`.
  protected var keys: Array[Int] = _
  protected var values: Array[T] = _

  /**
    * Initializes grouper. Note that this method is only for lazy initialization.
    *
    * @param specializedValue
    */
  protected def init(specializedValue: T): Unit = {
    PMVUtils.callGCMultipleTimes()
    bufferSize = (PMVUtils.availableMemoryBytes / (sizePerEntry + 8) / numCores * 0.75).toInt
    keys = Array.ofDim[Int](bufferSize)
    values = Array.ofDim[T](bufferSize)
  }

  private var currentFile = 0
  private val files = mutable.Queue[Int]()
  private val tempRoot = {
    val localDirs = PMVUtils.localDirs()
    localDirs(taskId % localDirs.length)
  }

  private val comparator = new AbstractIntComparator {
    override def compare(left: Int, right: Int): Int = {
      if (keys(left) < keys(right)) -1
      else if (keys(left) > keys(right)) 1
      else 0
    }
  }
  protected val swapper = new Swapper {
    override def swap(left: Int, right: Int): Unit = {
      val tempKey = keys(left)
      keys(left) = keys(right)
      keys(right) = tempKey

      // we use an external method for swapping values to avoid unnecessary boxing and unboxing.
      updateValue(left, values(right), right, values(left))
    }
  }

  /**
    * Updates values.
    *
    * @param left       The left index
    * @param leftValue  The left value
    * @param right      The right index
    * @param rightValue The right value
    */
  private def updateValue(left: Int, leftValue: T, right: Int, rightValue: T): Unit = {
    values(left) = leftValue
    values(right) = rightValue
  }

  /**
    * Returns a typed file reader.
    *
    * @param path The path to file
    * @return A reader object
    */
  protected def buildValueReader(path: String): FileReader[T]

  /**
    * Returns a typed file writer.
    *
    * @param path The path to file
    * @return A writer object
    */
  protected def buildValueWriter(path: String): FileWriter[T]

  /**
    * Appends a (key, value) pair.
    *
    * @param key   The key
    * @param value The value
    */
  def append(key: Int, value: T): Unit = {
    if (current == bufferSize) {
      flush()
    }

    keys(current) = key
    values(current) = value
    current += 1
  }

  /**
    * Flushes the values stored in memory cache to a file.
    */
  def flush(): Unit = {
    if (current > 0) {
      // sort indices
      Arrays.quickSort(0, current, comparator, swapper)

      // open files
      val keyWriter = new IntFileWriter(keyFilePath(currentFile))
      val countWriter = new IntFileWriter(countFilePath(currentFile))
      val valueWriter = buildValueWriter(valueFilePath(currentFile))

      // spill to disk
      var i = 0
      var count = 0
      var key = keys(i)
      while (i < current) {
        if (key == keys(i)) {
          count += 1
        } else {
          keyWriter.write(key)
          countWriter.write(count)
          key = keys(i)
          count = 1
        }
        valueWriter.write(values(i))
        i += 1
      }
      keyWriter.write(key)
      countWriter.write(count)

      // close files
      keyWriter.close()
      countWriter.close()
      valueWriter.close()

      files += currentFile
      current = 0
      currentFile += 1
    }
  }

  /**
    * Returns an iterator that groups all the values by key.
    *
    * @return An iterator object
    */
  def groupedIterator(): Iterator[(Int, Int, Array[T])] = {
    if (files.isEmpty) {
      // no spill files
      Arrays.quickSort(0, current, comparator, swapper)
      new MemoryGroupedIterator[T](keys, values, current)
    } else {
      flush()

      while (files.size > 1) {
        val bucket1 = files.dequeue()
        val bucket2 = files.dequeue()
        val newBucket = currentFile
        files += newBucket
        currentFile += 1

        merge(bucket1, bucket2, newBucket)

        File(keyFilePath(bucket1)).delete()
        File(countFilePath(bucket1)).delete()
        File(valueFilePath(bucket1)).delete()
        File(keyFilePath(bucket2)).delete()
        File(countFilePath(bucket2)).delete()
        File(valueFilePath(bucket2)).delete()
      }

      val lastBucket = files.dequeue()
      new FileGroupedIterator[T](
        keyFilePath(lastBucket), countFilePath(lastBucket), buildValueReader(valueFilePath(lastBucket)))
    }
  }

  /**
    * Copies a file to another file.
    *
    * @param keyReader   The key file reader
    * @param countReader The count file reader
    * @param valueReader The value file reader
    * @param keyWriter   The key file writer
    * @param countWriter The count file writer
    * @param valueWriter The value file writer
    */
  private def copy(
    keyReader: BufferedReader[Int], countReader: FileReader[Int], valueReader: FileReader[T],
    keyWriter: FileWriter[Int], countWriter: FileWriter[Int], valueWriter: FileWriter[T]) = {
    // write key
    keyWriter.write(keyReader.next())

    // write count
    val count = countReader.next()
    countWriter.write(count)

    // write values
    var i = 0
    while (i < count) {
      valueWriter.write(valueReader.next())
      i += 1
    }
  }

  /**
    * Merges two buckets into a bucket.
    *
    * @param bucket1   The first bucket id
    * @param bucket2   The second bucket id
    * @param outBucket The merged bucket id
    */
  private def merge(bucket1: Int, bucket2: Int, outBucket: Int) = {
    val key1Reader = new BufferedReader[Int](new IntFileReader(keyFilePath(bucket1)))
    val count1Reader = new IntFileReader(countFilePath(bucket1))
    val value1Reader = buildValueReader(valueFilePath(bucket1))

    val key2Reader = new BufferedReader[Int](new IntFileReader(keyFilePath(bucket2)))
    val count2Reader = new IntFileReader(countFilePath(bucket2))
    val value2Reader = buildValueReader(valueFilePath(bucket2))

    val keyWriter = new IntFileWriter(keyFilePath(outBucket))
    val countWriter = new IntFileWriter(countFilePath(outBucket))
    val valueWriter = buildValueWriter(valueFilePath(outBucket))

    while (key1Reader.hasNext && key2Reader.hasNext) {
      val key1 = key1Reader.head
      val key2 = key2Reader.head

      if (key1 < key2) {
        copy(key1Reader, count1Reader, value1Reader, keyWriter, countWriter, valueWriter)
      } else if (key1 > key2) {
        copy(key2Reader, count2Reader, value2Reader, keyWriter, countWriter, valueWriter)
      } else {
        // merge two keys
        keyWriter.write(key1Reader.next())
        key2Reader.next()

        val count1 = count1Reader.next()
        val count2 = count2Reader.next()
        countWriter.write(count1 + count2)

        var i = 0
        while (i < count1) {
          valueWriter.write(value1Reader.next())
          i += 1
        }
        i = 0
        while (i < count2) {
          valueWriter.write(value2Reader.next())
          i += 1
        }
      }
    }

    // copy remaining records
    while (key1Reader.hasNext) {
      copy(key1Reader, count1Reader, value1Reader, keyWriter, countWriter, valueWriter)
    }
    while (key2Reader.hasNext) {
      copy(key2Reader, count2Reader, value2Reader, keyWriter, countWriter, valueWriter)
    }

    // close readers
    key1Reader.close()
    count1Reader.close()
    value1Reader.close()
    key2Reader.close()
    count2Reader.close()
    value2Reader.close()

    // close writers
    keyWriter.close()
    countWriter.close()
    valueWriter.close()
  }

  private def keyFilePath(id: Int) = (File(tempRoot) / f"pmv-group-key-$taskId-$id%05d").toString()

  private def countFilePath(id: Int) = (File(tempRoot) / f"pmv-group-count-$taskId-$id%05d").toString()

  private def valueFilePath(id: Int) = (File(tempRoot) / f"pmv-group-value-$taskId-$id%05d").toString()
}
