/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: BitStrings.scala
  * - A bitstring for probablistic set
  * Version: 3.0
  */
package pegasus.spark.common.types

import pegasus.spark.common.Constants

import scala.annotation.tailrec

class BitStrings extends Serializable {
  private val bits = Array.ofDim[Long](Constants.DefaultNumBitstrings)

  def setBits(bits: Array[Long]): Unit = {
    System.arraycopy(bits, 0, this.bits, 0, Constants.DefaultNumBitstrings)
  }

  def getBits: Array[Long] = bits

  def getBit(i: Int) = bits(i)

  def setBit(i: Int, bit: Long): Unit = {
    bits(i) = bit
  }

  def or(other: BitStrings): Unit = {
    var i = 0
    while (i < Constants.DefaultNumBitstrings) {
      bits(i) |= other.bits(i)
      i += 1
    }
  }

  def isEmpty: Boolean = {
    @tailrec
    def recur(i: Int): Boolean = {
      if (i == Constants.DefaultNumBitstrings) true
      else if (bits(i) != 0) false
      else recur(i + 1)
    }

    recur(0)
  }

  def cardinality(): Double = {
    @tailrec
    def zeroBitpos(bit: Long, i: Int): Int = {
      if (i == 64) 64
      else if ((bit & (1L << (63 - i))) == 0) i
      else zeroBitpos(bit, i + 1)
    }

    val bitpos = bits.map(b => zeroBitpos(b, 0)).sum / Constants.DefaultNumBitstrings
    math.pow(2, bitpos) / Constants.Phi
  }


  def canEqual(other: Any): Boolean = other.isInstanceOf[BitStrings]

  override def equals(other: Any): Boolean = other match {
    case that: BitStrings =>
      (that canEqual this) &&
        (bits sameElements that.bits)
    case _ => false
  }

  override def hashCode(): Int = {
    val state = Seq(bits)
    state.map(_.hashCode()).foldLeft(0)((a, b) => 31 * a + b)
  }
  
  override def toString = s"BitStrings(${bits.map(_.toString).mkString(" ")})"
}
