/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: BitStringsVectorBlock.scala
  * - A vector block that contains bitstrings
  * Version: 3.0
  */
package pegasus.spark.common.types

import java.util

import pegasus.spark.common.{Constants, PMVUtils}

import scala.annotation.tailrec

class BitStringsVectorBlock(numBlocks: Int, numVertices: Long, _blockId: Int)
  extends VectorBlock[BitStrings](numBlocks, numVertices, _blockId) {
  private var values: Array[Array[Long]] = Array.fill(Constants.DefaultNumBitstrings)(Array.ofDim[Long](size))
  private var zeroValue = Array.ofDim[Long](Constants.DefaultNumBitstrings)
  private val value = new BitStrings
  clear()

  override def get(i: Int): BitStrings = {
    var k = 0
    while (k < Constants.DefaultNumBitstrings) {
      value.setBit(k, values(k)(i))
      k += 1
    }
    value
  }

  override def put(i: Int, v: BitStrings): Unit = {
    mark(i)

    var k = 0
    while (k < Constants.DefaultNumBitstrings) {
      values(k)(i) = v.getBit(k)
      k += 1
    }
  }

  override def addTo(i: Int, v: BitStrings): Unit = {
    mark(i)

    var k = 0
    while (k < Constants.DefaultNumBitstrings) {
      values(k)(i) |= v.getBit(k)
      k += 1
    }
  }

  override def clear(): Unit = {
    clearContainsSet()
    var k = 0
    while (k < Constants.DefaultNumBitstrings) {
      util.Arrays.fill(values(k), zeroValue(k))
      k += 1
    }
  }

  override def iterator(onlyNonzero: Boolean) = new VectorIterator[BitStrings] {
    private var currPos = 0
    private var currId = 0
    private val currVal = new BitStrings()

    override final def next(): Boolean = {
      var found = false

      while (!found && currPos < size && PMVUtils.localToGlobal(numBlocks, blockId, currPos) < numVertices) {
        if (!onlyNonzero || !isEmpty) {
          found = true
          currId = currPos

          var k = 0
          while (k < Constants.DefaultNumBitstrings) {
            currVal.setBit(k, values(k)(currPos))
            k += 1
          }
        }
        currPos += 1
      }

      found
    }

    override final def currentId: Int = currId

    override final def currentValue: BitStrings = currVal

    private def isEmpty: Boolean = {
      @tailrec
      def recur(k: Int): Boolean = {
        if (k == Constants.DefaultNumBitstrings) true
        else if (values(k)(currPos) != 0) false
        else recur(k + 1)
      }

      recur(0)
    }
  }

  override def iterator(onlyNonZero: Boolean, filterSet: BitSet) = new VectorIterator[BitStrings] {
    private val iter = iterator(onlyNonZero)

    override final def next(): Boolean = {
      var found = false

      while (!found && iter.next()) {
        found = filterSet.contains(iter.currentId)
      }

      found
    }

    override final def currentId: Int = iter.currentId

    override final def currentValue: BitStrings = iter.currentValue
  }
}
