/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: Long2IntOpenHashMapvVectorBlock.scala
  * - A vector block that contains map values
  * Version: 3.0
  */
package pegasus.spark.common.types

import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap
import pegasus.spark.common.PMVUtils

import scala.collection.JavaConverters._

class Long2IntOpenHashMapVectorBlock(numBlocks: Int, numVertices: Long, _blockId: Int)
  extends VectorBlock[Long2IntOpenHashMap](numBlocks, numVertices, _blockId) {
  private val values = Array.ofDim[Long2IntOpenHashMap](size)
  clear()

  override def get(i: Int): Long2IntOpenHashMap = {
    if (values(i) == null) {
      values(i) = new Long2IntOpenHashMap(1)
    }
    values(i)
  }

  override def put(i: Int, v: Long2IntOpenHashMap): Unit = {
    values(i) = v
  }

  override def addTo(i: Int, v: Long2IntOpenHashMap): Unit = {
    val map = get(i)
    v.long2IntEntrySet().iterator().asScala.foreach(entry => {
      map.addTo(entry.getLongKey, entry.getIntValue)
    })
  }

  override def clear(): Unit = {
    clearContainsSet()
    
    var i = 0
    while (i < size) {
      if (values(i) != null) {
        values(i).clear()
      }
      i += 1
    }
  }

  override def iterator(onlyNonzero: Boolean): VectorIterator[Long2IntOpenHashMap] = {
    new VectorIterator[Long2IntOpenHashMap] {
      private var currPos = 0
      private var currId = 0
      private var currVal: Long2IntOpenHashMap = _

      override final def next(): Boolean = {
        var found = false

        while (!found && currPos < size && PMVUtils.localToGlobal(numBlocks, blockId, currPos) < numVertices) {
          if (values(currPos) != null) {
            found = true
            currId = currPos
            currVal = values(currPos)
          }
          currPos += 1
        }

        found
      }

      override final def currentId: Int = currId

      override final def currentValue: Long2IntOpenHashMap = currVal
    }
  }

  override def iterator(onlyNonZero: Boolean, filterSet: BitSet): VectorIterator[Long2IntOpenHashMap] = {
    iterator(onlyNonZero)
  }
}
