/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: LongVectorBlock.scala
  * - A vector block that contains long values
  * Version: 3.0
  */
package pegasus.spark.common.types

import pegasus.spark.common.PMVUtils

class LongVectorBlock(numBlocks: Int, numVertices: Long, _blockId: Int, initValue: Long = -1)
  extends VectorBlock[Long](numBlocks, numVertices, _blockId) {
  private val values = Array.ofDim[Long](size)
  clear()

  override final def get(i: Int) = values(i)

  override final def put(i: Int, v: Long): Unit = {
    mark(i)
    values(i) = v
  }

  override final def addTo(i: Int, v: Long): Unit = {
    mark(i)
    values(i) += v
  }

  override final def clear(): Unit = {
    clearContainsSet()
    java.util.Arrays.fill(values, initValue)
  }

  override final def iterator(onlyNonzero: Boolean) = new VectorIterator[Long] {
    private var currPos = 0
    private var currId = 0
    private var currVal = 0L

    override final def next(): Boolean = {
      var found = false

      while (!found && currPos < size && PMVUtils.localToGlobal(numBlocks, blockId, currPos) < numVertices) {
        if (!onlyNonzero || values(currPos) != initValue) {
          found = true
          currId = currPos
          currVal = values(currPos)
        }
        currPos += 1
      }

      found
    }

    override final def currentId: Int = currId

    override final def currentValue: Long = currVal
  }

  override final def iterator(onlyNonZero: Boolean, filterSet: BitSet) = new VectorIterator[Long] {
    private val iter = iterator(onlyNonZero)

    override final def next(): Boolean = {
      var found = false

      while (!found && iter.next()) {
        found = filterSet.contains(iter.currentId)
      }

      found
    }

    override final def currentId: Int = iter.currentId

    override final def currentValue: Long = iter.currentValue
  }
}
