/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: UnweightedMatrixBlock.scala
  * - A matrix block that contains unweighted elements
  * Version: 3.0
  */
package pegasus.spark.common.types

import com.esotericsoftware.kryo.io.{Input, Output}
import com.esotericsoftware.kryo.{Kryo, Serializer}

class UnweightedMatrixBlock(
  override val blockId: Int,
  override val numCols: Int,
  override val numRows: Int
) extends MatrixBlock[Int] {
  private var lastCol = 0
  private var lastRow = 0
  private val cols = Array.ofDim[Int](numCols)
  private val degs = Array.ofDim[Int](numCols)
  private val rows = Array.ofDim[Int](numRows)
  private val rowSizes = Array.ofDim[Int](numCols)

  private[UnweightedMatrixBlock] def setMetadata(lastCol: Int, lastRow: Int): Unit = {
    this.lastCol = lastCol
    this.lastRow = lastRow
  }

  private[UnweightedMatrixBlock] def getArrays = (cols, degs, rows, rowSizes)

  override def +=(col: Int, deg: Int, rows: Array[Int]): Unit = {
    cols(lastCol) = col
    degs(lastCol) = deg
    Array.copy(rows, 0, this.rows, lastRow, rows.length)
    lastRow += rows.length
    rowSizes(lastCol) = lastRow
    lastCol += 1
  }

  override def iterator(): MatrixIterator[Int] = {
    new MatrixIterator[Int] {
      private var currPos = 0
      private var currCol = 0
      private var currDeg = 0
      private val currRows = new UnweightedMatrixRows()

      override def next(): Boolean = {
        if (currPos < lastCol) {
          currCol = cols(currPos)
          currDeg = degs(currPos)

          val begin = if (currPos == 0) 0 else rowSizes(currPos - 1)
          val end = rowSizes(currPos)
          currRows.setRows(rows, begin, end)

          currPos += 1
          true
        } else {
          false
        }
      }

      override def currentCol(): Int = currCol

      override def currentDeg(): Int = currDeg

      override def currentRows(): MatrixRows[Int] = currRows
    }
  }
}

object UnweightedMatrixBlock {

  class BlockSerializer extends Serializer[UnweightedMatrixBlock] {
    override def read(kryo: Kryo, in: Input, `type`: Class[UnweightedMatrixBlock]): UnweightedMatrixBlock = {
      val (blockId, numCols, numRows) = (in.readInt(), in.readInt(), in.readInt())
      val block = new UnweightedMatrixBlock(blockId, numCols, numRows)
      block.setMetadata(numCols, numRows)
      val (cols, degs, rows, rowSizes) = block.getArrays

      var i = 0
      while (i < numCols) {
        cols(i) = in.readInt()
        degs(i) = in.readInt()
        rowSizes(i) = in.readInt()
        i += 1
      }

      i = 0
      while (i < numRows) {
        rows(i) = in.readInt()
        i += 1
      }

      block
    }

    override def write(kryo: Kryo, out: Output, block: UnweightedMatrixBlock): Unit = {
      out.writeInt(block.blockId)
      out.writeInt(block.numCols)
      out.writeInt(block.numRows)
      val (cols, degs, rows, rowSizes) = block.getArrays

      var i = 0
      while (i < block.numCols) {
        out.writeInt(cols(i))
        out.writeInt(degs(i))
        out.writeInt(rowSizes(i))
        i += 1
      }

      i = 0
      while (i < block.numRows) {
        out.writeInt(rows(i))
        i += 1
      }
    }
  }
}
