/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: DegreeComputation.scala
  * - Computes degrees of graph
  * Version: 3.0
  */
package pegasus.spark.degree

import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.{Computation, GraphLoader}

object DegreeComputation extends Computation {

  /**
    * Represents a direction.
    */
  object Direction extends Enumeration {
    type Direction = Value

    val In, Out, Both = Value

    def fromString(str: String): Direction = str.toLowerCase match {
      case "in" => In
      case "out" => Out
      case "both" => Both
    }
  }

  import Direction.Direction

  case class Config(
    input: String = "",
    output: String = "",
    direction: Direction = null)

  private val parser = new scopt.OptionParser[Config]("Degree") {
    head("Degree")

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))

    arg[String]("direction")
      .validate(x => {
        x.toLowerCase match {
          case "in" | "out" | "both" => success
          case _ => failure("direction must be one of in, out, both")
        }
      })
      .action((x, c) => c.copy(direction = Direction.fromString(x)))
  }

  /**
    * Runs the degree computation with configuration object.
    * 
    * @param sc The SparkContext object
    * @param config The configuration object
    */
  def run(sc: SparkContext, config: Config): Unit = {
    val edges = GraphLoader.edgeListFile(sc, config.input)

    val vertices = config.direction match {
      case Direction.In => edges.map(e => (e._2, 1))
      case Direction.Out => edges.map(e => (e._1, 1))
      case Direction.Both => edges.mapPartitions(iter => {
        iter.flatMap(e => {
          if (e._1 == e._2) Seq((e._1, 1))
          else Seq((e._1, 1), (e._2, 1))
        })
      })
    }

    val degree = vertices.reduceByKey(_ + _)

    degree.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
  }

  /**
    * Runs the degree computation via CLI.
    * 
    * @param args The CLI arguments
    */
  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
