/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: DiameterComputation.scala
  * - Computes the diameter and radii of graph
  * Version: 3.0
  */
package pegasus.spark.diameter

import org.apache.spark.storage.StorageLevel
import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common._
import pegasus.spark.common.io._
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.types.{BitStrings, BitStringsVectorBlock, MatrixBlock, VectorBlock}

import scala.util.Random

object DiameterComputation extends Computation {

  case class Config(
    input: String = "",
    output: String = "",
    tempPath: Option[String] = None,
    numBlocks: Int = 40,
    denseThreshold: Int = 150,
    fieldDelimiter: String = "\t",
    maxIterations: Int = -1)

  private val parser = new scopt.OptionParser[Config]("diameter") {
    head("diameter")

    opt[Int]("numBlocks").abbr("b")
      .action((x, c) => c.copy(numBlocks = x))

    opt[Int]("denseThreshold").abbr("t")
      .action((x, c) => c.copy(denseThreshold = x))

    opt[Int]("maxIterations").abbr("i")
      .action((x, c) => c.copy(maxIterations = x))

    opt[String]("fieldDelimiter").abbr("d")
      .action((x, c) => c.copy(fieldDelimiter = x))

    opt[String]("temporaryPath").abbr("tp")
      .action((x, c) => c.copy(tempPath = Some(x)))

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))
  }

  /** Matrix-Vector multiplication for diameter computation.
    *
    * @param cardinalityPath The path to cardinality file
    */
  private def multiplication(cardinalityPath: String) = new Multiplication[Int, BitStrings, BitStrings] {
    private val random = new Random
    private val initValue = new BitStrings

    /** Given a vertex id, computes the initial bitstring for the vertex.
      *
      * @param id The vertex id
      * @return The initial bitstring
      */
    private def initBitStrings(id: Long): BitStrings = {
      random.setSeed(id)

      var i = 0
      while (i < Constants.DefaultNumBitstrings) {
        val threshold = random.nextDouble()
        var prob = 0.0
        var j = 0
        var flag = false
        while (j < 64 && !flag) {
          prob += math.pow(2, -j - 1)
          if (prob > threshold) {
            initValue.setBit(i, 1L << (63 - j))
            flag = true
          }
          j += 1
        }
        i += 1
      }

      initValue
    }

    /** @inheritdoc
      *
      * @param blockId     The block id of vector
      * @param numVertices The number of vertices
      * @param numBlocks   The number of block
      * @param init        Whether to initialize vector element or not
      * @param context     The multiplication context object
      * @return A created vector block
      */
    override def createVectorBlock(
      blockId: Int,
      numVertices: Long,
      numBlocks: Int,
      init: Boolean,
      context: MultiplicationContext
    ): VectorBlock[BitStrings] = {
      val block = new BitStringsVectorBlock(numBlocks, numVertices, blockId)
      val size = block.size
      var i = 0
      while (i < size) {
        val globalId = PMVUtils.localToGlobal(numBlocks, blockId, i)
        block.put(i, initBitStrings(globalId))
        i += 1
      }

      block
    }

    /** @inheritdoc
      *
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector reader
      */
    override def createVectorReader(vectorPath: String, blockId: Int): VectorReader[BitStrings] = {
      new BitStringsVectorReader(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      *
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector writer
      */
    override def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[BitStrings] = {
      new BitStringsVectorWriter(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      *
      * @param numVertices The number of vertices
      * @param numBlocks   The number of blocks
      * @param blockId     The block id
      * @param context     The multiplication context
      * @return A created intermediate vector block
      */
    override def createIntermediateVectorBlock(
      numVertices: Long,
      numBlocks: Int,
      blockId: Int,
      context: MultiplicationContext
    ): VectorBlock[BitStrings] = {
      new BitStringsVectorBlock(numBlocks, numVertices, blockId)
    }

    /** @inheritdoc
      *
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return The intermeidate vector reader
      */
    override def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[BitStrings] = {
      new BitStringsVectorReader(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      *
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return
      */
    override def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[BitStrings] = {
      new BitStringsVectorWriter(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * 
      * @param matrix  The matrix block
      * @param vector  The vector block
      * @param result  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combine2Op(
      matrix: MatrixBlock[Int],
      vector: VectorBlock[BitStrings],
      result: VectorBlock[BitStrings],
      context: MultiplicationContext): Unit = {
      val numBlocks = context.numBlocks
      val srcBlockId = matrix.blockId % numBlocks
      val dstBlockId = matrix.blockId / numBlocks
      val selfBlock = srcBlockId == dstBlockId

      val iter = matrix.iterator()
      while (iter.next()) {
        val col = iter.currentCol()
        val rows = iter.currentRows()
        if (!vector.contains(col)) {
          vector.addTo(col, initBitStrings(PMVUtils.localToGlobal(numBlocks, srcBlockId, col)))
        }
        val value = vector.get(col)

        var i = 0
        val size = rows.size()
        while (i < size) {
          result.addTo(rows.row(i), value)
          i += 1
        }
      }

      if (selfBlock) {
        val vecIter = vector.iterator(false)
        while (vecIter.next()) {
          result.addTo(vecIter.currentId, vecIter.currentValue)
        }
      }
    }

    /** @inheritdoc
      * 
      * @param input   The partial results
      * @param output  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combineAllOp(
      input: VectorReader[BitStrings],
      output: VectorBlock[BitStrings],
      context: MultiplicationContext
    ): Unit = {
      while (input.hasNext) {
        val (id, value) = input.next()
        output.addTo(id, value)
      }
    }

    /** @inheritdoc
      * 
      * @param result       The result from previous iteration
      * @param intermediate The intermediate result
      * @param context      The multiplication context
      */
    override def applyOp(
      result: VectorBlock[BitStrings],
      intermediate: VectorBlock[BitStrings],
      context: MultiplicationContext
    ): Unit = {
      val writer = new CardinalityWriter(cardinalityPath, result.blockId)
      val currIter = context.currentIter
      val iter = result.iterator(false)
      while (iter.next()) {
        val id = iter.currentId
        val oldValue = iter.currentValue
        val newValue = intermediate.get(id)

        if (newValue.isEmpty || oldValue == newValue) {
          val globalId = PMVUtils.localToGlobal(context.numBlocks, result.blockId, id)
          writer.write(globalId, currIter, oldValue.cardinality())
        } else {
          context.notifyChanged()
          result.put(id, newValue)
        }
      }
      writer.close()
    }
  }

  /** Runs the matrix-vector multiplication for diameter computation.
    *
    * @param sc     The SparkContext object
    * @param config The configuration object
    * @return The result
    */
  def run(sc: SparkContext, config: Config): (Double, Double, Array[(Int, Int)]) = {
    val partitionedPath = getPartitionedPath(config.output, config.tempPath)
    val intermediatePath = getIntermediatePath(config.output, config.tempPath)
    val cardinalityPath = s"$intermediatePath/cardinality"
    val edges = GraphLoader.edgeListFile(sc, config.input, config.fieldDelimiter)

    Prepartitioning.unweightedMatrix(edges, config.numBlocks, config.denseThreshold, partitionedPath)

    val (numVertices, problemRDD) = MatrixLoader.loadSubproblems(sc, partitionedPath)
    val (sparseRDD, denseRDD) = MatrixLoader.loadUnweightedMatrix(problemRDD)

    MultiplicationRunner.run(
      multiplication(cardinalityPath),
      problemRDD,
      sparseRDD,
      denseRDD,
      numVertices,
      config.maxIterations,
      intermediatePath
    )

    val rawCardinalities = sc.parallelize(0 until config.numBlocks, config.numBlocks).flatMap(blockId => {
      val reader = new CardinalityReader(cardinalityPath, blockId)
      new Iterator[(Long, (Int, Double))] {
        private var record: (Long, Int, Double) = _
        private var defined = false
        private var closed = false

        override def hasNext: Boolean = {
          if (!closed && !defined) {
            defined = reader.next()
            if (!defined) {
              reader.close()
            } else {
              record = reader.currentRecord
            }
          }

          defined
        }

        override def next(): (Long, (Int, Double)) = {
          if (!hasNext) throw new NoSuchElementException
          defined = false
          (record._1, (record._2, record._3))
        }
      }
    })
    
    val cardinalities = rawCardinalities.groupByKey().mapPartitions(iter => {
      val cardinalities = Array.ofDim[Double](Constants.MaxIterations)
      val sumCardinalities = Array.ofDim[Double](Constants.MaxIterations)
      iter.map(v => {
        val (id, values) = v
        var maxHop = Int.MinValue

        // collect cardinalities
        cardinalities.indices.foreach(i => cardinalities(i) = 0)
        values.foreach(cd => {
          cardinalities(cd._1 - 1) = cd._2
          maxHop = math.max(maxHop, cd._1)
        })
        (maxHop until Constants.MaxIterations).foreach(i => cardinalities(i) = cardinalities(i - 1))

        // estimate effective radii for each vertex
        val threshold = cardinalities(Constants.MaxIterations - 1)
        val radii = (0 until Constants.MaxIterations).filter(i => cardinalities(i) >= threshold).min

        (0 until Constants.MaxIterations).foreach(i => sumCardinalities(i) += cardinalities(i))

        (id, radii)
      })
    }).persist(StorageLevel.MEMORY_AND_DISK)
    
    val ccCard = rawCardinalities.map(_._2).reduceByKey(_ + _).collect()
    val minCC = ccCard.map(_._2).min
    val maxCC = ccCard.map(_._2).max
    val avgDiameter = ccCard.map(v => v._1 * v._2).sum / (maxCC - minCC + 1e-8)
    val thresholdCut = ccCard.filter(_._2 >= maxCC * 0.9).minBy(_._1)._1
    val cutValue = ccCard.find(_._1 == thresholdCut).get._2
    val prevCutValue = ccCard.find(_._1 == thresholdCut - 1).get._2
    val effDiameter = thresholdCut - 1 + (maxCC * 0.9 - prevCutValue) / (cutValue - prevCutValue)
    
    val dist = cardinalities.map(v => (v._2, 1)).reduceByKey(_ + _).collect()
    
    cardinalities.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
    
    (avgDiameter, effDiameter, dist)
  }

  /** Runs the matrix-vector multiplication for diameter computation via CLI.
    *
    * @param args The CLI arguments
    */
  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
