/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: LabelPropagationComputation.scala
  * - Computes the community via label propagation algorithm
  * Version: 3.0
  */
package pegasus.spark.labelprop

import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap
import org.apache.spark.storage.StorageLevel
import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common._
import pegasus.spark.common.io._
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.types.{Long2IntOpenHashMapVectorBlock, LongVectorBlock, MatrixBlock, VectorBlock}

import scala.collection.JavaConverters._

object LabelPropagationComputation extends Computation {

  case class Config(
    input: String = "",
    output: String = "",
    tempPath: Option[String] = None,
    numBlocks: Int = 40,
    denseThreshold: Int = 150,
    fieldDelimiter: String = "\t",
    maxIterations: Int = -1)

  private val parser = new scopt.OptionParser[Config]("labelprop") {
    head("labelprop")

    opt[Int]("numBlocks").abbr("b")
      .action((x, c) => c.copy(numBlocks = x))

    opt[Int]("denseThreshold").abbr("t")
      .action((x, c) => c.copy(denseThreshold = x))

    opt[Int]("maxIterations").abbr("i")
      .action((x, c) => c.copy(maxIterations = x))

    opt[String]("fieldDelimiter").abbr("d")
      .action((x, c) => c.copy(fieldDelimiter = x))

    opt[String]("temporaryPath").abbr("tp")
      .action((x, c) => c.copy(tempPath = Some(x)))

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))
  }

  private def multiplication() = new Multiplication[Int, Long, Long2IntOpenHashMap] {
    /** @inheritdoc
      * @param blockId     The block id of vector
      * @param numVertices The number of vertices
      * @param numBlocks   The number of block
      * @param init        Whether to initialize vector element or not
      * @param context     The multiplication context object
      * @return A created vector block
      */
    override def createVectorBlock(
      blockId: Int,
      numVertices: Long,
      numBlocks: Int,
      init: Boolean,
      context: MultiplicationContext
    ): VectorBlock[Long] = {
      val block = new LongVectorBlock(numBlocks, numVertices, blockId)
      if (init) {
        var i = 0
        while (i < block.size) {
          block.put(i, PMVUtils.localToGlobal(numBlocks, blockId, i))
          i += 1
        }
      }
      block
    }

    /** @inheritdoc
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector reader
      */
    override def createVectorReader(vectorPath: String, blockId: Int): VectorReader[Long] = {
      new LongVectorReader(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector writer
      */
    override def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Long] = {
      new LongVectorWriter(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * @param numVertices The number of vertices
      * @param numBlocks   The number of blocks
      * @param blockId     The block id
      * @param context     The multiplication context
      * @return A created intermediate vector block
      */
    override def createIntermediateVectorBlock(
      numVertices: Long,
      numBlocks: Int,
      blockId: Int,
      context: MultiplicationContext
    ): VectorBlock[Long2IntOpenHashMap] = {
      new Long2IntOpenHashMapVectorBlock(numBlocks, numVertices, blockId)
    }

    /** @inheritdoc
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return The intermeidate vector reader
      */
    override def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[Long2IntOpenHashMap] = {
      new Long2IntOpenHashMapVectorReader(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return
      */
    override def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Long2IntOpenHashMap] = {
      new Long2IntOpenHashMapVectorWriter(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * @param matrix  The matrix block
      * @param vector  The vector block
      * @param result  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combine2Op(
      matrix: MatrixBlock[Int],
      vector: VectorBlock[Long],
      result: VectorBlock[Long2IntOpenHashMap],
      context: MultiplicationContext)
    : Unit = {
      val iter = matrix.iterator()
      val voteVec = new Long2IntOpenHashMap(1)
      while (iter.next()) {
        val col = iter.currentCol()
        val rows = iter.currentRows()

        voteVec.clear()
        voteVec.put(vector.get(col), 1)

        var i = 0
        val size = rows.size()
        while (i < size) {
          val row = rows.row(i)
          result.addTo(row, voteVec)
          i += 1
        }
      }
    }

    /** @inheritdoc
      * @param input   The partial results
      * @param output  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combineAllOp(
      input: VectorReader[Long2IntOpenHashMap],
      output: VectorBlock[Long2IntOpenHashMap],
      context: MultiplicationContext
    ): Unit = {
      while (input.hasNext) {
        val (id, votes) = input.next()
        output.addTo(id, votes)
      }
    }

    /** @inheritdoc
      * @param result       The result from previous iteration
      * @param intermediate The intermediate result
      * @param context      The multiplication context
      */
    override def applyOp(
      result: VectorBlock[Long],
      intermediate: VectorBlock[Long2IntOpenHashMap],
      context: MultiplicationContext
    ): Unit = {
      val iter = result.iterator(false)
      while (iter.next()) {
        val id = iter.currentId
        val currVote = iter.currentValue
        var maxVote = currVote
        var maxCount = 0
        val votes = intermediate.get(id)
        votes.long2IntEntrySet().asScala.foreach(entry => {
          if (maxCount < entry.getIntValue) {
            maxVote = entry.getLongKey
            maxCount = entry.getIntValue
          }
        })

        if (currVote != maxVote && maxVote != -1) {
          result.put(id, maxVote)
          context.notifyChanged()
        }
      }
    }
  }

  /** Runs the matrix-vector multiplication for label propagation algorithm.
    *
    * @param sc     The SparkContext object
    * @param config The configuration object
    * @return The result
    */
  def run(sc: SparkContext, config: Config): (Long, Array[(Long, Int)]) = {
    val partitionedPath = getPartitionedPath(config.output, config.tempPath)
    val intermediatePath = getIntermediatePath(config.output, config.tempPath)

    val edges = GraphLoader.edgeListFile(sc, config.input, config.fieldDelimiter)

    Prepartitioning.unweightedMatrix(edges, config.numBlocks, config.denseThreshold, partitionedPath)

    val (numVertices, problemRDD) = MatrixLoader.loadSubproblems(sc, partitionedPath)
    val (sparseRDD, denseRDD) = MatrixLoader.loadUnweightedMatrix(problemRDD)

    val result = MultiplicationRunner.run(
      multiplication(),
      problemRDD,
      sparseRDD,
      denseRDD,
      numVertices,
      config.maxIterations,
      intermediatePath
    ).persist(StorageLevel.MEMORY_AND_DISK)

    val communityCount = result.map(_._2).distinct().count()
    val communities = result.map(v => (v._2, 1)).reduceByKey(_ + _).sortBy(_._2, ascending = false).take(10)

    result.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)

    (communityCount, communities)
  }

  /** Runs the matrix-vector multiplication for label propagation algorithm via CLI.
    *
    * @param args The CLI arguments
    */
  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
