/*
 * PegasusN: Peta-Scale Graph Mining System v3.0 (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: NetRayScatter.java
 * - Draw a scatter plot of the given data
 * Version: 3.0
 */
package pegasus.spark.netray

import org.apache.spark.rdd.RDD
import org.apache.spark.{SparkConf, SparkContext}

object NetRayScatter {

  case class Options(inputPath: String, outputPath: String, targetWidth: Int, fieldIds: Seq[Int], logScale: Boolean)

  private def parseOptions(args: Array[String]): Options = {
    val parser = new scopt.OptionParser[Options]("netray-scatter") {
      opt[String]("inputPath").abbr("i")
        .action((x, c) => c.copy(inputPath = x))
        .required()
      opt[String]("outputPath").abbr("o")
        .action((x, c) => c.copy(outputPath = x))
        .required()
      opt[Int]("targetWidth").abbr("w")
        .action((x, c) => c.copy(targetWidth = x))
        .required()
      opt[Seq[Int]]("fieldIds").abbr("f")
        .action((x, c) => c.copy(fieldIds = x))
        .validate(x => if (x.length == 2) success else failure("The number of specified fields must be 2."))
      opt[Unit]("logScale").abbr("l")
        .action((_, c) => c.copy(logScale = true))
    }

    parser.parse(args, Options("", "", -1, Seq(0, 1), logScale = false)) match {
      case Some(options) => options
      case None => sys.exit(-1)
    }
  }

  def main(args: Array[String]): Unit = {
    val options = parseOptions(args)
    val conf = new SparkConf()
      .setAppName(s"NetRayScatter - ${options.inputPath}")
    val sc = new SparkContext(conf)

    // accumulators for min/max values
    val (xMin, xMax, yMin, yMax) = (new MinAccumulator, new MaxAccumulator, new MinAccumulator, new MaxAccumulator)
    sc.register(xMin, "xMin")
    sc.register(xMax, "xMax")
    sc.register(yMin, "yMin")
    sc.register(yMax, "yMax")

    // load matrix
    val Seq(xField, yField) = options.fieldIds
    val initMat = sc.textFile(options.inputPath).mapPartitions(iter => {
      val delimiter = "\\s+".r
      iter.filter(!_.startsWith("#")).map(line => {
        val split = delimiter.split(line)
        val x = split(xField).toLong
        val y = split(yField).toLong
        xMin.add(x)
        xMax.add(x)
        yMin.add(y)
        yMax.add(y)
        (x, y)
      })
    }, preservesPartitioning = true).cache()

    // trigger empty action to compute min/max values
    initMat.foreachPartition(_ => ())

    // re-scale matrix
    val scaledMat = if (options.logScale) {
      // log scale
      val (diffX, diffY) = (math.log(xMin.value), math.log(yMin.value))
      val convertXRatio = (options.targetWidth - 1) / (math.log(xMax.value) - math.log(xMin.value))
      val convertYRatio = (options.targetWidth - 1) / (math.log(yMax.value) - math.log(yMin.value))
      initMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y) = mat
          val xBlk = (math.log(x) - diffX) * convertXRatio + 0.5
          val yBlk = (math.log(y) - diffY) * convertYRatio + 0.5
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val (diffX, diffY) = (xMin.value, yMin.value)
      val convertXRatio = (options.targetWidth - 1) / (xMax.value - xMin.value)
      val convertYRatio = (options.targetWidth - 1) / (yMax.value - yMin.value)
      initMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y) = mat
          val xBlk = (x - diffX) * convertXRatio + 0.5
          val yBlk = (y - diffY) * convertYRatio + 0.5
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc)


    // output
    val outputRDD = aggregatedMat.map(mat => s"${mat._1._1}\t${mat._1._2}\t${mat._2}")
    outputRDD.saveAsTextFile(options.outputPath)

    sc.stop()
  }

  def run(inputRDD: RDD[(Double, Double)], targetWidth: Int,
          logScale: Boolean, sc: SparkContext): RDD[(Int, Int, Long)] = {

    // accumulators for min/max values
    val (xNzMin, yNzMin) = (new MinAccumulator, new MinAccumulator)
    val (xMin, xMax, yMin, yMax) = (new MinAccumulator, new MaxAccumulator, new MinAccumulator, new MaxAccumulator)
    sc.register(xNzMin, "xNzMin")
    sc.register(xMin, "xMin")
    sc.register(xMax, "xMax")
    sc.register(yNzMin, "yNzMin")
    sc.register(yMin, "yMin")
    sc.register(yMax, "yMax")

    // load matrix
    val initMat = inputRDD.map{ case (x, y) =>
      if (x > 0) xNzMin.add(x)
      if (y > 0) yNzMin.add(y)
      xMin.add(x)
      xMax.add(x)
      yMin.add(y)
      yMax.add(y)
      (x, y)
    }.cache()

    // trigger empty action to compute min/max values
    initMat.count()

    // re-scale matrix
    val scaledMat = if (logScale) {
      // log scale
      val (diffX, diffY) = (math.log(xNzMin.value), math.log(yNzMin.value))
      val convertXRatio = (targetWidth - 1) / (math.log(xMax.value) - math.log(xNzMin.value))
      val convertYRatio = (targetWidth - 1) / (math.log(yMax.value) - math.log(yNzMin.value))
      initMat.mapPartitions(iter => {
        iter.filter(v => v._1 > 0 && v._2 > 0).map(mat => {
          val (x, y) = mat
          val xBlk = ((math.log(x) - diffX) * convertXRatio + 0.5).toInt
          val yBlk = ((math.log(y) - diffY) * convertYRatio + 0.5).toInt
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val (diffX, diffY) = (xMin.value, yMin.value)
      val convertXRatio = (targetWidth - 1) / (xMax.value - xMin.value)
      val convertYRatio = (targetWidth - 1) / (yMax.value - yMin.value)
      initMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y) = mat
          val xBlk = ((x - diffX) * convertXRatio + 0.5).toInt
          val yBlk = ((y - diffY) * convertYRatio + 0.5).toInt
          ((xBlk, yBlk), 1L)
        })
      }, preservesPartitioning = true)
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc)

    // output
    aggregatedMat.map{case ((u, v), k) => (u, v, k)}
  }
}
