/*
 * PegasusN: Peta-Scale Graph Mining System v3.0 (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: NetRaySpy.java
 * - Draw a spy plot of the given matrix
 * Version: 3.0
 */
package pegasus.spark.netray

import org.apache.spark.rdd.RDD
import org.apache.spark.{SparkConf, SparkContext}

object NetRaySpy {

  case class Options(inputPath: String, outputPath: String, numNodes: Long, targetWidth: Int, logScale: Boolean)

  private def parseOptions(args: Array[String]) = {
    val parser = new scopt.OptionParser[Options]("netray-spy") {
      opt[String]("inputPath").abbr("i")
        .action((x, c) => c.copy(inputPath = x))
        .required()
      opt[String]("outputPath").abbr("o")
        .action((x, c) => c.copy(outputPath = x))
        .required()
      opt[Long]("numNodes").abbr("n")
        .action((x, c) => c.copy(numNodes = x))
        .required()
      opt[Int]("targetWidth").abbr("w")
        .action((x, c) => c.copy(targetWidth = x))
        .required()
      opt[Unit]("logScale").abbr("l")
        .action((_, c) => c.copy(logScale = true))
    }

    parser.parse(args, Options("", "", -1L, -1, logScale = false)) match {
      case Some(options) => options
      case None => sys.exit(-1)
    }
  }

  def main(args: Array[String]): Unit = {
    val options = parseOptions(args)
    val conf = new SparkConf()
      .setAppName(s"NetRaySpy - ${options.inputPath}")
    val sc = new SparkContext(conf)

    // load adjacency matrix
    val adjMat = sc.textFile(options.inputPath).mapPartitions(iter => {
      val delimiter = "\\s+".r
      iter.filter(!_.startsWith("#")).map(line => {
        val split = delimiter.split(line)
        if (split.length == 2) {
          (split(0).toLong, split(1).toLong, 1L)
        } else {
          (split(0).toLong, split(1).toLong, split(2).toLong)
        }
      })
    }, preservesPartitioning = true)

    // re-scale matrix
    val scaledMat = if (options.logScale) {
      // log scale
      val convertRatio = (options.targetWidth - 1) / math.log10(options.numNodes)
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = (math.log10(x + 1) * convertRatio + 0.5).toInt
          val yBlk = (math.log10(y + 1) * convertRatio + 0.5).toInt
          ((xBlk, yBlk), value)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val convertRatio = options.targetWidth / options.numNodes.toDouble
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = ((x + 1) * convertRatio).toInt
          val yBlk = ((y + 1) * convertRatio).toInt
          ((xBlk, yBlk), value)
        })
      })
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc)

    // output
    val outputRDD = aggregatedMat.map(mat => s"${mat._1._1}\t${mat._1._2}\t${mat._2}")
    outputRDD.saveAsTextFile(options.outputPath)

    sc.stop()
  }

  def run(inputPath: String, numNodes: Long, targetWidth: Int,
          isLogScale: Boolean, sc:SparkContext): RDD[(Int, Int, Long)] = {

    // load adjacency matrix
    val adjMat = sc.textFile(inputPath).mapPartitions(iter => {
      val delimiter = "\\s+".r
      iter.filter(!_.startsWith("#")).map(line => {
        val split = delimiter.split(line)
        if (split.length == 2) {
          (split(0).toLong, split(1).toLong, 1L)
        } else {
          (split(0).toLong, split(1).toLong, split(2).toLong)
        }
      })
    }, preservesPartitioning = true)

    // re-scale matrix
    val scaledMat = if (isLogScale) {
      // log scale
      val convertRatio = (targetWidth - 1) / math.log10(numNodes)
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = (math.log10(x + 1) * convertRatio + 0.5).toInt
          val yBlk = (math.log10(y + 1) * convertRatio + 0.5).toInt
          ((xBlk, yBlk), value)
        })
      }, preservesPartitioning = true)
    } else {
      // linear scale
      val convertRatio = targetWidth / numNodes.toDouble
      adjMat.mapPartitions(iter => {
        iter.map(mat => {
          val (x, y, value) = mat
          val xBlk = ((x + 1) * convertRatio).toInt
          val yBlk = ((y + 1) * convertRatio).toInt
          ((xBlk, yBlk), value)
        })
      })
    }

    // aggregate
    val aggFunc = (v1: Long, v2: Long) => v1 + v2
    val aggregatedMat = scaledMat.aggregateByKey(0L)(aggFunc, aggFunc).map{
      case ((u, v), k) =>
        (u, v, k)
    }
    aggregatedMat
  }

}
