/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PageRankComputation.scala
  * - Computes the PageRank scores
  * Version: 3.0
  */
package pegasus.spark.pagerank

import org.apache.spark.storage.StorageLevel
import org.apache.spark.{SparkConf, SparkContext}
import pegasus.spark.common._
import pegasus.spark.common.io.{DoubleVectorReader, DoubleVectorWriter, VectorReader, VectorWriter}
import pegasus.spark.common.serialization.PMVRegistrator
import pegasus.spark.common.types.{DoubleVectorBlock, MatrixBlock, VectorBlock}

object PageRankComputation extends Computation {

  case class Config(
    input: String = "",
    output: String = "",
    tempPath: Option[String] = None,
    alpha: Double = 0.85,
    sourceVertex: Option[Long] = None,
    numBlocks: Int = 40,
    denseThreshold: Int = 150,
    fieldDelimiter: String = "\t",
    maxIterations: Int = -1)

  private val parser = new scopt.OptionParser[Config]("PageRank") {
    head("PageRank")

    opt[Int]("numBlocks").abbr("b")
      .action((x, c) => c.copy(numBlocks = x))

    opt[Int]("denseThreshold").abbr("t")
      .action((x, c) => c.copy(denseThreshold = x))

    opt[Int]("maxIterations").abbr("i")
      .action((x, c) => c.copy(maxIterations = x))

    opt[Double]("alpha").abbr("a")
      .action((x, c) => c.copy(alpha = x))

    opt[Long]("sourceVertex").abbr("sv")
      .action((x, c) => c.copy(sourceVertex = Some(x)))

    opt[String]("fieldDelimiter").abbr("d")
      .action((x, c) => c.copy(fieldDelimiter = x))

    opt[String]("temporaryPath").abbr("tp")
      .action((x, c) => c.copy(tempPath = Some(x)))

    arg[String]("input")
      .action((x, c) => c.copy(input = x))

    arg[String]("output")
      .action((x, c) => c.copy(output = x))
  }

  private def multiplication(
    alpha: Double,
    sourceVertex: Option[Long],
    threshold: Double
  ) = new Multiplication[Int, Double, Double] {
    /** @inheritdoc
      * 
      * @param blockId     The block id of vector
      * @param numVertices The number of vertices
      * @param numBlocks   The number of block
      * @param init        Whether to initialize vector element or not
      * @param context     The multiplication context object
      * @return A created vector block
      */
    override def createVectorBlock(
      blockId: Int,
      numVertices: Long,
      numBlocks: Int,
      init: Boolean,
      context: MultiplicationContext
    ): VectorBlock[Double] = {
      val block = new DoubleVectorBlock(numBlocks, numVertices, blockId)
      if (init) {
        sourceVertex match {
          case Some(v) =>
            if (v % numBlocks == blockId) {
              block.put(PMVUtils.globalToLocal(numBlocks, v), 1.0)
            }
          case None =>
            var i = 0
            while (i < block.size) {
              block.put(i, 1.0 / numVertices)
              i += 1
            }
        }
      }
      block
    }

    /** @inheritdoc
      * 
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector reader
      */
    override def createVectorReader(vectorPath: String, blockId: Int): VectorReader[Double] = {
      new DoubleVectorReader(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * 
      * @param vectorPath The path to vector blocks
      * @param blockId    The id of vector block
      * @return A created vector writer
      */
    override def createVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Double] = {
      new DoubleVectorWriter(f"$vectorPath/${Constants.DenseVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * 
      * @param numVertices The number of vertices
      * @param numBlocks   The number of blocks
      * @param blockId     The block id
      * @param context     The multiplication context
      * @return A created intermediate vector block
      */
    override def createIntermediateVectorBlock(
      numVertices: Long,
      numBlocks: Int,
      blockId: Int,
      context: MultiplicationContext
    ): VectorBlock[Double] = {
      new DoubleVectorBlock(numBlocks, numVertices, blockId)
    }

    /** @inheritdoc
      * 
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return The intermeidate vector reader
      */
    override def createIntermediateVectorReader(vectorPath: String, blockId: Int): VectorReader[Double] = {
      new DoubleVectorReader(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * 
      * @param vectorPath The path to intermediate vector file
      * @param blockId    The block id
      * @return
      */
    override def createIntermediateVectorWriter(vectorPath: String, blockId: Int): VectorWriter[Double] = {
      new DoubleVectorWriter(f"$vectorPath/${Constants.IntermediateVectorBlock}-$blockId%05d")
    }

    /** @inheritdoc
      * 
      * @param matrix  The matrix block
      * @param vector  The vector block
      * @param result  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combine2Op(
      matrix: MatrixBlock[Int],
      vector: VectorBlock[Double],
      result: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      val iter = matrix.iterator()
      while (iter.next()) {
        val col = iter.currentCol()
        val rows = iter.currentRows()
        val value = vector.get(col) / iter.currentDeg()

        var i = 0
        val size = rows.size()
        while (i < size) {
          val row = rows.row(i)
          result.addTo(row, value)
          i += 1
        }
      }
    }

    /** @inheritdoc
      * 
      * @param input   The partial results
      * @param output  A vector block to contain the result
      * @param context The multiplication context
      */
    override def combineAllOp(
      input: VectorReader[Double],
      output: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      while (input.hasNext) {
        val (col, value) = input.next()
        output.addTo(col, value)
      }
    }

    /** @inheritdoc
      * 
      * @param result       The result from previous iteration
      * @param intermediate The intermediate result
      * @param context      The multiplication context
      */
    override def applyOp(
      result: VectorBlock[Double],
      intermediate: VectorBlock[Double],
      context: MultiplicationContext
    ): Unit = {
      val iter = result.iterator(false)
      val (personalized, srcId) = sourceVertex match {
        case Some(v) => (true, v)
        case None => (false, -1L)
      }
      while (iter.next()) {
        val id = iter.currentId
        val jumpProb = if (!personalized) 1 - alpha else if (id == srcId) 1 - alpha else 0
        val oldValue = iter.currentValue
        val newValue = if (intermediate.contains(id)) {
          alpha * intermediate.get(id) + jumpProb / context.numVertices
        } else {
          jumpProb / context.numVertices
        }

        if (math.abs(oldValue - newValue) > threshold) {
          context.notifyChanged()
          result.put(id, newValue)
        }
      }
    }
  }

  /** Runs the matrix-vector multiplication for PageRank algorithm.
    *
    * @param sc     The SparkContext object
    * @param config The configuration object
    * @return The top 10 PageRank scores
    */
  def run(sc: SparkContext, config: Config): Array[(Long, Double)] = {
    val partitionedPath = getPartitionedPath(config.output, config.tempPath)
    val intermediatePath = getIntermediatePath(config.output, config.tempPath)
    val edges = GraphLoader.edgeListFile(sc, config.input, config.fieldDelimiter)

    Prepartitioning.unweightedMatrix(edges, config.numBlocks, config.denseThreshold, partitionedPath)

    val (numVertices, problemRDD) = MatrixLoader.loadSubproblems(sc, partitionedPath)
    val (sparseRDD, denseRDD) = MatrixLoader.loadUnweightedMatrix(problemRDD)

    val result = MultiplicationRunner.run(
      multiplication(config.alpha, config.sourceVertex, 1e-7),
      problemRDD,
      sparseRDD,
      denseRDD,
      numVertices,
      config.maxIterations,
      intermediatePath
    ).persist(StorageLevel.MEMORY_AND_DISK)
    
    val top10Scores = result.sortBy(_._2, ascending = false).take(10)

    result.map(v => s"${v._1}\t${v._2}").saveAsTextFile(config.output)
    
    top10Scores
  }

  /** Runs the matrix-vector multiplication for PageRank score computation via CLI.
    *
    * @param args The CLI arguments
    */
  def main(args: Array[String]): Unit = {
    parser.parse(args, Config()) match {
      case Some(config) =>
        val conf = new SparkConf()
          .setAppName("PMV - PageRank")
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", classOf[PMVRegistrator].getName)
          .set("spark.kryoserializer.buffer", "24m")
        val sc = new SparkContext(conf)

        run(sc, config)

        sc.stop()
      case _ => System.exit(-1)
    }
  }
}
