/*
 * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
 * Authors: Chiwan Park, Ha-Myung Park, U Kang
 *
 * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Seoul National University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * -------------------------------------------------------------------------
 * File: EdgeOrder.scala
 * - Make an oriented graph from an undirected graph by vertex degree.
 * Version: 3.0
 */

package pegasus.spark.subgraph

import java.util.StringTokenizer

import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark.{SparkConf, SparkContext}

/**
  * Created by hmpark on 17. 3. 14.
  */
object EdgeOrder {

  Logger.getLogger("org").setLevel(Level.OFF)
  Logger.getLogger("akka").setLevel(Level.OFF)

  val logger = Logger.getLogger(EdgeOrder.getClass)

  def main(args: Array[String]): Unit = {
    val inputPath = args(0)
    val outputPath = args(1)

    val local = if(args.length >= 3) args(2).toBoolean else false

    logger.info(f"inputPath: $inputPath, output: $outputPath")

    run(inputPath, outputPath, local)

  }

  def run(inputPath: String, outputPath: String, local: Boolean): Unit = {

    var time_startup = System.currentTimeMillis()

    val conf = new SparkConf().setAppName("[EdgeOrder]" + inputPath) //.setMaster("local")
      .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
      .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
      .set("spark.kryoserializer.buffer.mb", "24")

    if(local) conf.setMaster("local[2]")

    val sc = new SparkContext(conf)

    val fs = FileSystem.get(sc.hadoopConfiguration)

    //find numNodes
    sc.textFile(inputPath).flatMap { line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong

      if ( u != v ) Seq((u, v), (v, u))
      else Seq()

    }.groupByKey().flatMap {
      case (u, _uN) =>
        val uN = _uN.toArray.sorted

        val uDeg = uN.foldLeft((-1L, 0)){case ((prev, uDeg), v) => (v, if(prev != v) uDeg + 1 else uDeg) }._2

        var prev = -1L
        uN.filter{v => val res = v != prev; prev = v; res}.map{v => ((u, v), uDeg)}
    }.map{
      case ((u, v), uDeg) =>
        if( u < v ) ((u,v), uDeg)
        else ((v, u), -uDeg)
    }.groupByKey().map{
      case ((u, v), degs) =>

        val it = degs.iterator
        val tmp = it.next
        val (uDeg, vDeg) = if (tmp < 0) (it.next, -tmp)
                           else (tmp, -it.next)

        if (uDeg < vDeg || (uDeg == vDeg && u < v))
          (u, v)
        else
          (v, u)
    }.saveAsSequenceFile(outputPath)

    sc.stop()
  }

  def run(inputPath: String, outputPath: String, sc: SparkContext): Unit = {

    var time_startup = System.currentTimeMillis()

    //find numNodes
    sc.textFile(inputPath).flatMap { line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong

      if ( u != v ) Seq((u, v), (v, u))
      else Seq()

    }.groupByKey().flatMap {
      case (u, _uN) =>
        var prev = -1L
        val uN = _uN.toArray.sorted.filter{v => val res = v != prev; prev = v; res}
        val uDeg = uN.length

        uN.map{v => ((u, v), uDeg)}
    }.map{
      case ((u, v), uDeg) =>
        if( u < v ) ((u,v), uDeg)
        else ((v, u), -uDeg)
    }.groupByKey().map{
      case ((u, v), degs) =>

        val it = degs.iterator
        val tmp = it.next
        val (uDeg, vDeg) = if (tmp < 0) (it.next, -tmp)
        else (tmp, -it.next)

        if (uDeg < vDeg || (uDeg == vDeg && u < v))
          (u, v)
        else
          (v, u)
    }.saveAsSequenceFile(outputPath)

  }

}
