/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PSE.scala
  * - Find subgraphs in a graph.
  * Version: 3.0
  */
package pegasus.spark.subgraph

import java.io.FileNotFoundException
import java.lang.management.ManagementFactory
import java.util
import java.util.StringTokenizer

import com.esotericsoftware.kryo.KryoException
import com.esotericsoftware.kryo.io.{Input, Output}
import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark._
import org.apache.spark.broadcast.Broadcast
import org.apache.spark.rdd.RDD
import org.xerial.snappy.{SnappyInputStream, SnappyOutputStream}

class PSE extends Serializable{

  Logger.getLogger("org").setLevel(Level.WARN)
  Logger.getLogger("akka").setLevel(Level.WARN)

  val APP_NAME: String = "pse"
  val EXEC_NAME: String = "pse"
  val VERSION: String = "3.0"

  /**
    * command line interface with default parameters
    * @param numColors the number of vertex colors
    * @param inputPath input file path
    * @param outputPath output file path
    * @param patternString the binary string of a query graph
    */
  case class Config(numColors: Int = 10, inputPath: String = "",
                      outputPath: String = null, patternString: String = null)

  /** run by command line interpace */
  def runByCLI(args: Array[String]): Unit = {
    val parser = new scopt.OptionParser[Config](EXEC_NAME){
      head(APP_NAME, VERSION)

      opt[Int]('p', "numColors")
        .action((x, c) => c.copy(numColors = x))
        .text("the number of colors. (default: 80)")

      opt[String]('q', "patternString").required()
        .action((x, c) => c.copy(patternString = x))
        .text("pattern string.")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output").optional()
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")
    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        PSE.logger.info(f"input: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"num colors: ${opts.numColors}")

        val conf = new SparkConf().setAppName("[" + APP_NAME + "]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")

        val sc = new SparkContext(conf)

        val res = run(opts.inputPath, opts.patternString, opts.numColors.toShort, sc)

        if(opts.outputPath == null) {
          println("opts.patternString: " + opts.patternString)
          println(f"the number of subgraphs (pattern: ${opts.patternString}): ${res.count()}") // count mode
        }
        else{
          FileSystem.get(sc.hadoopConfiguration).delete(new Path(opts.outputPath), true)
          res.map{x => x.mkString("\t")}.saveAsTextFile(opts.outputPath)
        } // list mode

        sc.stop()

      case None =>
    }
  }

  /**
    * submit the spark job.
    *
    * @param inputPath input file path
    * @param patternString the binary string of a query graph
    * @param numColors the number of colors
    * @param sc spark context
    * @return an RDD containing triangles
    */
  def run(inputPath: String, patternString: String, numColors: Short,
          sc: SparkContext): RDD[Array[Long]] = {

    val partPath = inputPath + ".part/"

    val fs = FileSystem.get(sc.hadoopConfiguration)
    fs.delete(new Path(partPath), true)
    fs.mkdirs(new Path(partPath))

    //find numNodes
    val numNodes = sc.broadcast(sc.textFile(inputPath).map { line =>
      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong

      if (u < v) v
      else u
    }.max() + 1)

    partitionGraph(inputPath, partPath, numColors, sc)

    fs.deleteOnExit(new Path(partPath))

    enumerateSubgraphs(partPath, patternString, numColors, numNodes, sc)
  }

  /**
    * Find subgraphs that match a query graph.
    *
    * @param partPath partitioned file path
    * @param patternString the binary string of a query graph
    * @param numColors the number of colors
    * @param numNodes the number of nodes in the input graph
    * @param sc spark context
    * @return
    */
  def enumerateSubgraphs(partPath: String, patternString: String, numColors: Short,
                         numNodes: Broadcast[Long], sc: SparkContext): RDD[Array[Long]] = {

    val numSubgraphs = sc.longAccumulator("numSubgraphs")

    var patternSize = 1
    while (patternSize * (patternSize - 1) / 2 < patternString.length()) {
      patternSize += 1
    }

    val set_subproblems = SeedGenerator.dual(numColors, patternSize).map { arr => arr(0) -> arr }
      .zipWithIndex
      .map { case (a, b) => (b, a) }

    val parts = sc.parallelize(set_subproblems).partitionBy(new PSESubproblemPartitioner(set_subproblems.length))

    parts.mapPartitions { it =>

      printGC()

      // in the first item of the iterator is the problemGroup.

      val part = new Path(partPath)

      val _problem = it.next()

      val problemGroup = _problem._2._1
      val problems = _problem._2._2

      val colors = (0 until numColors).filter { i => (problemGroup & 1 << i) != 0 }.toArray

      val numLocalColors = colors.length

      val nodeSize = numNodes.value
      val nodeSize_edgeSet = (nodeSize / numColors + (if (nodeSize % numColors == 0) 0 else 1)).toInt
      val nodeSize_problemGroup = nodeSize_edgeSet * numLocalColors

      val adj = new Array[Array[Int]](nodeSize_problemGroup)

      PSE.logger.info("Colors of this problem: (" + colors.mkString(", ") + ")")

      for (i <- colors.indices; j <- colors.indices if i != j || colors.length != patternSize) {
        loadAdj(part.suffix("/edgeset-" + (colors(i) * numColors + colors(j))),
          adj, i, j, colors.length, FileSystem.get(new Configuration))
        PSE.logger.info(f"edgeset-(${colors(i)}, ${colors(j)}) is loaded" +
          f"(free-mem: ${VF2_degree.availableMemoryBytes(false) / 1000000.0}%.2fMB)")
      }

      PSE.logger.info("Block loading complete.")

      val emptyArr = new Array[Int](0)

      for (i <- adj.indices) {
        adj(i) =
          if (adj(i) != null)
            adj(i).sortBy { x => (adj(x).length.toLong << 32) | x }
          else emptyArr

      }

      PSE.logger.info(f"Adj sorting complete. (free-mem ${VF2_degree.availableMemoryBytes(true) / 1000000.0}%.2fMB)")

      val filter = problems.map { problem =>
        var colorbit = 0L

        for (i <- colors.indices) {
          colorbit |= ((problem >>> colors(i)) & 1) << i
        }

        colorbit
      }

      PSE.logger.info("filters: (" + filter.map{x=>x.toBinaryString}.mkString(",") + ")")


      val vf2 = new VF2(adj, patternString, filter, colors, numColors)

      val subgraphs = vf2.subgraphs()

      PSE.logger.info("Subgraph enumerate complete.")

      subgraphs
    }
  }



  /**
    * partition the input graph.
    * The partitioned graph are stored in partPath
    * A partition with colors i and j is stored in partPath/edgeset-{i*numColors + j} as a binary file
    * whose format is as follows:
    * node1 #neighbors-of-node1 neighbor1 neighbor2 ...
    * node2 #neighbors-of-node2 neighbor1 ...
    *
    * @param inputPath input file path
    * @param partPath partition file path
    * @param numColors the number of colors
    * @param sc spark context
    */
  def partitionGraph(inputPath: String, partPath: String, numColors: Short, sc: SparkContext): Unit = {
    sc.textFile(inputPath).flatMap(line => {

      val st = new StringTokenizer(line)
      val u = st.nextToken().toLong
      val v = st.nextToken().toLong

      val u_local = (u / numColors).toInt
      val v_local = (v / numColors).toInt
      val u_color = (u - u_local * numColors).toShort
      val v_color = (v - v_local * numColors).toShort

      Seq(
        ((u_color * numColors + v_color).toShort, (u_local.toLong << 32) | v_local),
        ((v_color * numColors + u_color).toShort, (v_local.toLong << 32) | u_local)
      )

    })
      .partitionBy(new PSEPartitioner(numColors))
      .foreachPartition { it =>

        var blockColor: Short = -1


        val edgeList = it.map { item =>
          blockColor = item._1
          item._2
        }.toArray

        util.Arrays.sort(edgeList)

        val fs = FileSystem.get(new Configuration)
        val out = new Output(
          new SnappyOutputStream(
            fs.create(
              new Path(partPath + "/edgeset-" + blockColor)
            ).getWrappedStream)
        )

        var uStart = 0
        var uEnd = 0
        while (uStart < edgeList.length) {

          val edge = edgeList(uStart)
          val u = (edge >>> 32).toInt
          val limit = (u + 1).toLong << 32

          var count_uniq = 0
          try {
            var prev = -1L
            var cur = 0L
            while ( {
              cur = edgeList(uEnd)
              cur < limit
            }) {
              if (prev != cur) {
                count_uniq += 1
                prev = cur
              }
              uEnd += 1
            }
          } catch {
            case e: ArrayIndexOutOfBoundsException =>
          }

          out.writeInt(u, true)
          out.writeInt(count_uniq, true)

          var prev = -1L
          var cur = 0L
          for (i <- uStart until uEnd) {
            cur = edgeList(i)
            if (prev != cur) {
              out.writeInt(cur.toInt, true)
              prev = cur
            }
          }

          uStart = uEnd
        }

        out.close()
      }
  }

  /**
    * load a graph of an adj format.
    *
    * @param graphPath graph file path
    * @param adj to contain graph
    * @param localColor_src the color of source nodes
    * @param localColor_dst the color of destination nodes
    * @param numColors the number of colors
    * @param fs hadoop file system
    */
  def loadAdj(graphPath: Path, adj: Array[Array[Int]], localColor_src: Int,
              localColor_dst: Int, numColors: Int, fs: FileSystem): Unit = {

    try{

      val in = new Input(new SnappyInputStream(fs.open(graphPath).getWrappedStream))

      try{
        while(true){
          val u_local = in.readInt(true)
          val uSize = in.readInt(true)
          val u = u_local * numColors + localColor_src

          var length_prev = 0
          val adj_u = if(adj(u) != null){
            length_prev = adj(u).length
            val adj_u_tmp = new Array[Int](length_prev + uSize)
            System.arraycopy(adj(u), 0, adj_u_tmp, 0, length_prev)
            adj_u_tmp
          }
          else{
            new Array[Int](uSize)
          }

          adj(u) = adj_u

          for(i <- length_prev until adj_u.length){
            adj_u(i) = in.readInt(true) * numColors + localColor_dst
          }

        }
      } catch{
        case e: KryoException =>
      }

      in.close()

    } catch{
      case e: FileNotFoundException =>
    }
  }

  /**
    * print gc type
    */
  def printGC(): Unit = {
    val gcMxBeans = ManagementFactory.getGarbageCollectorMXBeans.iterator

    PSE.logger.info("garbage collector info=====")
    while (gcMxBeans.hasNext) {
      val gcMxBean = gcMxBeans.next()

      PSE.logger.info(gcMxBean.getName)
      PSE.logger.info(gcMxBean.getObjectName)
    }
    PSE.logger.info("===========================")
  }

  class PSESubproblemPartitioner(numSubproblems: Int) extends Partitioner {
    def getPartition(key: Any): Int = key.asInstanceOf[Int]
    def numPartitions: Int = numSubproblems
  }

  class PSEPartitioner(numColors: Int) extends Partitioner {
    def getPartition(key: Any): Int = key.asInstanceOf[Short]
    def numPartitions: Int = { numColors * numColors }
  }
}

object PSE{
  private val logger = Logger.getLogger(getClass)
  logger.setLevel(Level.INFO)

  def main(args: Array[String]): Unit = new PSE().runByCLI(args)
}