/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PTE.java
  * - Find Triangles in a graph.
  * Version: 3.0
  */
package pegasus.spark.triangle

import java.lang.management.ManagementFactory
import java.util

import com.esotericsoftware.kryo.KryoException
import com.esotericsoftware.kryo.io.{Input, Output}
import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark.broadcast.Broadcast
import org.apache.spark.rdd.RDD
import org.apache.spark.{Partitioner, SparkConf, SparkContext}
import org.xerial.snappy.{SnappyInputStream, SnappyOutputStream}
import pegasus.spark.subgraph.SeedGenerator

class PTE extends Serializable{

  val APP_NAME: String = "pte"
  val EXEC_NAME: String = "pte"
  val VERSION: String = "3.0"

  /**
    * command line interface with default parameters
    * @param numColors the number of vertex colors
    * @param inputPath input file path
    * @param outputPath output file path
    */
  case class Config(numColors: Int = 10, inputPath: String = "", outputPath: String = null)

  /** run by command line interface */
  def runByCLI(args: Array[String]): Unit = {

    val parser = new scopt.OptionParser[Config](EXEC_NAME){
      head(APP_NAME, VERSION)

      opt[Int]('p', "numColors")
        .action((x, c) => c.copy(numColors = x))
        .text("the number of colors. (default: 10)")

      arg[String]("input")
        .action((x, c) => c.copy(inputPath = x))
        .text("input file path.")

      arg[String]("output").optional()
        .action((x,c) => c.copy(outputPath = x))
        .text("output file path.")
    }

    parser.parse(args, Config()) match {
      case Some(opts) =>

        PTE.logger.info(f"input: ${opts.inputPath}, output: ${opts.outputPath}, " +
          f"num colors: ")

        val conf = new SparkConf().setAppName("[" + APP_NAME + "]" + opts.inputPath)
          .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
          .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
          .set("spark.kryoserializer.buffer", "24m")

        val sc = new SparkContext(conf)


        val res = run(opts.inputPath, opts.numColors.toShort, sc).map{ case (u, v, w) => u + "\t" + v + "\t" + w }

        if(opts.outputPath == null) println("the number of triangles: " + res.count()) // count mode
        else{
          FileSystem.get(sc.hadoopConfiguration).delete(new Path(opts.outputPath), true)
          res.saveAsTextFile(opts.outputPath)
        } // list mode

        sc.stop()

      case None =>
    }
  }

  /**
    * submit the spark job.
    *
    * @param inputPath input file path.
    * @param numColors the number of colors
    * @param sc spark context.
    * @return an RDD containing triangles
    */
  def run(inputPath: String, numColors: Short, sc: SparkContext): RDD[(Long, Long, Long)] = {

    val partPath = inputPath + ".part/"

    val fs = FileSystem.get(sc.hadoopConfiguration)
    fs.mkdirs(new Path(partPath))

    //find numNodes
    val numNodes = sc.broadcast(sc.sequenceFile[Long,Long](inputPath).map { case (u, v) =>
      if (u < v) v else u
    }.max() + 1)

    //Pre-partition of PTE
    partitionGraph(inputPath, partPath, numColors, sc)

    //Triangle enumeration of PTE
    enumerateTriangles(partPath, numColors, numNodes, sc)
  }

  /**
    * partition the input graph.
    * The partitioned graph are stored in partPath
    * A partition with colors i and j is stored in partPath/edgeset-{i*numColors + j} as a binary file
    * whose format is as follows:
    * node1 #out-neighbors-of-node1 out-neighbor1 out-neighbor2 ...
    * node2 #out-neighbors-of-node2 out-neighbor1 ...
    *
    * @param inputPath input file path
    * @param partPath partition file path
    * @param numColors the number of colors
    * @param sc spark context
    */
  def partitionGraph(inputPath: String, partPath: String, numColors: Short, sc: SparkContext): Unit = {
    sc.sequenceFile[Long,Long](inputPath).map{ case (u, v) =>

      val u_local = u / numColors
      val v_local = v / numColors
      val u_color = (u - u_local * numColors).toShort
      val v_color = (v - v_local * numColors).toShort

      ((u_color * numColors + v_color).toShort, (u_local << 32) | v_local)

    }
      .partitionBy(new PTEPartitioner(numColors))
      .foreachPartition { it =>

        var blockColor: Short = -1


        val edgeList = it.map { item =>
          blockColor = item._1
          item._2
        }.toArray

        util.Arrays.sort(edgeList)


        val fs = FileSystem.get(new Configuration)
        val out = new Output(
          new SnappyOutputStream(
            fs.create(
              new Path(partPath + "/edgeset-" + blockColor)
            ).getWrappedStream)
        )

        var uStart = 0
        var uEnd = 0
        while (uStart < edgeList.length) {

          val edge = edgeList(uStart)
          val u = (edge >>> 32).toInt
          val limit = (u + 1).toLong << 32

          try {
            while (edgeList(uEnd) < limit) uEnd += 1
          } catch {case _: ArrayIndexOutOfBoundsException =>}

          out.writeInt(u, true)
          out.writeInt(uEnd - uStart, true)

          for (i <- uStart until uEnd) {
            out.writeInt(edgeList(i).toInt, true)
          }

          uStart = uEnd
        }

        out.close()

      }
  }


  /**
    * Find triangles.
    *
    * @param partPath partitioned file path
    * @param numColors the number of colors
    * @param numNodes the number of nodes in the input graph
    * @param sc spark context
    * @return an RDD containing triangles
    */
  def enumerateTriangles(partPath: String,
                         numColors: Short, numNodes: Broadcast[Long], sc: SparkContext): RDD[(Long, Long, Long)]= {

    val set_subproblems = SeedGenerator.dual(numColors, 3).map { arr => arr(0) -> arr }
      .zipWithIndex
      .map { case (a, b) => (b, a) }

    val parts = sc.parallelize(set_subproblems).partitionBy(new PSESubproblemPartitioner(set_subproblems.length))

    parts.mapPartitions { it =>

      printGC()

      // the first item of the iterator contains the problemGroup.
      val part = new Path(partPath)

      val _problem = it.next()

      val problemGroup = _problem._2._1
      val problems = _problem._2._2

      val colors: Array[Short] = (0 until numColors).filter { i => (problemGroup & 1 << i) != 0 }.map{_.toShort}.toArray

      val nodeSize = numNodes.value
      val nodeSize_edgeSet = (nodeSize / numColors + (if (nodeSize % numColors == 0) 0 else 1)).toInt

      val fs = FileSystem.get(new Configuration)

      val count = if(colors.length == 2)
        enumType2(numColors, part, problems, colors, nodeSize_edgeSet, fs)
      else
        enumType3(numColors, part, colors, nodeSize_edgeSet, fs)

      count
    }

  }

  /**
    * Find type-3 triangles
    *
    * @param numColors the number of colors
    * @param partPath partitioned file path
    * @param colors of the given subproblem
    * @param nodeSize_edgeSet the maximum id of a node.
    * @param fs hadoop file system
    * @return triangle iterator
    */
  def enumType3(numColors: Short, partPath: Path, colors: Array[Short], nodeSize_edgeSet: Int,
                fs: FileSystem): Iterator[(Long, Long, Long)] = new Iterator[(Long, Long, Long)]{

    val adjs: Array[Array[Array[Array[Int]]]] = Array.ofDim[Array[Array[Int]]](3, 3)

    var state = 0
    var it_now: Iterator[(Long, Long, Long)] = Iterator.empty

    override def hasNext: Boolean = if(it_now.hasNext) {true} else {
      state match {
        case 0 =>
          state = 1
          adjs(0)(1) = loadAdj(partPath.suffix("/edgeset-" + (colors(0) * numColors + colors(1))), nodeSize_edgeSet, fs)
          adjs(0)(2) = loadAdj(partPath.suffix("/edgeset-" + (colors(0) * numColors + colors(2))), nodeSize_edgeSet, fs)
          adjs(1)(2) = loadAdj(partPath.suffix("/edgeset-" + (colors(1) * numColors + colors(2))), nodeSize_edgeSet, fs)
          it_now = triangleIterator(adjs(0)(1), adjs(0)(2), adjs(1)(2), colors(0), colors(1), colors(2), numColors)
          hasNext
        case 1 =>
          state = 2
          adjs(2)(1) = loadAdj(partPath.suffix("/edgeset-" + (colors(2) * numColors + colors(1))), nodeSize_edgeSet, fs)
          it_now = triangleIterator(adjs(0)(2), adjs(0)(1), adjs(2)(1), colors(0), colors(2), colors(1), numColors)
          hasNext
        case 2 =>
          state = 3
          adjs(1)(0) = loadAdj(partPath.suffix("/edgeset-" + (colors(1) * numColors + colors(0))), nodeSize_edgeSet, fs)
          it_now = triangleIterator(adjs(1)(0), adjs(1)(2), adjs(0)(2), colors(1), colors(0), colors(2), numColors)
          hasNext
        case 3 =>
          state = 4
          adjs(0)(2) = null
          System.gc()
          adjs(2)(0) = loadAdj(partPath.suffix("/edgeset-" + (colors(2) * numColors + colors(0))), nodeSize_edgeSet, fs)
          it_now = triangleIterator(adjs(1)(2), adjs(1)(0), adjs(2)(0), colors(1), colors(2), colors(0), numColors)
          hasNext
        case 4 =>
          state = 5
          adjs(1)(2) = null
          System.gc()
          it_now = triangleIterator(adjs(2)(0), adjs(2)(1), adjs(0)(1), colors(2), colors(0), colors(1), numColors)
          hasNext
        case 5 =>
          state = 6
          adjs(0)(1) = null
          System.gc()
          it_now = triangleIterator(adjs(2)(1), adjs(2)(0), adjs(1)(0), colors(2), colors(1), colors(0), numColors)
          hasNext
        case 6 =>
          adjs(1)(0) = null
          adjs(2)(1) = null
          adjs(2)(0) = null
          System.gc()
          false
      }

    }

    override def next(): (Long, Long, Long) = if(hasNext) it_now.next() else Iterator.empty.next()
  }


  /**
    * Find type-1 and type-2 triangles
    *
    * @param numColors the number of colors
    * @param partPath partitioned file path
    * @param problems to solve
    * @param colors of the given subproblem
    * @param nodeSize_edgeSet the maximum id number of a node.
    * @param fs hadoop file system
    * @return triangle iterator
    */
  def enumType2(numColors: Short, partPath: Path, problems: Array[Long], colors: Array[Short],
                nodeSize_edgeSet: Int, fs: FileSystem): Iterator[(Long, Long, Long)] = {

    val adjs = Array.ofDim[Array[Array[Int]]](2, 2)

    adjs(0)(0) = loadAdj(partPath.suffix("/edgeset-" + (colors(0) * numColors + colors(0))), nodeSize_edgeSet, fs)
    adjs(0)(1) = loadAdj(partPath.suffix("/edgeset-" + (colors(0) * numColors + colors(1))), nodeSize_edgeSet, fs)
    adjs(1)(0) = loadAdj(partPath.suffix("/edgeset-" + (colors(1) * numColors + colors(0))), nodeSize_edgeSet, fs)
    adjs(1)(1) = loadAdj(partPath.suffix("/edgeset-" + (colors(1) * numColors + colors(1))), nodeSize_edgeSet, fs)

    problems.iterator.flatMap{ problem =>
      val problem_colors = (0 until numColors).filter { i => (problem & 1 << i) != 0 }.toArray

      if(problem_colors.length == 1){
        val i_or_j = colors.indexOf(problem_colors(0)).toShort
        val xx = adjs(i_or_j)(i_or_j)
        val color = colors(i_or_j)

        triangleIterator(xx, xx, xx, color, color, color, numColors)
      }
      else{
        Iterator(triangleIterator(adjs(0)(0), adjs(0)(1), adjs(0)(1), colors(0), colors(0), colors(1), numColors),
        triangleIterator(adjs(0)(1), adjs(0)(0), adjs(1)(0), colors(0), colors(1), colors(0), numColors),
        triangleIterator(adjs(0)(1), adjs(0)(1), adjs(1)(1), colors(0), colors(1), colors(1), numColors),
        triangleIterator(adjs(1)(0), adjs(1)(0), adjs(0)(0), colors(1), colors(0), colors(0), numColors),
        triangleIterator(adjs(1)(0), adjs(1)(1), adjs(0)(1), colors(1), colors(0), colors(1), numColors),
        triangleIterator(adjs(1)(1), adjs(1)(0), adjs(1)(0), colors(1), colors(1), colors(0), numColors)).flatten
      }
    }

  }

  /**
    * find triangles whose pivot edge is in 'ij', port edge is in 'ik', and starboard edge is in 'jk'
    *
    * @param ij pivot edge set
    * @param ik port edge set
    * @param jk starboard edge set
    * @param i the first color
    * @param j the second color
    * @param k the last color
    * @param numColors the number of colors
    * @return triangles whose pivot edge is in 'ij', port edge is in 'ik', and starboard edge is in 'jk'
    */
  def triangleIterator(ij: Array[Array[Int]], ik: Array[Array[Int]], jk: Array[Array[Int]],
                               i: Short, j: Short, k: Short, numColors: Short): Iterator[(Long, Long, Long)] = {
    ij.indices.iterator.flatMap{ u =>
      if(ij(u) != null){
        val uOrigin = u.toLong * numColors + i
        ij(u).iterator.flatMap{ v =>
          val vOrigin = v.toLong * numColors + j
          intersectIterator(ik(u), jk(v)).map{w => (uOrigin, vOrigin, w.toLong * numColors + k)}
        }
      }
      else Iterator.empty
    }
  }


  /**
    * find intersection of two adj lists
    *
    * @param uAdj an adj list
    * @param vAdj another adj list
    * @return intersection of two adj lists
    */
  def intersectIterator(uAdj: Array[Int], vAdj: Array[Int]): Iterator[Int] = {
    if (uAdj == null || vAdj == null) Iterator.empty
    else {
      new Iterator[Int] {
        var uCur: Int = 0
        var vCur: Int = 0
        val uD: Long = uAdj.length
        val vD: Long = vAdj.length

        var hd: Int = _
        var hdDefined: Boolean = false

        override def hasNext: Boolean = {
          while (!hdDefined && uCur < uD && vCur < vD){
            if (uAdj(uCur) < vAdj(vCur)) uCur += 1
            else if (vAdj(vCur) < uAdj(uCur)) vCur += 1
            else {
              hd = uAdj(uCur)
              hdDefined = true
              uCur += 1
              vCur += 1
            }
          }
          hdDefined
        }

        override def next(): Int = if(hasNext){ hdDefined = false; hd} else Iterator.empty.next
      }
    }
  }

  /**
    * load a graph of an adj format.
    *
    * @param graphPath graph file path
    * @param nodeSize the maximum id number of a node
    * @param fs hadoop file system
    * @return a graph of an adj format
    */
  private def loadAdj(graphPath: Path, nodeSize: Int, fs: FileSystem): Array[Array[Int]] = {
    val t: Long = System.currentTimeMillis
    val adj: Array[Array[Int]] = new Array[Array[Int]](nodeSize)

    val in: Input = new Input(new SnappyInputStream(fs.open(graphPath).getWrappedStream))

    try {
      while (true) {
        val u_local = in.readInt(true)
        val uSize = in.readInt(true)

        adj(u_local) = new Array[Int](uSize)

        val uAdj: Array[Int] = adj(u_local)

        for (i <- 0 until uSize) {
          uAdj(i) = in.readInt(true)
        }
      }
    }
    catch {
      case _: KryoException =>
    }

    in.close()

    PTE.logger.info(f"loading[${graphPath.toString}]: ${(System.currentTimeMillis - t) / 1000.0}%.3fs " +
      f"(free memory: ${availableMemoryBytes(true) / 1000000.0}%.2fMB)")

    adj
  }

  /**
    * print gc type
    */
  def printGC(): Unit = {
    val gcMxBeans = ManagementFactory.getGarbageCollectorMXBeans.iterator

    PTE.logger.info("garbage collector info=====")
    while (gcMxBeans.hasNext) {
      val gcMxBean = gcMxBeans.next()

      PTE.logger.info(gcMxBean.getName)
      PTE.logger.info(gcMxBean.getObjectName)
    }
    PTE.logger.info("===========================")
  }


  /**
    * get available memory size
    * @param gc if true, operate a system garbage collector
    * @return available memory size
    */
  def availableMemoryBytes(gc: Boolean): Long = {
    if (gc) {
      System.gc()
      System.gc()
    }
    Runtime.getRuntime.maxMemory - (Runtime.getRuntime.totalMemory - Runtime.getRuntime.freeMemory)
  }

  class PSESubproblemPartitioner(numSubproblems: Int) extends Partitioner {
    def getPartition(key: Any): Int = key.asInstanceOf[Int]
    def numPartitions: Int = numSubproblems
  }
}

object PTE{
  private val logger = Logger.getLogger(getClass)
  logger.setLevel(Level.INFO)

  def main(args: Array[String]): Unit = new PTE().runByCLI(args)
}


