/**
  * PegasusN: Peta-Scale Graph Mining System (Pegasus v3.0)
  * Authors: Chiwan Park, Ha-Myung Park, U Kang
  *
  * Copyright (c) 2018, Ha-Myung Park, Chiwan Park, and U Kang
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of the Seoul National University nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * -------------------------------------------------------------------------
  * File: PTE_sc.java
  * - Find Triangles in a graph.
  * Version: 3.0
  */
package pegasus.spark.triangle

import java.lang.management.ManagementFactory
import java.util

import com.esotericsoftware.kryo.KryoException
import com.esotericsoftware.kryo.io.{Input, Output}
import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.{FileSystem, Path}
import org.apache.log4j.{Level, Logger}
import org.apache.spark._
import org.apache.spark.broadcast.Broadcast
import org.xerial.snappy.{SnappyInputStream, SnappyOutputStream}
import pegasus.spark.subgraph.SeedGenerator

/**
  * Created by hmpark on 17. 2. 27.
  */
object PTE_sc {

  Logger.getLogger("org").setLevel(Level.OFF)
  Logger.getLogger("akka").setLevel(Level.OFF)

  val logger = Logger.getLogger(PTE_sc.getClass)

  def main(args: Array[String]): Unit = {

    val inputPath = args(0)
    val numColors = args(1).toShort

    val local = if(args.length >= 3) args(2).toBoolean else false

    logger.info(f"inputPath: $inputPath, numColors: $numColors")

    run(inputPath, numColors, local)

  }

  def run(inputPath: String, numColors: Short, local: Boolean): Long = {
    //
    // start up, count num nodes in the input graph
    var time_startup = System.currentTimeMillis()

    val partPath = inputPath + ".part/"

    val conf = new SparkConf().setAppName("[PTE]" + inputPath) //.setMaster("local")
      .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
      .set("spark.kryo.registrator", "pegasus.spark.subgraph.MyRegistrator")
      .set("spark.kryoserializer.buffer.mb", "24")

    if(local) conf.setMaster("local[2]")

    val sc = new SparkContext(conf)

    val fs = FileSystem.get(sc.hadoopConfiguration)
    fs.mkdirs(new Path(partPath))



    //find numNodes
    val numNodes = sc.broadcast(sc.sequenceFile[Long,Long](inputPath).map { case (u, v) =>
      if (u < v) v else u
    }.max() + 1)


    time_startup = System.currentTimeMillis() - time_startup






    //
    //Pre-partition of PSE
    var time_partition = System.currentTimeMillis()

    val edgeBlocks = partitionGraph(inputPath, numColors, sc)

    time_partition = System.currentTimeMillis() - time_partition





    //
    //Subgraph enumeration of PSE
    var time_enumeration = System.currentTimeMillis()

    val numSubgraphs = enumerateTriangles(partPath, numColors, numNodes, sc)

    time_enumeration = System.currentTimeMillis() - time_enumeration

    printf("PSE-spark\t%s\t%.2f\t%.2f\t%.2f\t%d\n",
      inputPath, time_startup / 1000.0, time_partition / 1000.0, time_enumeration / 1000.0, numSubgraphs)


    sc.stop()


    numSubgraphs
  }







  def enumerateTriangles(partPath: String,
                         numColors: Short, numNodes: Broadcast[Long], sc: SparkContext): Long = {

    val numTriangles = sc.longAccumulator("numTriangles")

    val set_subproblems = SeedGenerator.dual(numColors, 3).map { arr => arr(0) -> arr }
      .zipWithIndex
      .map { case (a, b) => (b, a) }

    val parts = sc.parallelize(set_subproblems).partitionBy(new PSESubproblemPartitioner(set_subproblems.length))

    parts.foreachPartition { it =>

      printGC()

      // in the first item of the iterator, the problemGroup is there.

      val part = new Path(partPath)

      val _problem = it.next()

      val problemGroup = _problem._2._1
      val problems = _problem._2._2

      val colors = (0 until numColors).filter { i => (problemGroup & 1 << i) != 0 }.toArray

      val numLocalColors = colors.length

      val nodeSize = numNodes.value
      val nodeSize_edgeSet = (nodeSize / numColors + (if (nodeSize % numColors == 0) 0 else 1)).toInt

      val fs = FileSystem.get(new Configuration)

      val count = if(colors.length == 2)
        enumType2(numColors, part, problems, colors, nodeSize_edgeSet, fs)
      else
        enumType3(numColors, part, colors, nodeSize_edgeSet, fs)

      numTriangles.add(count)
    }

    logger.info("Triangle enumerate complete.")
    numTriangles.value
  }

  def enumType3(numColors: Short, part: Path, colors: Array[Int], nodeSize_edgeSet: Int, fs: FileSystem): Long = {
    var count = 0L

    val adjs = Array.ofDim[Array[Array[Int]]](3, 3)

    adjs(0)(1) = loadAdj(part.suffix("/edgeset-" + (colors(0) * numColors + colors(1))), nodeSize_edgeSet, fs)
    adjs(0)(2) = loadAdj(part.suffix("/edgeset-" + (colors(0) * numColors + colors(2))), nodeSize_edgeSet, fs)
    adjs(1)(2) = loadAdj(part.suffix("/edgeset-" + (colors(1) * numColors + colors(2))), nodeSize_edgeSet, fs)

    count += countTriangles(adjs(0)(1), adjs(0)(2), adjs(1)(2))

    adjs(2)(1) = loadAdj(part.suffix("/edgeset-" + (colors(2) * numColors + colors(1))), nodeSize_edgeSet, fs)

    count += countTriangles(adjs(0)(2), adjs(0)(1), adjs(2)(1))

    adjs(1)(0) = loadAdj(part.suffix("/edgeset-" + (colors(1) * numColors + colors(0))), nodeSize_edgeSet, fs)

    count += countTriangles(adjs(1)(0), adjs(1)(2), adjs(0)(2))

    adjs(0)(2) = null
    System.gc()

    adjs(2)(0) = loadAdj(part.suffix("/edgeset-" + (colors(2) * numColors + colors(0))), nodeSize_edgeSet, fs)

    count += countTriangles(adjs(1)(2), adjs(1)(0), adjs(2)(0))

    adjs(1)(2) = null
    System.gc()

    count += countTriangles(adjs(2)(0), adjs(2)(1), adjs(0)(1))

    adjs(0)(1) = null
    System.gc()

    count += countTriangles(adjs(2)(1), adjs(2)(0), adjs(1)(0))

    adjs(1)(0) = null
    adjs(2)(1) = null
    adjs(2)(0) = null
    System.gc()
    count
  }

  def enumType2(numColors: Short, part: Path, problems: Array[Long], colors: Array[Int],
                nodeSize_edgeSet: Int, fs: FileSystem): Long = {

    var count = 0L
    val adjs = Array.ofDim[Array[Array[Int]]](2, 2)

    adjs(0)(0) = loadAdj(part.suffix("/edgeset-" + (colors(0) * numColors + colors(0))), nodeSize_edgeSet, fs)
    adjs(0)(1) = loadAdj(part.suffix("/edgeset-" + (colors(0) * numColors + colors(1))), nodeSize_edgeSet, fs)
    adjs(1)(0) = loadAdj(part.suffix("/edgeset-" + (colors(1) * numColors + colors(0))), nodeSize_edgeSet, fs)
    adjs(1)(1) = loadAdj(part.suffix("/edgeset-" + (colors(1) * numColors + colors(1))), nodeSize_edgeSet, fs)


    problems.foreach { problem =>
      val problem_colors = (0 until numColors).filter { i => (problem & 1 << i) != 0 }.toArray

      if (problem_colors.length == 1) {
        val i_or_j = colors.indexOf(problem_colors(0))
        val xx = adjs(i_or_j)(i_or_j)

        logger.info(s"solve (${i_or_j})")

        count += countTriangles(xx, xx, xx)

      }
      else {
        // two colors.
        count += countTriangles(adjs(0)(0), adjs(0)(1), adjs(0)(1))
        count += countTriangles(adjs(0)(1), adjs(0)(0), adjs(1)(0))
        count += countTriangles(adjs(0)(1), adjs(0)(1), adjs(1)(1))
        count += countTriangles(adjs(1)(0), adjs(1)(0), adjs(0)(0))
        count += countTriangles(adjs(1)(0), adjs(1)(1), adjs(0)(1))
        count += countTriangles(adjs(1)(1), adjs(1)(0), adjs(1)(0))
      }
    }

    count
  }

  private def countTriangles(ij: Array[Array[Int]], ik: Array[Array[Int]], jk: Array[Array[Int]]): Long = {
    val t: Long = System.currentTimeMillis

    var count = 0L
    var u = 0
    val nodeSize = ij.length
    for(u <- ij.indices){
      if (ij(u) != null) for (v <- ij(u)) {
        count += countIntersect(ik(u), jk(v))
      }
    }

    System.out.println("counting: " + ((System.currentTimeMillis - t) / 1000.0) + "s")
    count
  }

  private def countIntersect(uAdj: Array[Int], vAdj: Array[Int]): Long = {

    if (uAdj == null || vAdj == null) 0
    else {
      var count = 0L
      var uCur: Int = 0
      var vCur: Int = 0
      val uD: Long = uAdj.length
      val vD: Long = vAdj.length
      while (uCur < uD && vCur < vD){
        if (uAdj(uCur) < vAdj(vCur)) uCur += 1
        else if (vAdj(vCur) < uAdj(uCur)) vCur += 1
        else {
          count += 1
          uCur += 1
          vCur += 1
        }
      }
      count
    }
  }



  def partitionGraph(inputPath: String, numColors: Short, sc: SparkContext): Unit = {
    sc.sequenceFile[Long,Long](inputPath).map{ case (u, v) =>

      val u_local = u / numColors
      val v_local = v / numColors
      val u_color = (u - u_local * numColors).toShort
      val v_color = (v - v_local * numColors).toShort

      ((u_color * numColors + v_color).toShort, (u_local << 32) | v_local)

    }
      .partitionBy(new PTEPartitioner(numColors))
      .foreachPartition { it =>

        var blockColor: Short = -1


        val edgeList = it.map { item =>
          blockColor = item._1
          item._2
        }.toArray

        util.Arrays.sort(edgeList)


        val fs = FileSystem.get(new Configuration)
        val out = new Output(
          new SnappyOutputStream(
            fs.create(
              new Path(inputPath + ".part/edgeset-" + blockColor)
            ).getWrappedStream)
        )

        var uStart = 0
        var uEnd = 0
        while (uStart < edgeList.length) {

          val edge = edgeList(uStart)
          val u = (edge >>> 32).toInt
          val limit = (u + 1).toLong << 32

          try {
            while (edgeList(uEnd) < limit) uEnd += 1
          } catch {case ignored: ArrayIndexOutOfBoundsException =>}

          out.writeInt(u, true)
          out.writeInt(uEnd - uStart, true)

          for (i <- uStart until uEnd) {
            out.writeInt(edgeList(i).toInt, true)
          }

          uStart = uEnd
        }

        out.close()

      }
  }

  private def loadAdj(graphPath: Path, nodeSize: Int, fs: FileSystem): Array[Array[Int]] = {
    val t: Long = System.currentTimeMillis
    val adj: Array[Array[Int]] = new Array[Array[Int]](nodeSize)

    val in: Input = new Input(new SnappyInputStream(fs.open(graphPath).getWrappedStream))

    try {
      while (true) {
        val u_local = in.readInt(true)
        val uSize = in.readInt(true)

        adj(u_local) = new Array[Int](uSize)

        val uAdj: Array[Int] = adj(u_local)

        for (i <- 0 until uSize) {
          uAdj(i) = in.readInt(true)
        }
      }
    }
    catch {
      case e: KryoException =>
    }

    in.close()

    logger.info(f"loading[${graphPath.toString}]: ${(System.currentTimeMillis - t) / 1000.0}%.3fs " +
      f"(free memory: ${availableMemoryBytes(true) / 1000000.0}%.2fMB)")

    adj
  }

  def printGC(): Unit = {
    val gcMxBeans = ManagementFactory.getGarbageCollectorMXBeans.iterator

    logger.info("garbage collector info=====")
    while (gcMxBeans.hasNext) {
      val gcMxBean = gcMxBeans.next()

      logger.info(gcMxBean.getName)
      logger.info(gcMxBean.getObjectName)
    }
    logger.info("===========================")
  }

  def availableMemoryBytes(gc: Boolean): Long = {
    if (gc) {
      System.gc()
      System.gc()
    }
    Runtime.getRuntime.maxMemory - (Runtime.getRuntime.totalMemory - Runtime.getRuntime.freeMemory)
  }

  class PSESubproblemPartitioner(numSubproblems: Int) extends Partitioner {
    def getPartition(key: Any): Int = key.asInstanceOf[Int]
    def numPartitions: Int = numSubproblems
  }

}



class PTEPartitioner(numColors: Short) extends Partitioner {
  def getPartition(key: Any): Int = {
    key.asInstanceOf[Short]
  }
  def numPartitions: Int = { numColors * numColors }
}




















