package pegasus.web

import java.util.StringTokenizer

import fs2.Task
import fs2.interop.cats._
import org.apache.hadoop._
import org.apache.hadoop.fs.{FileSystem, LocalFileSystem}
import org.apache.hadoop.hdfs.DistributedFileSystem
import org.apache.spark.storage.StorageLevel
import org.apache.spark.ui.jobs.JobProgressListener
import org.apache.spark.{SparkConf, SparkContext}
import org.http4s._
import org.http4s.dsl._
import org.http4s.server.blaze.BlazeBuilder
import org.http4s.server.middleware.CORS
import org.http4s.util.StreamApp
import pegasus.spark.cc.PACC
import pegasus.spark.diameter.DiameterComputation
import pegasus.spark.graphgen.{KroneckerGenerator, RMATGenerator, RandomGenerator}
import pegasus.spark.labelprop.LabelPropagationComputation
import pegasus.spark.netray.{NetRayScatter, NetRaySpy}
import pegasus.spark.pagerank.PageRankComputation
import pegasus.spark.sssp.SSSPComputation
import pegasus.spark.subgraph.PSE
import spray.json._

object MyJsonProtocol extends DefaultJsonProtocol {
  case class JobInfo(jobId: Int, jobGroup: String, name: String,
                     numTasks: Int, numCompletedTasks: Int,
                     numFailedTasks: Int, numSkippedTasks: Int,
                     status: String
                    )

  implicit val personFormat = jsonFormat8(JobInfo)
}

import pegasus.web.MyJsonProtocol._

object Web extends StreamApp {



  object ParamLevel extends QueryParamDecoderMatcher[Int]("level")
  object ParamNumEdges extends QueryParamDecoderMatcher[Long]("numedges")
  object ParamA extends QueryParamDecoderMatcher[Double]("a")
  object ParamB extends QueryParamDecoderMatcher[Double]("b")
  object ParamC extends QueryParamDecoderMatcher[Double]("c")
  object ParamD extends QueryParamDecoderMatcher[Double]("d")
  object ParamOutput extends QueryParamDecoderMatcher[String]("output")
  object ParamInput extends QueryParamDecoderMatcher[String]("input")
  object ParamSeed extends QueryParamDecoderMatcher[String]("seed")
  object ParamNumNodes extends QueryParamDecoderMatcher[Long]("numnodes")
  object ParamMaxIter extends QueryParamDecoderMatcher[Int]("maxiter")
  object ParamSourceVertexId extends QueryParamDecoderMatcher[Long]("sid")
  object ParamCliqueNum extends QueryParamDecoderMatcher[Int]("clqnum")
  object ParamCycleLength extends QueryParamDecoderMatcher[Int]("cyclelen")
  object ParamPatternString extends QueryParamDecoderMatcher[String]("pattern")
  object ParamTargetWidth extends QueryParamDecoderMatcher[Int]("target_width")
  object ParamLogscale extends QueryParamDecoderMatcher[Boolean]("logscale")
  object ParamInput1 extends QueryParamDecoderMatcher[String]("input1")
  object ParamInput2 extends QueryParamDecoderMatcher[String]("input2")

  val conf = new SparkConf().setAppName("PegasusN").setMaster("local[20]")
    .set("spark.executor.cores", "20")
  val sc = new SparkContext(conf)

  val jobListener = new JobProgressListener(conf)
  sc.addSparkListener(jobListener)

  val GRAPH_PATH = "file:///tmp/pegasusn/graphs"
  val RESULT_PATH = "file:///tmp/pegasusn/results"

  sc.hadoopConfiguration.set("fs.hdfs.impl", classOf[DistributedFileSystem].getName)
  sc.hadoopConfiguration.set("fs.file.impl", classOf[LocalFileSystem].getName)

  def static(file: String, request: Request) =
    StaticFile.fromResource("/web-client/" + file, Some(request)).getOrElseF(NotFound())

  val helloWorldService = HttpService {

    case request @ GET -> Root / path if List(".js", ".css", ".map", ".html", ".webm", ".png").exists(path.endsWith) =>
      static(path, request)

    case GET -> Root / "api" / "job-info" => {

      val all = jobListener.activeJobs.values.toArray.reverseIterator ++
        jobListener.completedJobs.reverseIterator
      val allJson = all.map{x =>
        JobInfo(x.jobId, x.jobGroup.getOrElse(""),
          jobListener.stageIdToInfo(x.stageIds.last).name,
          x.numTasks, x.numCompletedTasks,
          x.numFailedTasks, x.numSkippedTasks, x.status.toString)
      }.toArray.toJson

      Ok(allJson.prettyPrint)
    }

    case GET -> Root / "api" / "result-list" => {
      val xx = FileSystem.get(sc.hadoopConfiguration).listStatus(new fs.Path(RESULT_PATH))
      Ok(xx.map{x => x.getPath.getName}.toJson.prettyPrint)
    }

    case GET -> Root / "api" / "graph-list" => {
      val xx = FileSystem.get(sc.hadoopConfiguration).listStatus(new fs.Path(GRAPH_PATH))
      Ok(xx.map{x => x.getPath.getName}.toJson.prettyPrint)
    }

    case GET -> Root / "api" / "spark-app-id" =>
      Ok(sc.applicationId);

    case GET -> Root / "api" / "graphgen" / "rmat" :?
      ParamLevel(level) +& ParamNumEdges(numEdges) +&
        ParamA(a) +& ParamB(b) +& ParamC(c) +& ParamD(d) +& ParamOutput(out) => {

      val abcd = (a, b, c, d)
      val numTasks = 60
      val noise = 0
      val output = GRAPH_PATH + "/" + out

      sc.setJobGroup("R-MAT", s"level: $level, numEdges: $numEdges, abcd: $abcd")
      RMATGenerator.run(level, numEdges, numTasks, abcd, noise, sc)
        .map { case (u, v) => u + "\t" + v }.saveAsTextFile(output)

      Ok(s"graphgen/rmat. lv=$level, edges=$numEdges, abcd=$abcd, out=$output\n")

    }

    case GET -> Root / "api" / "graphgen" / "kronecker" :?
      ParamLevel(level) +& ParamSeed(seed) +& ParamOutput(out) => {

      val numTasks = 60
      val seedGraph = seed.split(",").map{y=> val z = y.split(" "); (z(0).toInt, z(1).toInt)}
      val output = GRAPH_PATH + "/" + out

      sc.setJobGroup("Kronecker", s"level: $level")
      KroneckerGenerator.run(level, seedGraph, numTasks, sc)
        .map { case (u, v) => u + "\t" + v }.saveAsTextFile(output)

      Ok(s"graphgen/kronecker. lv=$level, seed=$seed, out=$output")
    }

    case GET -> Root / "api" / "graphgen" / "random" :?
      ParamNumNodes(nodes) +& ParamNumEdges(edges) +& ParamOutput(out) => {

      val numTasks = 60
      val output = GRAPH_PATH + "/" + out

      sc.setJobGroup("Random", s"$output")
      RandomGenerator.run(nodes, edges, numTasks, sc)
        .map { case (u, v) => u + "\t" + v }.saveAsTextFile(output)

      Ok(s"graphgen/random. nodes=$nodes, edges=$edges, out=$output")
    }

    case GET -> Root / "api" / "gsa" / "pagerank" :?
      ParamMaxIter(maxiter) +& ParamInput(in) +& ParamOutput(out) => {

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      sc.setJobGroup("PageRank", out)
      val config = PageRankComputation.Config(input = input, output = output, maxIterations = maxiter, numBlocks = 4)
      val top10prs = PageRankComputation.run(sc, config)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output + ".tmp"), true)

      val idjson = "\"id\""
      val prjson = "\"pr\""
      Ok(s"[${top10prs.map(v => s"{$idjson: ${v._1}, $prjson: ${v._2}}").mkString(", ")}]")
    }

    case GET -> Root / "api" / "gsa" / "diameter" :?
      ParamInput(in) +& ParamOutput(out) => {
      
      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out
      
      sc.setJobGroup("Diameter", out)
      val config = DiameterComputation.Config(input = input, output = output, numBlocks = 4)
      val (avgDiameter, effDiameter, dist) = DiameterComputation.run(sc, config)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output + ".tmp"), true)
      
      val avgJson = "\"avg\""
      val effJson = "\"eff\""
      val distJson = "\"dist\""
      val radiiJson = "\"radii\""
      val countJson = "\"count\""
      Ok(s"{$avgJson: $avgDiameter, $effJson: $effDiameter, $distJson: [${dist.sortBy(_._1)
        .map(d => s"{$radiiJson: ${d._1}, $countJson: ${d._2}}").mkString(", ")}]}")
    }

    case GET -> Root / "api" / "gsa" / "sssp" :?
      ParamSourceVertexId(sid) +& ParamInput(in) +& ParamOutput(out) => {
      
      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out
      
      sc.setJobGroup("SSSP", out)
      val config = SSSPComputation.Config(input = input, output = output, sourceVertex = sid, numBlocks = 4)
      val (reachableNodes, _) = SSSPComputation.run(sc, config)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output + ".tmp"), true)
      
      val reachableNodesJson = "\"reachableNodes\""
      Ok(s"{$reachableNodesJson: $reachableNodes}")
    }

    case GET -> Root / "api" / "gsa" / "rwr" :?
      ParamSourceVertexId(sid) +& ParamMaxIter(maxiter) +& ParamInput(in) +& ParamOutput(out) => {
      
      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      sc.setJobGroup("RWR", out)
      val config = PageRankComputation.Config(
        input = input, output = output, maxIterations = maxiter, numBlocks = 4, sourceVertex = Some(sid))
      val top10prs = PageRankComputation.run(sc, config)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output + ".tmp"), true)
      
      val idjson = "\"id\""
      val prjson = "\"pr\""
      Ok(s"[${top10prs.map(v => s"{$idjson: ${v._1}, $prjson: ${v._2}}").mkString(", ")}]")
    }

    case GET -> Root / "api" / "gsa" / "cc" :?
      ParamInput(in) +& ParamOutput(out) => {

      val localThreshold = 1000
      val numPartitions = 40
      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      sc.setJobGroup("Connected Components", s"output: $output")
      val result = new PACC().run(input, localThreshold, numPartitions, sc).persist(StorageLevel.MEMORY_AND_DISK)
      result.map { case (u, v) => u + "\t" + v }.saveAsTextFile(output)
      
      val ccCount = result.map(_._2).distinct().count()
      val top10CCs = result.map(v => (v._2, 1L)).reduceByKey(_ + _).sortBy(_._2, ascending = false).take(10)

      FileSystem.get(sc.hadoopConfiguration)
        .delete(new fs.Path(input + ".pacc.tmp"), true)
      
      val top10CCJson ="\"top10\""
      val idJson = "\"id\""
      val countJson = "\"count\""
      Ok(s"{$top10CCJson: [${top10CCs.map(v => s"{$idJson: ${v._1}, $countJson: ${v._2}}").mkString(", ")}], " +
        s"$countJson: $ccCount}")
    }

    case GET -> Root / "api" / "gsa" / "lp" :?
      ParamMaxIter(maxiter) +& ParamInput(in) +& ParamOutput(out) => {

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      sc.setJobGroup("LabelProp", out)
      val config = LabelPropagationComputation.Config(
        input = input, output = output, maxIterations = maxiter, numBlocks = 4)
      val (commCount, top10CommDist) = LabelPropagationComputation.run(sc, config)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output + ".tmp"), true)
      
      val idJson = "\"id\""
      val countJson = "\"count\""
      val commJson = "\"communities\""
      Ok(s"{$countJson: $commCount, $commJson: [${top10CommDist
        .map(v => s"{$idJson: ${v._1}, $countJson: ${v._2}}").mkString(", ")}]}")
    }

    case GET -> Root / "api" / "se" / "triangle" :?
      ParamInput(in) +& ParamOutput(out) => {

      sc.setJobGroup("Triangle", "")

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output), true)


      val triAccu = sc.longAccumulator

      val pse = new PSE()
      val tt = pse.run(input, "111", 9, sc).flatMap { x => triAccu.add(1); x.map(y => (y, 1L)) }.reduceByKey(_ + _)
        .map { case (a, b) => a + "\t" + b }.saveAsTextFile(output)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(input + ".part/"), true)

      Ok(s"""{\"triangles\":${triAccu.count / 3}}""")
    }

    case GET -> Root / "api" / "se" / "clique" :?
      ParamCliqueNum(clqnum) +& ParamInput(in) +& ParamOutput(out) => {

      sc.setJobGroup("Clique", "")

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out
      val clique_pattern = Array.fill[Int](clqnum * (clqnum-1) / 2)(1).mkString("")

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output), true)

      val accu = sc.longAccumulator

      val pse = new PSE()
      val tt = pse.run(input, clique_pattern, 9, sc).flatMap{x=> accu.add(1); x.map(y=>(y,1L))}.reduceByKey(_ + _)
        .map{case (a, b) => a + "\t" + b}.saveAsTextFile(output)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(input + ".part/"), true)

      Ok(s"""{\"cliques\":${accu.count / clqnum}}""")
    }

    case GET -> Root / "api" / "se" / "cycle" :?
      ParamCycleLength(cyclelen) +& ParamInput(in) +& ParamOutput(out) => {

      sc.setJobGroup("Cycle", "")

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out
      val cycle_pattern = if(cyclelen == 3) "111"
      else if(cyclelen == 4) "110011"
      else if(cyclelen == 5) "1100010011"
      else if(cyclelen == 6) "110000100010011"
      else "0"

      if(cycle_pattern != "0") {

        FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output), true)

        val accu = sc.longAccumulator

        val pse = new PSE()
        val tt = pse.run(input, cycle_pattern, 9, sc)
          .flatMap { x => accu.add(1); x.map(y => (y, 1L)) }.reduceByKey(_ + _)
          .map { case (a, b) => a + "\t" + b }.saveAsTextFile(output)

        FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(input + ".part/"), true)

        Ok(s"""{\"cycles\":${accu.count / cyclelen}}""")
      }
      else
        Ok(s"""{}""")
    }

    case GET -> Root / "api" / "se" / "etc" :?
      ParamPatternString(pattern) +& ParamInput(in) +& ParamOutput(out) => {

      sc.setJobGroup("SubgraphEnumeration", "")

      val input = GRAPH_PATH + "/" + in
      val output = RESULT_PATH + "/" + out

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(output), true)

      val accu = sc.longAccumulator

      val pse = new PSE()
      val tt = pse.run(input, pattern, 9, sc).flatMap{x=> accu.add(1); x.map(y=>(y,1L))}.reduceByKey(_ + _)
        .map{case (a, b) => a + "\t" + b}.saveAsTextFile(output)

      FileSystem.get(sc.hadoopConfiguration).delete(new fs.Path(input + ".part/"), true)

      val numNodes = ((1 + Math.sqrt(1 + 8 * pattern.length))/2).toInt

      Ok(s"""{\"patterns\":${accu.count / numNodes}}""")
    }

    case GET -> Root / "api" / "vis" / "spy" :?
      ParamTargetWidth(target_width) +& ParamLogscale(logscale)
        +& ParamInput(in) => {

      sc.setJobGroup("SpyPlot", s"$in")

      val input = GRAPH_PATH + "/" + in

      val numNodes = sc.textFile(input).map{line =>
        val st = new StringTokenizer(line)
        val u = st.nextToken().toLong
        val v = st.nextToken().toLong
        u max v
      }.reduce{case (a, b) => a max b}

      val adj = NetRaySpy.run(input, numNodes, target_width-1, logscale, sc).collect()

      val width = adj.flatMap{case (u,v,k) => Seq(u,v)}.max + 1

      val arr = Array.ofDim[Long](width, width)

      adj.foreach{case (u, v, k) =>
        arr(u)(v) = k
      }

      Ok(arr.toJson.compactPrint)
    }
    case GET -> Root / "api" / "vis" / "dist" :?
      ParamTargetWidth(target_width) +& ParamLogscale(logscale)
        +& ParamInput(in) => {

      sc.setJobGroup("DistributionPlot", "")

      val input = RESULT_PATH + "/" + in

      val inputRDD = sc.textFile(input).map{line =>
        val st = new StringTokenizer(line)
        st.nextToken()
        val v = st.nextToken().toDouble
        (v, 1.0)
      }.reduceByKey(_ + _)

      val adj = NetRayScatter.run(inputRDD, target_width, logscale, sc)

      Ok(adj.collect().toJson.compactPrint)
    }

    case GET -> Root / "api" / "vis" / "corr" :?
      ParamTargetWidth(target_width) +& ParamLogscale(logscale)
        +& ParamInput1(in1) +& ParamInput2(in2) => {
      sc.setJobGroup("DistributionPlot", "")

      val input1 = RESULT_PATH + "/" + in1
      val input2 = RESULT_PATH + "/" + in2

      val input1RDD = sc.textFile(input1).map{line =>
        val st = new StringTokenizer(line)
        val u = st.nextToken().toLong
        val v = st.nextToken().toDouble
        (u, v)
      }

      val input2RDD = sc.textFile(input2).map{line =>
        val st = new StringTokenizer(line)
        val u = st.nextToken().toLong
        val v = st.nextToken().toDouble
        (u, v)
      }

      val joined = input1RDD.join[Double](input2RDD).map(_._2)



      val adj = NetRayScatter.run(joined, target_width, logscale, sc)

      Ok(adj.collect().toJson.compactPrint)
    }
  }

  override def stream(args: List[String]): fs2.Stream[Task, Nothing] = {
    BlazeBuilder
      .bindHttp(8084, "0.0.0.0")
      .mountService(CORS(helloWorldService), "/")
      .serve
  }
}

